#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       PyCI.py and all related original code...
#
#       Copyright 2009 Dan McDougall <YouKnowWho@YouKnowWhat.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; Version 3 of the License
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, the license can be downloaded here:
#
#       http://www.gnu.org/licenses/gpl.html

# Meta
__version__ = '0.5'
__license__ = "GNU General Public License (GPL) Version 3"
__version_info__ = (0, 5)
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

# Standard Python modules
import os, sys, gettext

# Debugging
#import dowser

# TODO:  Have this try to import each module before sys.path.appends and if successful, remove the respective zip file.
# Note:  The current python-openssl package for OpenWRT is broken.  It is missing rand.so and crypto.so.  I'm trying to figure that out.
#        Once that package is fixed PyCI will do SSL just fine.

# I tried to use python-mini with the extra needed modules in a zip file but gettext and locale
# can't get along without functools which requires a compiled shared library (_functools.so)
# ...so I've commented out the following three lines which tried to keep things smaller that the full python install

# Append our dependencies to Python's import path
sys.path.append('required_modules/beaker.zip')
sys.path.append('required_modules/mako.zip')
sys.path.append('required_modules/routes.zip')
sys.path.append('required_modules/cherrypy.zip')
sys.path.append('required_modules/ipaddr.zip')

try: # Note: The json module is only present in Python 2.6+
    import json
except: # If not found, try importing simplejson which will be required if using Python 2.4 or Python 2.5
    import simplejson as json # simplejson was included in Python 2.6 as 'json'

# This needs to be set up top here since we use _() before __main__
locale_dir = os.path.join(os.getcwd(), 'i18n')
def get_translation():
    try:
        trans = gettext.translation('PyCI', locale_dir)
    except: # Use a null translation (i.e. English) if no language is found matching the current LANG environment variable
        #print "PyCI had to fall back to a NullTranslation"
        trans = gettext.NullTranslations()
    return trans
trans = get_translation()
_ = lambda s: trans.ugettext(s).encode('utf-8') # The encode() fixes UnicodeDecodeError exceptions on the console
# Note:  If your LANG environment variable isn't set to your preferred language you might get some messages in English before __main__

# 3rd party module imports
import cherrypy

# PyCI imports
try:
    from auth import AuthController, require, has_read, has_write
    from plugin_utils import init_plugin_system, get_plugins, Plugin, find_plugins
    from utils import lookup_template, connect_tarball_routes, serve_static, detect_mime_type
    from pyci_db import connect_db, disconnect_db
except:
    try: # Try importing the zipped versions
        sys.path.append('pyci_modules.zip')
        from auth import AuthController, require, has_read, has_write
        from plugin_utils import init_plugin_system, get_plugins, Plugin, find_plugins
        from utils import lookup_template, connect_tarball_routes, serve_static, detect_mime_type
        from pyci_db import connect_db, disconnect_db
    except Exception as e:
        print _("Error: Could not import PyCI's own modules.  It is likely that something is wrong with your PyCI installation.")
        print e

# TODO:  Get this grabbing the preferred language setting from the client's browser
class PyCI(object):
    """This is the base class of PyCI.  It is used by CherryPy to load the index (/), 'static', 'html', and 'gettext_json' requests."""
    _cp_config = {
            'tools.auth.on': True
        }

    @require()
    def index(self):
        trans = get_translation()
        _ = lambda s: trans.ugettext(s)
        template = lookup_template("html", "index.mako")
        plugin_description = _("Index")
        try:
            language = os.environ['LANG']
        except KeyError: # LANG isn't set
            language = u'en_US' # Fall back/default to US English
        rendered_template = template.render_unicode(plugin_name='index', language=language, plugin_object=Plugin, plugin_description=plugin_description, gettext_domain='PyCI', _=_)
        return rendered_template

    #def oldstatic(self, filepath, **kwargs):
        #"""Returns files from within 'static' or static.tgz"""
        #return serve_static('static.tgz', filepath)

    #def html(self, filepath, **kwargs):
        #"""Returns files from within 'html' or html.tgz (which is our currently installed theme)"""
        #return serve_static('html/static.tgz', filepath)

    def static(self, filepath, **kwargs):
        """Returns files from the given filepath.  If the file is gzip-compressed the
        Content-Encoding will be set to 'gzip' and the file will be served as-is.
        under the assumption that the browser can handle gzip compression.

        The serve-gzipped-files-as-is method saves a lot of CPU in compressing and
        decompressing files.  Very nice for embedded systems."""
        if os.path.exists(filepath):
            detect_mime_type(filepath)
            return open(filepath).read()
        else:
            # Check if the file exists with a .gz extension
            if os.path.exists(filepath + '.gz'):
                detect_mime_type(filepath)
                cherrypy.response.headers['Content-Encoding'] = 'gzip'
                return open(filepath+ '.gz').read()
            else:
                raise cherrypy.HTTPError(404)

    def version(self, **kwargs):
        """Replies with the current version of PyCI (in json format)"""
        # Note:  This is used by the Admin/Reboot page to figure out when PyCI comes back up
        version_dict = {'name': 'PyCI', 'version': __version__}
        return json.dumps(version_dict)

    def restart_pyci(self, from_page="/"):
        """Re-initializes the plugin system (i.e. when you change language preferences)"""
        cherrypy.engine.reexec()
        raise cherrypy.HTTPRedirect(from_page or "/")

    def gettext_json(self, domain, locale_dir=None, indent=False, **kwargs):
        """Returns a json-encoded version of the locale in question (for the jQuery gettext plugin).

        Based on code from http://plugins.jquery.com/project/gettext
        Thanks Sabin Iacob!"""
        # Note:  Since we use the LANG environment variable to set the language we don't have to worry gettext's 'languages' option
        try:
            if not locale_dir:
                locale_dir = os.path.join(os.getcwd(), 'i18n')
                if domain == 'PyCI':
                    tr = gettext.translation(domain, locale_dir)
                else:
                    # Check if the domain matches the name of a plugin
                    plugins = find_plugins()
                    for plugin in plugins:
                        if plugin.__module__ == domain:
                            locale_dir = os.path.join(os.getcwd(), 'plugins_enabled/%s/i18n' % plugin.__module__)
                            tr = gettext.translation(domain, locale_dir)
            keys = tr._catalog.keys()
            keys.sort()
            ret = {}
            for k in keys:
                v = tr._catalog[k]
                if type(k) is tuple:
                    if k[0] not in ret:
                        ret[k[0]] = []
                    ret[k[0]].append(v)
                else:
                    ret[k] = v
            # Even though this is correct, jquery.gettext doesn't like it
            #cherrypy.response.headers['Content-Type'] = "application/json"
            return json.dumps(ret, ensure_ascii=False, indent=indent)
        except IOError:
            return None

def setup_plugin_routes(dispatcher):
    plugin_dir = "%s/plugins_enabled" % os.getcwd()
    plugin_list = get_plugins(plugin_dir)
    init_plugin_system({'plugin_path': plugin_dir, 'plugins': plugin_list})
    plugins = find_plugins()
    for plugin in plugins:
        plugin_path = "/%s/:action" % plugin.__name__
        print _("Enabling plugin: %s") % plugin.__name__
        dispatcher.connect(plugin.__module__, plugin_path, controller=plugin())

def setup_routes(root=None):
    """Setup all base routes as well as plugin routes."""
    d = cherrypy.dispatch.RoutesDispatcher()
    if not root:
        root = PyCI()
    auth = AuthController()
    #if os.path.exists('static'):
        #pass
    #else:
        #connect_tarball_routes(d, root, 'static.tgz', action='static')
    #if os.path.exists('html/static'):
        #pass
    #else:
        #connect_tarball_routes(d, root, 'html/static.tgz', action='html')
    setup_plugin_routes(d)
    d.connect('main', ':action', controller=root)
    d.connect('auth', '/auth/:action', controller=auth)
    #print "connecting dowser..."
    #cherrypy.tree.mount(dowser.Root(), '/dowser')
    return d

def verify_db():
    """Verifies that pyci.db exists and contains at least one permission value.
    If there's something amiss this method will attempt to fix it automatically."""
    print _("Verifying the PyCI database...")
    pyci_db = connect_db('pyci.db')
    dbcursor = pyci_db.cursor()

    # These silently set up the database if it doesn't exist
    # Create tables...
    dbcursor.execute('CREATE TABLE IF NOT EXISTS user_permissions (id INTEGER PRIMARY KEY, user TEXT, plugin TEXT, permission TEXT)')
    # "group" is a reserved word in sqlite3 so we use "_group"
    dbcursor.execute('CREATE TABLE IF NOT EXISTS group_permissions (id INTEGER PRIMARY KEY, _group TEXT, plugin TEXT, permission TEXT)')
    dbcursor.execute('CREATE TABLE IF NOT EXISTS pyci_settings (id INTEGER PRIMARY KEY, option TEXT, setting TEXT)')
    # Create indexes to keep things speedy...
    dbcursor.execute('CREATE INDEX IF NOT EXISTS user_index ON user_permissions (user)')
    dbcursor.execute('CREATE INDEX IF NOT EXISTS user_plugin_index ON user_permissions (plugin)')
    dbcursor.execute('CREATE INDEX IF NOT EXISTS group_index ON group_permissions (_group)')
    dbcursor.execute('CREATE INDEX IF NOT EXISTS group_plugin_index ON group_permissions (plugin)')
    dbcursor.execute('CREATE INDEX IF NOT EXISTS settings_index ON pyci_settings (setting)')
    dbcursor.execute('CREATE INDEX IF NOT EXISTS options_index ON pyci_settings (option)')

    # The following sets up the default permissions if the database was just created or accidentally deleted
    # TODO: Make this also set the language preference based on the LANG environment variable
    try: # This will raise an exception if there's no record
        dbcursor.execute('SELECT * FROM user_permissions').fetchone()[0]
        print _("Everything is fine.  Moving on...")
    except:
        print _("Error: The user_permissions table is empty.  Adding default permissions for user 'root'...")
        dbcursor.execute("INSERT INTO user_permissions VALUES (null, 'root', 'all', 'rw')")
        print _("Done adding default root permissions.  Moving on...")
    disconnect_db(pyci_db)

def start(config=None):
    """This starts up the cherrypy webserver"""
    if config:
        cherrypy.config.update(config)

    cherrypy.tree.mount(None, config=conf)
    cherrypy.engine.start()
    cherrypy.engine.block()

def main():
    """Main program method"""

    global conf
    # Unfortunately we can only globally set the language preference on start or restart
    try:
        pyci_db = connect_db('pyci.db')
        dbcursor = pyci_db.cursor()
        language = dbcursor.execute('SELECT setting FROM pyci_settings WHERE option="language"').fetchone()[0]
        disconnect_db(pyci_db)
        os.environ['LANG'] = language
    except: # This just means there's no language set
        pass
    session_dir = "/tmp/pyci_sessions"
    verify_db()
    conf = {
        '/': {
            'request.dispatch': setup_routes(),
            'tools.sessions.on': True,
            'tools.sessions.storage_type': "file",
            'tools.sessions.storage_path': session_dir,
            #'tools.sessions.timeout': 60,
        },
    }

    # Make our session dir if it doesn't already exist
    if not os.path.exists(session_dir):
        os.mkdir(session_dir)
    # Clean up any old .lock files in the session dir to be safe
    for filename in os.listdir(session_dir):
        if filename.endswith('.lock'):
            os.remove(os.path.join(session_dir, filename))

    start(os.path.join(os.getcwd(), 'cherrypy.conf'))

if __name__ == '__main__':
    main()
