# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

import os, sys, time, gettext
try: # Note: The json module is only present in Python 2.6+
    import json
except: # If not found, try importing simplejson which will be required if using Python 2.4 or Python 2.5
    import simplejson as json # simplejson was included in Python 2.6 as 'json'
# 3rd party modules
import cherrypy
# PyCI modules
import plugin_utils
from auth import require, has_read, has_write
from pyci_db import check_user_permission, check_group_permission, change_user_permission, change_group_permission, connect_db, disconnect_db, delete_user_permission_by_id, delete_group_permission_by_id
from utils import lookup_template, search_cells, jqgrid_select_options, plugin_template, jquery_ui_error_msg, jquery_ui_highlight_msg, jqgrid_json
from openwrt import ConfigFileOpenWRT, OpenWRTConfig
from odict import odict
import quagga

"""Admin Interface Plugin (for configuring things like user/group permissions, themes, etc)"""

# TODO:  Make PasswdFile use the make_backup and undo_from_backup methods along with some timing events
# TODO:  Figure out a way to migrate all but the base Admin object out of this file (might need to re-do how base.mako figures out the plugin navigation tree)

# Take care of internationalization
trans = plugin_utils.get_translation(__file__)
_ = lambda s: trans.ugettext(s)

class Services(plugin_utils.Plugin):
    """A plugin for managing various services (usually 3rd party daemons)."""
    display = True
    permissions = ['Services',]
    sort_priority = 1
    navigable_functions = odict([
        (_('Advanced Routing (Quagga)'), 'Quagga'),
        (_('Coova Chilli'), 'CoovaChilli'),
        (_('Cron'), 'Cron'),
        (_('Dropbear SSH'), 'Dropbear'),
        (_('NTP Client'), 'NTPClient'),
        (_('SNMP Agent'), 'SNMPAgent'),
    ])

    # TODO:  Write an SNMP_Conf class that can parse and save /etc/snmp/snmpd.conf
    @require(has_read('Services'))
    def SNMPAgent(self):
        """Render the SNMP Agent configuration page"""
        _template = plugin_template(self, "templates/snmp_agent.mako")
        return self.standard_render(
            template=_template,
            help=_("If you do not have either the snmpd or snmpd-static packages installed you will not be able to use this interface.  For reference, the snmpd-static package is slightly smaller if you wish to save some space when choosing which one to install."),
            long_name=_('SNMP Agent Configuration'),
            description=_("The SNMP Agent can be configured with this interface."),
        )

    @require(has_read('Services'))
    def Dropbear(self):
        """Render the dropbear (SSH daemon) configuration page"""
        _template = plugin_template(self, "templates/dropbear.mako")
        return self.standard_render(
            template=_template,
            help=_("<p>This interface allows you to configure the port dropbear listens on as well as whether or not to permit password authentication.</p> <p>Why would you want to disallow password authentication?  If you had pre-shared SSH keys setup, disabling password authentication would provide an extra security measure.</p>"),
            long_name=_('Dropbear SSH Daemon Configuration'),
            description=_("The dropbear SSH daemon can be configured with this interface."),
        )

    @require(has_read('Services'))
    def dropbear_status(self):
        """Render the dropbear status mini-page."""
        from commands import getoutput
        result = getoutput('ps aux | grep dropbear | grep -v grep')
        process_count = len(result.split('\n'))
        if result:
            return "Dropbear is %s (%s processes)" % (result, process_count)
        else:
            return "Dropbear is not running"

    @require(has_read('Services'))
    def dropbear_conf_tab(self):
        """Render the dropbear configuration tab"""
        _template = plugin_template(self, "templates/dropbear_conf_tab.mako")

        return self.standard_render(
            template=_template,
            dropbear_conf = ConfigFileOpenWRT(filepath='/etc/config/dropbear'),
        )

    @require(has_read('Services'))
    def raw_dropbear_tab(self, **data):
        """Render the raw dropbear file configuration page"""
        _template = plugin_template(self, "templates/raw_dropbear_tab.mako")
        dropbear_conf = ConfigFileOpenWRT(filepath='/etc/config/dropbear')
        raw_dropbear = dropbear_conf.raw()
        del dropbear_conf # Clean up memory

        return self.standard_render(
            template=_template,
            raw_dropbear=raw_dropbear,
        )

    @require(has_read('Services'))
    def dropbear_service_control_tab(self):
        """Render the dropbear service control tab"""
        _template = plugin_template(self, "templates/dropbear_service_control_tab.mako")

        return self.standard_render(
            template=_template,
        )

    @require(has_read('Services'))
    def dropbear_options_help(self, option, **data):
        """Return help text for the provided dropbear option (for jTip)"""
        options_dict = {
            'Port': _('<strong>%s</strong>: The port you want dropbear to listen for connections.  Default: 22') % option,
            'PasswordAuth': _('<strong>%s</strong>: This controls whether or not you want dropbear to accept password authentication.  Without this option checked dropbear will only accept key-based authentication.') % option,
        }
        try:
            return options_dict[option]
        except:
            return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_write('Services'))
    def update_dropbear_conf(self, port, password_auth=None):
        """Save the dropbear config as submitted."""
        try:
            dropbear_conf = ConfigFileOpenWRT(filepath='/etc/config/dropbear')
            dropbear_conf.sections[0]['options']['Port'] = port
            if password_auth:
                dropbear_conf.sections[0]['options']['PasswordAuth'] = password_auth
            else:
                dropbear_conf.sections[0]['options']['PasswordAuth'] = 'off'
            dropbear_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/config/dropbear was saved.'))
        except Exception as e:
            return jquery_ui_error_msg(_('Could not save /etc/config/dropbear.  Please make sure PyCI is running as root.<br /> %s') % e)

    @require(has_write('Services'))
    def update_raw_dropbear(self, dropbear):
        """Save the raw dropbear config file text as submitted"""
        try:
            dropbear_conf = ConfigFileOpenWRT(filepath='/etc/config/dropbear')
            dropbear_conf.save_raw(dropbear)
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/config/dropbear was saved.'))
        except Exception as e:
            return jquery_ui_error_msg(_('Could not save /etc/config/dropbear.  Please make sure PyCI is running as root.<br /> %s') % e)

    @require(has_write('Services'))
    def dropbear_service_control(self, oper):
        """Stop, Start, or Restart dropbear based on oper which should be one of:
        'start'
        'stop'
        'restart'"""
        from utils import service_control
        valid_operations = ['start', 'stop', 'restart']
        if oper in valid_operations:
            result = service_control('dropbear', oper)
            if result == True:
                return jquery_ui_highlight_msg(
                    _('<strong>Success:</strong> "/etc/init.d/dropbear %s" executed successfully.') % oper
                )
            else:
                return jquery_ui_error_msg(
                    _('Could not execute /etc/init.d/dropbear %s.<br />%s') % (oper, result)
                )

    @require(has_read('Services'))
    def snmp_config(self):
        """Render the SNMP Agent configuration tab"""
        _template = plugin_template(self, "templates/snmp_conf.mako")

        return self.standard_render(
            template=_template,
        )

    @require(has_write('Services'))
    def restart_snmpd(self):
        """Save the submitted text (snmpd_conf) to the /etc/snmp/snmpd.conf file"""
        from subprocess import check_call
        if cherrypy.request.method == 'POST': # Only restart if we were POSTED to
            try:
                check_call(["/etc/init.d/snmpd", "restart"])
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> snmpd was restarted.'))
            except CalledProcessError, e:
                return jquery_ui_error_msg(_('"/etc/init.d/smnpd restart" failed: %s') % e)

    @require(has_read('Services'))
    def raw_snmpd_conf(self):
        """Render the raw snmpd_conf configuration tab."""
        _template = plugin_template(self, "templates/raw_snmpd_conf.mako")
        if os.path.exists('/etc/snmp/snmpd.conf'):
            raw_snmpd_conf = open('/etc/snmp/snmpd.conf', 'r').read()
        else:
            raw_snmpd_conf = _("/etc/snmp/snmpd.conf does not exist on this system.")

        return self.standard_render(
            template=_template,
            raw_snmpd_conf=raw_snmpd_conf,
        )

    @require(has_write('Services'))
    def update_raw_snmpd_conf(self, snmpd_conf):
        """Save the submitted text (snmpd_conf) to the /etc/snmp/snmpd.conf file"""
        try:
            f = open('/etc/snmp/snmpd.conf', 'w')
            f.write(snmpd_conf)
            f.close()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/snmp/snmpd.conf was saved.  You must restart snmpd in order for changes to take effect.'))
        except Exception as e:
            return jquery_ui_error_msg(_('Could not save /etc/snmp/snmpd.conf.  Please make sure PyCI is running as root.<br /> %s') % e)

    @require(has_read('Services'))
    def Cron(self):
        """Render the Cron configuration page"""
        _template = plugin_template(self, "templates/cron.mako")

        return self.standard_render(
            template=_template,
            help=_("Each section of the accordion represents a user's crontab file in /etc/crontabs.  You can edit existing crontabs using the 'Crontabs' tab or you can create a new user crontab using the 'New Crontab' tab.  Crontabs are laid out using the following format:<br /><br />* * * * * &lt;command to execute&gt;<br /><br />Each asterisk represents a time frame to execute the task.  The order is as follows: Minute, Hour, Day of the Month, Month, Day of the Week (0 or 7 are valid for Sunday).  For more information on how to use crontabs see the <a href='http://en.wikipedia.org/wiki/Cron'>Wikipedia entry on cron</a>"),
            long_name=_('Cron Configuration'),
            description=_("This tool lets you add or edit crontabs for each user on your system."),

        )

    @require(has_read('Services'))
    def new_crontab(self):
        """Render the Cron configuration page"""
        _template = plugin_template(self, "templates/new_crontab.mako")

        return self.standard_render(
            template=_template,
        )

    @require(has_read('Services'))
    def crontabs(self):
        """Render the crontabs tab"""
        _template = plugin_template(self, "templates/crontabs.mako")
        # Assemble the existing crontabs into a dict...
        crontabs = {}
        for crontab in os.listdir('/etc/crontabs/'):
            raw_crontab = open('/etc/crontabs/%s' % crontab, 'r').read()
            crontabs.update({crontab: raw_crontab})

        return self.standard_render(
            template=_template,
            crontabs=crontabs,
        )

    @require(has_write('Services'))
    def update_raw_crontab(self, username, crontab, new=False, delete=False):
        """Save the raw root crontab file text as submitted"""
        try:
            if new: # This is supposed to be a new crontab but we'll check just in case
                if os.path.exists('/etc/crontabs/%s' % username):
                    return jquery_ui_error_msg(_('A crontab for %s already exists.  You can edit existing crontabs using the "Crontabs" tab.') % username)
            if delete == "true" or delete == "True":
                os.remove('/etc/crontabs/%s' % username)
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/crontabs/%s was deleted.') % username)

            f = open('/etc/crontabs/%s' % username, 'w')
            f.write(crontab)
            f.close()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/crontabs/%s was saved.') % username)
        except Exception as e:
            return jquery_ui_error_msg(_('Could not save /etc/crontabs/%s.  Please make sure PyCI is running as root.<br />%s') % (username, e))

    @require(has_read(permissions[0]))
    def crontab_help(self, option, **data):
        """Return help text for the provided option (for jTip)"""
        options_dict = {
            'Username': _('<strong>%s</strong>: The user associated with this crontab (e.g. "root").<br />The crontab will be saved using this name as the file name: /etc/crontabs/&lt;username&gt;<br /><strong>Note:</strong> If the user in question already has a crontab this form will report an error saying the crontab for this user already exists.') % option,
        }
        try:
            return options_dict[option]
        except:
            return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_read('Services'))
    def NTPClient(self):
        """Render the NTP client configuration page"""
        _template = plugin_template(self, "templates/ntpclient.mako")
        return self.standard_render(
            template=_template,
            help=_("<p>The Network Time Protocol (NTP) client synchronizes the system clock with NTP servers.  This interface allows you to configure the NTP client built into OpenWRT.</p>"),
            long_name=_('NTP Client Configuration'),
            description=_("Use this interface to configure the Network Time Protocol client."),
        )

    @require(has_read('Services'))
    def ntpclient_configuration(self):
        """Render the NTP client configuration page"""
        _template = plugin_template(self, "templates/ntpclient_configuration.mako")
        ntpclient_conf = ConfigFileOpenWRT('/etc/config/ntpclient')
        validation_obj = ntpclient_conf.validation_obj()
        validation_dict = ntpclient_conf.validation_dict(validation_obj)
        return self.standard_render(
            template=_template,
            ntpclient_conf=ntpclient_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def ntpclient_options_meta(self, option=None, section=None, help=True, values=False, **data):
        """By default, return help text for the provided ntpclient option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section.
        If 'section' is provided, attempt to obtain valid values from the specified section
        of the validation file."""
        ntpclient_conf = ConfigFileOpenWRT('/etc/config/ntpclient')
        if values:
            try:
                values_dict = {}
                for value in ntpclient_conf.valid_options(section, option):
                    if value == '0':
                        values_dict.update({value:"false"})
                    elif value == '1':
                        values_dict.update({value:"true"})
                    else:
                        values_dict.update({value:value})
                return json.dumps(values_dict)
            except:
                # This is so we can determine that no list values are available for this option
                # ...in JavaScript: if (!json.empty) { do stuff... }
                return "{empty:true}"
        if help:
            try:
                validation_obj = ntpclient_conf.validation_obj()
                validation_dict = ntpclient_conf.validation_dict(validation_obj)
                return '<strong>%s</strong>: %s' % (option, validation_dict[section][option]['title'])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_write('Network'))
    def update_ntpclient_config(self, config=None, **kwargs):
        """Updates ntpclient settings in /etc/config/ntpclient with the supplied options."""
        if not config:
            return jquery_ui_error_msg(_('Could not update NTP client settings.  Please make sure to pass this function a "config" parameter.'))
        try:
            conf_position = None
            ntpclient_conf = ConfigFileOpenWRT('/etc/config/ntpclient')
            for index, section in enumerate(ntpclient_conf.sections):
                if config in section['config']: # config ntpclient
                    conf_position = index
                    break
                elif 'hostname' in section['options'] and section['options']['hostname'] == config:
                    # config ntpserver
                    conf_position = index
                    break
            if 'delete' in kwargs.keys(): # This interface has been marked for deletion
                del ntpclient_conf.sections[conf_position] # Remove this config section
                ntpclient_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> ntpserver %s has been deleted.') % config)
            ntpclient_conf.sections[conf_position]['options'] = {} # We'll be replacing these
            for key, value in kwargs.iteritems():
                #print "%s\t%s" % (key, value) # For debugging
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    ntpclient_conf.sections[conf_position]['options'].update({key: value})
            ntpclient_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> NTP Client settings have been updated.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not update NTP client settings.<br />%s') % e)

    @require(has_write('Network'))
    def add_ntpserver(self, hostname, port=None, **kwargs):
        """Adds a new ntpserver section to /etc/config/ntpclient"""
        ntpclient_conf = ConfigFileOpenWRT('/etc/config/ntpclient')
        # First make sure there isn't already an ntpserver with this name
        for section in ntpclient_conf.sections:
            if 'hostname' in section['options'] and hostname == section['options']['hostname']:
                return jquery_ui_error_msg(_('An ntpserver already exists with this name.  Please choose a different name and re-submit.'))
        try: # Now try creating the new ntpserver
            options = {}
            options.update({'hostname': hostname})
            if port:
                options.update({'hostname': hostname})
            config_obj = OpenWRTConfig(config=['ntpserver'], options=options)
            ntpclient_conf.sections.append(config_obj)
            ntpclient_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> ntpserver %s was added.') % hostname)
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not add ntpserver.<br />%s') % e)

    @require(has_read('Services'))
    def raw_ntpclient(self):
        """Render the raw ntpclient configuration tab."""
        _template = plugin_template(self, "templates/raw_ntpclient.mako")
        ntpclient_conf = ConfigFileOpenWRT('/etc/config/ntpclient')

        return self.standard_render(
            template=_template,
            raw_ntpclient=ntpclient_conf.raw(),
        )

    @require(has_write('Services'))
    def update_raw_ntpclient(self, ntpclient):
        """Save the submitted text (ntpclient) to the /etc/config/ntpclient file"""
        try:
            ntpclient_conf = ConfigFileOpenWRT('/etc/config/ntpclient')
            ntpclient_conf.save_raw(ntpclient)
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/config/ntpclient was saved.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not save /etc/config/ntpclient. <br />%s') % e)

    @require(has_write('Services'))
    def validate_option(self, config=None, section=None, **kwargs):
        """Returns a json-encoded error message if the given option:value pair does not validate"""
        config_obj = ConfigFileOpenWRT('/etc/config/%s' % config)
        result = None
        # This should only ever loop once
        for key, value in kwargs.items():
            result = config_obj.validate_option(section, key, value)
        if result:
            return json.dumps(result)

    @require(has_read('Services'))
    def Quagga(self):
        """Render the Quagga configuration page"""
        _template = plugin_template(self, "templates/quagga.mako")
        return self.standard_render(
            template=_template,
            # TODO:  Re-word these...
            help=_("<p>This interface allows you to configure the Quagga advanced routing daemon on this system.  Each accordion section represents a Quagga configuration file in /etc/quagga.  These are typically named after the routing protocol they support:  zebra.conf, ospf.conf, etc...</p>"),
            long_name=_('Quagga (Advanced Routing) Configuration'),
            description=_("This interface allows you to connect to any Quagga vtysh running on this system or a remote system.  You can also configure Quagga."),
        )

    @require(has_read('Services'))
    def quagga_shell(self):
        """Render the Quagga shell (vtysh) tab."""
        _template = plugin_template(self, "templates/quagga_shell.mako")
        # Find out which daemons are enabled
        daemons = []
        all_daemons = ['zebra', 'ripd', 'ripngd', 'ospfd', 'ospf6d', 'bgpd']
        etc_quagga = os.listdir('/etc/quagga/')
        for config in etc_quagga:
            if config.rstrip('.conf') in all_daemons:
                daemons.append(config.rstrip('.conf'))
        # Get the read-only passwords for each daemon
        passwords = {}
        for daemon in daemons:
            config = quagga.QuaggaConfig('/etc/quagga/%s.conf' % daemon)
            passwords.update({daemon: config.password})

        return self.standard_render(
            template=_template,
            daemons=daemons,
            passwords=passwords,
        )
# TODO:  Add Quagga Service Control
    @require(has_write('Services'))
    def quagga_execute(self, command, password=None, host="127.0.0.1", port=None, enable=False, enable_password=None, daemon=None, **kw):
        """Execute the given command (and options) via Quagga's vtysh and stream the output"""
        try:
            output = quagga.vtysh_command(command, password, enable=enable, enable_password=enable_password, host=host, port=port, daemon=daemon)
            return output
        except Exception as e:
            return _("Got an exception.  Please make sure that %s is running.<br />%s") % (daemon, e)

    @require(has_read('Services'))
    def quagga_config(self):
        """Render the Quagga configuration tab."""
        _template = plugin_template(self, "templates/quagga_config.mako")
        import quagga
        # Find out which daemons are enabled
        daemons = []
        config_files = []
        all_daemons = ['zebra', 'ripd', 'ripngd', 'ospfd', 'ospf6d', 'bgpd']
        etc_quagga = os.listdir('/etc/quagga/')
        for config in etc_quagga:
            if config.rstrip('.conf') in all_daemons:
                daemons.append(config.rstrip('.conf'))
                config_files.append(quagga.QuaggaConfig('/etc/quagga/%s' % config))

        return self.standard_render(
            template=_template,
            daemons=daemons,
            config_files=config_files,
        )

    @require(has_read('Services'))
    def quagga_options_meta(self, option=None, filepath=None, help=True, values=False, **data):
        """By default, return help text for the provided quagga option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section.
        If 'section' is provided, attempt to obtain valid values from the specified section
        of the validation file."""
        import quagga
        config_file = quagga.QuaggaConfig(filepath)
        if values:
            try:
                values_dict = {}
                for value in config_file.valid_options(option):
                    if value == '0' or value == 'off' or value == 'disabled':
                        values_dict.update({value:"off"})
                    elif value == '1' or value == 'on' or value == 'enabled':
                        values_dict.update({value:"on"})
                    else:
                        values_dict.update({value:value})
                if values_dict:
                    return json.dumps(values_dict)
                else:
                    return "{empty:true}"
            except:
                # This is so we can determine that no list values are available for this option
                # ...in JavaScript: if (!json.empty) { do stuff... }
                return "{empty:true}"
        if help:
            try:
                if '<dl>' in config_file.help[option]:
                    import re
                    term = re.findall('\<dt\>(.*)\<\/dt\>', config_file.help[option])[0]
                    help_text = re.findall('\<dd\>(.*)\<\/dd\>', config_file.help[option])[0]
                    return '<strong>%s</strong>: %s' % (term, help_text)
                else:
                    return '<strong>%s</strong>: %s' % (option, config_file.help[option])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_read('Services'))
    def update_quagga_config(self, filepath, config, **kwargs):
        """Updates the given filepath with the given arguments."""
        try:
            f = open(filepath, 'w')
            f.write(config)
            f.close()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> %s was saved.') % filepath)
        except Exception as e:
            return jquery_ui_error_msg(_('Could not save %s.  Please make sure PyCI is running as root.<br /> %s') % (filepath, e))

    @require(has_read('Services'))
    def CoovaChilli(self):
        """Render the main Coova Chilli page (tabs)."""
        _template = plugin_template(self, "templates/coova_chilli.mako")
        return self.standard_render(
            template=_template,
            help=_("<p>This interface allows you to configure the Coova Chilli HotSpot daemon.</p>"),
            long_name=_('Coova Chilli Configuration'),
            description=_("This interface allows you to configure the Coova Chilli HotSpot daemon."),
        )

    @require(has_read('Services'))
    def coova_chilli_config(self):
        """Render the Coova Chilli configuration tab.
        Note:  If /etc/chilli/config doesn't exist this function will pre-populate it with defaults."""
        _template = plugin_template(self, "templates/coova_chilli_conf.mako")
        from coova_chilli import CoovaChilliConfig
        if not os.path.exists('/etc/chilli/config'):
            from commands import getoutput
            output = getoutput('cp /etc/chilli/defaults /etc/chilli/config')

        return self.standard_render(
            template=_template,
            chilli_config=CoovaChilliConfig(),
        )

    @require(has_read('Services'))
    def coova_chilli_config_raw(self):
        """Render the Coova Chilli raw file edit tab."""
        _template = plugin_template(self, "templates/coova_chilli_conf_raw.mako")
        from coova_chilli import CoovaChilliConfig

        return self.standard_render(
            template=_template,
            raw_coova_chilli=CoovaChilliConfig().raw(),
        )

    @require(has_read('Services'))
    def coova_chilli_status(self):
        """Render the Coova Chilli status tab."""
        _template = plugin_template(self, "templates/coova_chilli_status.mako")

        return self.standard_render(
            template=_template,
        )

    @require(has_read('Services'))
    def coova_chilli_options_meta(self, option=None, help=True, values=False, **data):
        """By default, return help text for the provided Coova Chilli option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section."""
        from coova_chilli import CoovaChilliConfig
        chilli_config = CoovaChilliConfig()
        if values: # No validation or enumeration yet
            return "{empty:true}"

        if help:
            try:
                return '<strong>%s</strong>: %s' % (option, chilli_config.settings_help[option])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_write('Services'))
    def update_raw_coova_chilli(self, coova_chilli):
        """Save the raw Coova Chilli config file text as submitted"""
        try:
            from coova_chilli import CoovaChilliConfig
            chilli_config = CoovaChilliConfig()
            chilli_config.save_raw(coova_chilli)
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/chilli/config was saved.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not save /etc/chilli/config. <br />%s') % e)

    @require(has_write('Services'))
    def update_coova_chilli(self, **kwargs):
        """Updates /etc/chilli/config with the supplied options.  Requires 'config' (i.e. 'dhcp lan' or 'dnsmasq')"""
        try:
            from coova_chilli import CoovaChilliConfig
            chilli_config = CoovaChilliConfig()
            for key, value in kwargs.items():
                if key in chilli_config.settings_list: # Only accept valid options
                    chilli_config.settings.update({key: value})
            chilli_config.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Coova Chilli settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update Coova Chilli settings.<br />%s') % e)

    @require(has_read('Services'))
    def coova_chilli_status_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns the output of 'chilli_query list' in a json format that's compatible with jqGrid."""
        import coova_chilli
        chilli_status = coova_chilli.coova_chilli_status()
        header = ["mac", "ip", "state", "session_id", "auth_status", "user", "duration", "idle", "in", "out", "max", "swap_octets", "percent_up", "percent_down", "original_url"]
        return jqgrid_json(self, chilli_status, header, id_field_index=0, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)

    @require(has_write('Services'))
    def coova_chilli_authorize_user(self, sessionid, **kwargs):
        """Authorizes the given session (sessionid) with the provided settings (passed via kwargs)."""
        from coova_chilli import authorize_user
        result = authorize_user(sessionid, **kwargs)
        if result == 0:
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> User has been authenticated.'))
        else:
            return jquery_ui_error_msg(_('An error was encountered trying to run the chilli_query command.<br />'))

    @require(has_write('Services'))
    def coova_chilli_logout_user(self, mac, **kwargs):
        """Logs out the given MAC address (mac) using the chilli_query command."""
        from coova_chilli import logout_user
        result = logout_user(mac)
        if result == 0:
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> User has been authenticated.'))
        else:
            return jquery_ui_error_msg(_('An error was encountered trying to run the chilli_query command.<br />'))