/*
eftw.js - Escape From The Web!  Access a terminal anywhere.

Author:  Dan McDougall - <YouKnowWho@YouKnowWhat.com>

Note: This script was built using MochiKit (original version had some jQuery).
      It makes extensive use of Mochikit.Signal, MochiKit.Async, and MochiKit.DOM

For those wishing to port this to jQuery I've already thought about what would be necessary:
     * js-Hotkeys (http://code.google.com/p/js-hotkeys/).  NOTE: Changing the way keyboard interaction works will take a considerable effort.
     * "getElementDimensions(elem)" becomes: "$(elem).width(); $(elem).height();"
*/

// Globals
eftwTimer = null;
escapeTimer = null;
escapePressed = false;
charBuffer = new Array;
xhrDeferred = succeed(0);  // Global deferred for universal control of screen updates
doingUpdate = false; // Used to prevent out-of-order character events.
keyEvent = new Array; // This is meant to hold key events (2 at a time)

// Utility functions
function queue(character) {
    // Adds 'character' to the charBuffer Array
    charBuffer.unshift(character);
}
function getRowsAndColumns(elementID) {
/*  Calculates and returns the number of text rows and colunmns that will fit in the given element ID (elementID).
    Important:  elementID must be a basic block element such as DIV, SPAN, P, PRE, etc.
                Elements that require sub-elements such as TABLE (requires TRs and TDs) probably won't work.
    Note:  This function only works properly with monospaced fonts but it does work high-resolution displays.
           (so users with properly-configured high-DPI displays will be happy =).
           The other similar functions I found on the web had hard-coded pixel widths for known fonts
           at certain point sizes.  They break on any display higher than 96dpi.
*/

    // Get the full width and height of the given element:
    var elementDimensions = getElementDimensions(elementID);
    var elementNode = getElement(elementID);
    var fontSize = elementNode.style.fontSize;

    // Create a div to use to calculate sizing
    var sizingDiv = DIV();
    sizingDiv.innerHTML = "M"; //Fill it with a single character

    // Set the attributes of our copy to reflect a minimal-size block element
    minimalAttributes = {
        'display': 'block',
        'float': 'left',
        'position': 'absolute',
        'font-family': 'monospace',
        'font-size': fontSize,
        'top': '0px',
        'left': '0px',
        'width': 'auto',
        'height': 'auto'
    };
    setNodeAttribute(sizingDiv, 'style', minimalAttributes);

    // We have to append sizingDiv to the body to get the dimensions:
    appendChildNodes(window.document.body, sizingDiv);

    // getElementDimensions() should return the exact width and height of our single character (M):
    var textDimensions = getElementDimensions(sizingDiv);
    // Don't need this anymore:
    removeElement(sizingDiv);
    // Now calculate the rows and columns:
    var rows = Math.floor(elementDimensions.h / textDimensions.h);
    var cols = Math.floor(elementDimensions.w / textDimensions.w);
    dimensionsObj = {'rows': rows, 'cols': cols};
    return dimensionsObj
};
function bufferSpecial(character) {
    // Converts special characters (e.g. PgUp, PgDown, Arrow keys, etc) to strings before adding them to the charBuffer
    queue(String.fromCharCode(27) + character);
};
function updateTerminal(xhrResponse) {
    // Replaces the contents of the eftwTerminal div with any '<pre>' element inside the given xhrResponse.
    xml = xhrResponse.responseXML.documentElement;
    if (xml.tagName == "pre") {
        replaceChildNodes('eftwTerminal', xml);
    };
};
function eftwUpdate() {
    // Performs a one-time update of the EFTW terminal and pop()s out the current charBuffer.
    // Normally this function is called every second by pollServer() and every time a key is pressed.
    if (xhrDeferred.fired != -1){ // Only perform a screen update/character push if the last doSimpleXMLHttpRequest completed.
        // Note: This conditional is why we xhrDeferred is created using succeed().
        if (!doingUpdate) { // Only perform a screen update/character push if we're *positive* the last run is done.
            // I had to add this extra check because doSimpleXMLHttpRequest returns a non-blocking deferred and as a result
            // fast typers could wind up with characters lost.
//             log("charBuffer: " + charBuffer);
            doingUpdate = true;
            var charString = "";
            for (character in charBuffer) { charString += charBuffer.pop() };
            xhrDeferred = doSimpleXMLHttpRequest("/System/eftw", {chars: charString});
            xhrDeferred.addCallbacks(updateTerminal, eftwError);
            xhrDeferred.addCallback(function() {
                doingUpdate = false;
            });
        };
    };
};
function pollServer() {
    // Polls the server for changes every second and updates the terminal with those changes.
    // It uses a global Javascript interval timer, 'eftwTimer' which will stop EFTW from doing screen updates if cleared.
    eftwTimer = setInterval(function() {
        eftwUpdate();
    }, 1000);
};
function eftwInit(meta) {
    // This function initiates the Escape From The Web terminal on the server
    // (if there isn't already one associated with this user's session).
    // ...and starts the 1-second interval screen updates by running pollServer().
    // If an error is encountered it will kick off eftwError().
    xhrDeferred = doSimpleXMLHttpRequest("/System/eftw");
    xhrDeferred.addCallbacks(updateTerminal, eftwError);
    xhrDeferred.addCallback(pollServer());
};
function eftwError(meta) {
    // This gets called if we have trouble communicating with the server.
    // If there *is* a problem it will stop the 1-second screen refresh and retry connecting to the server every 5 seconds.
    getElement('eftwTerminal').innerHTML = "Error communicating with the server.  Will retry every 5 seconds...";
    clearInterval(eftwTimer);
    wait(5, function() {
        dimensions = getRowsAndColumns('eftwTerminal');
        var result = doSimpleXMLHttpRequest("/System/eftw_terminal_size", {rows: dimensions.rows, cols: dimensions.cols});
        result.addCallbacks(eftwInit, eftwError);
    });
};
function processKeyEvent() {
    // Expected to be run along with simultaneous onkeydown and onkeypress events...
    // This function determines whether or not a key can be sent to the server as-is or if it needs special handling.
    // It uses the keyEvent global to make sure that both types of events are received before processing
    // Note:  Because onkeydown always kicks off before onkeypress, keyEvent[0] will always be onkeydown and keyEvent[1] will always be onkeypress
//     log(repr(keyEvent));
    if (keyEvent.length != 2) {
        noop(); // Must have both an onkeydown and an onkeypress event
    } else {
        if (keyEvent[1].key().code) { // onkeypress actually registered a code
            var key = keyEvent[1].key()
            var modifiers = keyEvent[1].modifier();
            if (!modifiers.ctrl && !modifiers.alt && !modifiers.meta) { // Only send a key if no modifiers are held (other than shift)
                // Add the key to the buffer and update the terminal
//                 log("keyEvent[1] is a vanilla onkeypress event");
//                 log("Added '" + key.string + "' to charBuffer");
                queue(key.string);
                eftwUpdate();
                keyEvent = new Array; // Empty it out
            } else {
                // This is a ctrl/alt/meta key combo
//                 log("Invoking special key combo handler");
                eftwKeyEvents.emulateKeyCombo(keyEvent[1]);
                eftwUpdate();
                keyEvent = new Array; // Empty it out
            }
        } else { // This is an event that can only be handled by onkeydown
//             log("keyEvent[1] (onkeypress) didn't have a code. Handling " + keyEvent[0].key().string);
            eftwKeyEvents.emulateSpecialKey(keyEvent[0]);
            eftwUpdate();
            keyEvent = new Array; // Empty it out
        }
    }
};
eftwKeyEvents = {
    // This object holds all of the special key handlers and controls the "escape"/"escape escape" sequences
    emulateSpecialKey: function(e) {
        // This method handles all special keys that are only registered via onkeydown events (not onkeypress)
        var key = e.key();
        var modifiers = e.modifier();
        // We always want to intercept the escape key
        // Note: If any input element has focus the escape key will first un-focus that element (and apply any
        //       related functions) before this function will receive the escape key event.  So we don't have
        //       to worry about conflicts with functions that cancel things or close dialogs when the user
        //       presses the escape key.
        if (key.string == 'KEY_ESCAPE') {
            if (getElement('eftw').style.display != "none") { // If it is visible
                if (escapePressed) { // This is the *second* time escape was pressed within 1.25 seconds.
                    toggle('eftw', 'slide', { // Hide the Escape From The Web terminal.
                        beforeStart: function() {
                            clearInterval(eftwTimer); // Stop updating the screen every second.
                            clearTimeout(escapeTimer); // Kill the escape timer (probably not necessary but good coding practices)
                        },
                        afterFinish: function () {
                            escapePressed = false; // Reset the escapePressed state.
                            return true;
                        }
                    });
                } else { // This is a new escape key event.
                    // First set the escape key state:
                    escapePressed = true;
                    // Next, kick off a global timeout that:
                    //  1) Resets escapePressed if the user doesn't press it again in 1.25 seconds
                    //  2) Adds the escape key to the charBuffer to preserve the escape key's terminal functionality.
                    escapeTimer = setTimeout(function() {
                        escapePressed = false;
//                         log("adding Esc to buffer");
                        queue(String.fromCharCode(key.code));
                    }, 1250);
                    return true;
                };
            } else {
                toggle('eftw', 'slide', { // Slide down the EFTW terminal, figure out the text dimensions, and kick off the screen refreshing
                    afterFinish: function () {
                        dimensions = getRowsAndColumns('eftwTerminal');
                        var result = doSimpleXMLHttpRequest("/System/eftw_terminal_size", {rows: dimensions.rows, cols: dimensions.cols});
                        result.addCallbacks(eftwInit, eftwError);
                        return true;
                    }
                });
            };
        } else { // If *any* other key is pressed reset the escapePressed state and clear the escapeTimer
            escapePressed = false;
            // If this hasn't been set yet nothing will happen (no big deal)
            clearTimeout(escapeTimer);
            if (getElement('eftw')) {
                // Only send keypresses to the server if eftw is actually visible:
                if (getElement('eftw').style.display != "none") {
                    if (key.string == 'KEY_ENTER') queue(String.fromCharCode(13));
                    else if (key.string == 'KEY_TAB') queue(String.fromCharCode(9));
                    else if (key.code == 'KEY_BACKSPACE') queue(String.fromCharCode(127));
                    else {
                        if (key.string == 'KEY_PAGE_UP') bufferSpecial("[5~");
                        else if (key.string == 'KEY_PAGE_DOWN') bufferSpecial("[6~");
                        else if (key.string == 'KEY_END') bufferSpecial("[4~");
                        else if (key.string == 'KEY_HOME') bufferSpecial("[1~");
                        else if (key.string == 'KEY_ARROW_LEFT') bufferSpecial("[D");
                        else if (key.string == 'KEY_ARROW_UP') bufferSpecial("[A");
                        else if (key.string == 'KEY_ARROW_RIGHT') bufferSpecial("[C");
                        else if (key.string == 'KEY_ARROW_DOWN') bufferSpecial("[B");
                        else if (key.string == 'KEY_INSERT') bufferSpecial("[2~");
                        else if (key.string == 'KEY_DELETE') bufferSpecial("[3~");
                        else if (key.string == 'KEY_F1') bufferSpecial("[[A");
                        else if (key.string == 'KEY_F2') bufferSpecial("[[B");
                        else if (key.string == 'KEY_F3') bufferSpecial("[[C");
                        else if (key.string == 'KEY_F4') bufferSpecial("[[D");
                        else if (key.string == 'KEY_F5') bufferSpecial("[[E");
                        else if (key.string == 'KEY_F6') bufferSpecial("[17~");
                        else if (key.string == 'KEY_F7') bufferSpecial("[18~");
                        else if (key.string == 'KEY_F8') bufferSpecial("[19~");
                        else if (key.string == 'KEY_F9') bufferSpecial("[20~");
                        else if (key.string == 'KEY_F10') bufferSpecial("[21~");
                        else if (key.string == 'KEY_F11') bufferSpecial("[23~");
                        else if (key.string == 'KEY_F12') bufferSpecial("[24~");
                        else { // For everything else the default code converted into a string should work fine
    //                         log("key is being handled by onkeypress using String.fromCharCode(key.code)");
                            queue(String.fromCharCode(key.code));
                        };
                    }
                };
            }
        }
    },
    emulateKeyCombo: function(e) {
        // This method translates ctrl/alt/meta key combos such as ctrl-c into their string equivalents.
        var key = e.key();
        var modifiers = e.modifier();
        if (key.code >= 97 && key.code <= 122) queue(String.fromCharCode(key.code - 96)); // Ctrl-[a-z]
        else if (key.code >= 65 && key.code <= 90) queue(String.fromCharCode(key.code - 64)); // Ctrl-Shift-[a-z]
        else if (key.string == "^")  queue(String.fromCharCode(30));
            else if (key.string == "_") queue(String.fromCharCode(31));
            else if (key.string == "[") queue(String.fromCharCode(27));
            else if (key.code == 92) queue(String.fromCharCode(28)); // Ctrl-\
            else if (key.string == "]") queue(String.fromCharCode(29));
            else if (key.string == "[") queue(String.fromCharCode(27));
            else if (key.string == "@") queue(String.fromCharCode(0));
    }
};
// Note:  onkeydown always fires before onkeypress
connect(document, 'onkeydown', function(e) {
    // Special keys (enter, escape, control-<key>, alt-<key>, meta-<key>, F-keys, etc) must be converted before being sent to the server
    if (getElement('eftw')) { // This extra check prevents an exception when someone presses a key before the document has been fully loaded
        if (getElement('eftw').style.display != "none") {
            e.preventDefault();
        };
    };
    if (keyEvent.length == 0) {
        keyEvent.push(e);
        processKeyEvent();
    };
});
connect(document, 'onkeypress', function(e) {
    // For regular keys (single keypress with or without shift) the 'onkeypress' string works just fine
    if (getElement('eftw')) {
        if (keyEvent.length == 1) {
            keyEvent.push(e);
            processKeyEvent();
        }
    }
});
connect(window, 'onresize', function() {
    if (getElement('eftw')) {
        if (getElement('eftw').style.display != "none") {
            log('window resize detected');
            clearInterval(eftwTimer);
            getElement('eftwTerminal').innerHTML = ""; // clear it out so we don't end up with leftovers (that give us scroll bars).
            dimensions = getRowsAndColumns('eftwTerminal');
            log('dimensions, rows: ' + dimensions.rows + ' cols: ' + dimensions.cols);
            var result = doSimpleXMLHttpRequest("/System/eftw_terminal_size", {rows: dimensions.rows, cols: dimensions.cols});
            result.addCallbacks(eftwInit, eftwError);
        }
    }
});
addLoadEvent(function() {
// Create the EFTW container and append it to the document body
    var eftw = DIV({'id': 'eftw', 'style': 'display: none; z-index:140; width:100%; height:99%; position:absolute; top:0; left:0;'},
        DIV({'id': 'eftwTerminal', 'style': 'width:100%; height:100%; z-index:140; background-color:rgba(0,0,0,0.8); margin: 0px; display: block; font-family: monospace; font-size: 1.2em; white-space: pre; color: white;'})
    );
    appendChildNodes(document.body, eftw);
});