# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

from plugin_utils import get_translation
from configfile import ConfigFile

# Take care of internationalization
trans = get_translation(__file__)
_ = lambda s: trans.ugettext(s)

class CherryPyConfig(ConfigFile):
    """Class to handle manipulation of the cherrypy.conf file"""
    def __init__(self, filepath="cherrypy.conf"):
        self.filepath = filepath
        self.non_configurable_lines = []
        # Initialize the SSL values as None since we may not have them set
        self.ssl_certificate = None
        self.ssl_private_key = None
        self.read_conf()

    def read_conf(self):
        """Parses cherrypy.conf and stores the settings in local variables.
            self.port
            self.listen_address
            self.num_threads
            self.sessions_timeout
            self.ssl_certificate
            self.ssl_private_key
            self.log_to_screen
            self.access_log
            self.error_log"""
        f = open(self.filepath, 'r').readlines()
        for line in f:
            line = line.rstrip('\n') # Don't need newlines in our variables
            if line.startswith('['): # i.e. [global]
                self.non_configurable_lines.append(line) # Make sure this always goes first
            elif 'server.socket_host' in line:
                line = line.split('=')[1].strip() # 'server.socket_host = "0.0.0.0"' becomes '"0.0.0.0"'
                if "'" or '"' in line: # This option has quotes around it
                    line = line.strip('"').strip("'") # Get rid of any surrounding quotes
                self.listen_address = line
            elif 'server.socket_port' in line:
                line = line.split('=')[1].strip()
                line = line.strip('"').strip("'")
                self.port = line
            elif 'server.thread_pool' in line:
                line = line.split('=')[1].strip()
                line = line.strip('"').strip("'")
                self.num_threads = line
            elif 'tools.sessions.timeout' in line:
                line = line.split('=')[1].strip()
                line = line.strip('"').strip("'")
                self.session_timeout = line
            elif 'log.screen' in line:
                line = line.split('=')[1].strip()
                line = line.strip('"').strip("'")
                self.log_to_screen = line
            elif 'log.access_file' in line:
                line = line.split('=')[1].strip()
                line = line.strip('"').strip("'")
                self.access_log = line
            elif 'log.error_file' in line:
                line = line.split('=')[1].strip()
                line = line.strip('"').strip("'")
                self.error_log = line
            elif 'server.ssl_certificate' in line:
                line = line.split('=')[1].strip()
                if "'" or '"' in line: # This option has quotes around it
                    line = line.strip('"').strip("'") # Get rid of any surrounding quotes
                self.ssl_certificate = line
            elif 'server.ssl_private_key' in line:
                line = line.split('=')[1].strip()
                if "'" or '"' in line: # This option has quotes around it
                    line = line.strip('"').strip("'") # Get rid of any surrounding quotes
                self.ssl_private_key = line
            # Commented this out because I don't think CherryPy 3.1.2 supports this yet
            #elif 'server.ssl_certificate_chain' in line:
                #line = line.split('=')[1].strip()
                #if "'" or '"' in line: # This option has quotes around it
                    #line = line.strip('"').strip("'") # Get rid of any surrounding quotes
                #self.ssl_certificate_chain = line
            else:
                self.non_configurable_lines.append(line)

    def save(self):
        """Save the file"""
        out_string = self.non_configurable_lines[0] + '\n' # should be [global]
        out_string += 'server.socket_host = "' + self.listen_address + '"\n'
        out_string += 'server.socket_port = ' + str(self.port) + '\n'
        out_string += 'server.thread_pool = ' + self.num_threads + '\n'
        out_string += 'tools.sessions.timeout = ' + self.session_timeout + '\n'
        out_string += 'log.log_to_screen = ' + self.log_to_screen + '\n'
        out_string += 'log.access_file = ' + self.access_log + '\n'
        out_string += 'log.error_file = ' + self.error_log + '\n'
        out_string += 'server.ssl_certificate = "' + str(self.ssl_certificate) + '"\n'
        out_string += 'server.ssl_private_key = "' + str(self.ssl_private_key) + '"\n'
        for line in self.non_configurable_lines[1:]:
            out_string += line + '\n'
        f = open(self.filepath, 'w')
        f.write(out_string)
        f.close()

    def asList(self):
        """Returns our parameters as a list of lists."""
        config_list = []
        config_list.append([_('Listen Address'), self.listen_address])
        config_list.append([_('Port'), self.port])
        config_list.append([_('Number of Threads'), self.num_threads])
        config_list.append([_('Session Timeout'), self.session_timeout])
        config_list.append([_('Log to Screen'), self.log_to_screen])
        config_list.append([_('Access Log'), self.access_log])
        config_list.append([_('Error Log'), self.error_log])
        config_list.append([_('SSL Certificate'), self.ssl_certificate])
        config_list.append([_('SSL Private Key'), self.ssl_private_key])
        return config_list