#!/usr/bin/env python
# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

#
# Description: Network plugin

import datetime, os
try:
    from commands import getstatusoutput, getoutput
except: # getstatusoutput was moved to subprocess in Python 3
    from subprocess import getstatusoutput
try: # Note: The json module is only present in Python 2.6+
    import json
except: # If not found, try importing simplejson which will be required if using Python 2.4 or Python 2.5
    import simplejson as json # simplejson was included in Python 2.6 as 'json'
import cherrypy
from openwrt import ConfigFileOpenWRT, OpenWRTConfig
from plugin_utils import Plugin, get_translation
from auth import require, has_read, has_write
from utils import lookup_template, search_cells, jqgrid_format, jqgrid_json, plugin_template, jquery_ui_error_msg, jquery_ui_highlight_msg, sortby_field, split_seq
from iptables import IPTables
from odict import odict

# Internationalization
trans = get_translation(__file__)
_ = lambda s: trans.ugettext(s)

# Globals
# These are the targets built into the Linux kernel
iptables_targets = set(['ACCEPT', 'CLASSIFY', 'CLUSTERIP', 'CONNMARK', 'CONNSECMARK', 'DNAT',
    'DROP', 'DSCP', 'ECN', 'IPV4OPTSSTRIP', 'LOG', 'MARK', 'MASQUERADE',
    'MIRROR', 'NETMAP', 'NFQUEUE', 'NOTRACK', 'REDIRECT', 'REJECT', 'RETURN',
    'ROUTE', 'SAME', 'SECMARK', 'SET', 'SNAT', 'TARPIT', 'TCPMSS', 'TOS',
    'TRACE', 'TTL', 'ULOG'
    ]) # Not using them yet but I plan to

class dnsmasqException(Exception): pass

def delete_dhcp_lease(lease_id):
    """Deletes the given lease ID from dnsmasq's DHCP lease list."""
    # Note: dnsmasq only uses it's lease file when it starts up.
    #       ...so to truly delete a lease we have to do this:
    #
    #       1) Stop dnsmasq
    #       2) Remove the specified lease
    #       3) Start dnsmasq again
    new_lease_file = ""
    dhcp_conf = ConfigFileOpenWRT(filepath='/etc/config/dhcp')
    # get the dnsmasq config section
    dnsmasq_config = [a for a in dhcp_conf.sections if 'dnsmasq' in a['config']][0]
    # ...so we can get the lease file path
    lease_filepath = dnsmasq_config['options']['leasefile']
    # Don't need these anymore
    del dnsmasq_config
    del dhcp_conf
    # First we make a copy of the existing lease file minus the entry we're removing
    for line in open(lease_filepath, 'r'):
        if line: # skip blank lines
            if lease_id in line:
                pass # skip it since we're deleting it
            else:
                new_lease_file += line
    # Next we stop dnsmasq...
    return_code, output = getstatusoutput('/etc/init.d/dnsmasq stop')
    if return_code != 0:
        raise dnsmasqException
    # Now we write our new lease file...
    lease_file = open(lease_filepath, 'w')
    lease_file.write(new_lease_file)
    lease_file.close()
    # Finally we start dnsmasq back up...
    return_code, output = getstatusoutput('/etc/init.d/dnsmasq start')
    if return_code == 0:
        return True # Success
    else:
        raise dnsmasqException

# TODO:  Enable apply/undo functions in the interface
# TODO:  Come up with a good way to apply settings without rebooting
class Network(Plugin):
    """A plugin for OpenWRT's network configuration"""
    display = True
    sort_priority = 1
    # Using an ordered dictionary to control the order:
    navigable_functions = odict([
        (_('DHCP Server'), 'DHCPServer'),
        (_('Dynamic DNS'), 'DDNS'),
        (_('Firewall'), 'Firewall'),
        (_('Interfaces'), 'Interfaces'),
        (_('QOS (Traffic Shaping)'), 'QOS'),
        (_('Static Routes'), 'StaticRoutes'),
        (_('Switch Configuration'), 'Switch'),
        (_('Wireless'), 'Wireless'),
        ])
    wireless_device_options = [
        'type', 'channel', 'country', 'disabled', 'distance', 'diversity', 'frameburst', 'hwmode', 'macfilter', 'maclist', 'maxassoc', 'rxantenna', 'txantenna', 'txpower'
    ]
    wireless_interface_options = [
        "mode", "ar", "auth", "bgscan", "bssid", "bursting", "ca_cert", "compression", "device", "doth", "eap_type", "encryption", "ff", "frag", "hidden", "identity", "isolate", "key", "maclist", "macpolicy", "maxrate", "mcast_rate", "minrate", "nasid", "network", "nosbeacon", "password", "port", "priv_key", "priv_key_pwd", "probereq", "rate", "rts", "server", "ssid", "sw_merge", "turbo", "txpower", "wds", "wdssep", "wmm", "xr"
    ]
    dnsmasq_options = [
        "addnhosts", "authoritative", "boguspriv", "dhcp_boot", "dhcpleasemax", "dnsforwardmax", "domain", "domainneeded", "ednspacket_max", "enable_tftp", "expandhosts", "filterwin2k", "leasefile", "local", "localise_queries", "logqueries", "nohosts", "nonegcache", "noresolv", "port", "queryport", "readethers", "resolvfile", "strictorder", "tftp_root"
    ]
    dhcp_interface_options = [
        "interface", "dhcp_option", "dynamicdhcp", "force", "ignore", "leasetime", "limit", "name", "netmask", "start"
    ]

    @require(has_read('Network'))
    def Interfaces(self):
        """Render the network configuration page"""
        interfaces_template = plugin_template(self, "templates/interfaces.mako")
        network_conf = ConfigFileOpenWRT(filepath='/etc/config/network')

        return self.standard_render(
            template=interfaces_template,
            help=_('This is the help for the LAN method'),
            long_name=_('LAN Configuration'),
            description=_("Configure your router's LAN below."),
            network_conf=network_conf,
        )

    @require(has_read('Network'))
    def interfaces_tab(self):
        """Render the interfaces tab"""
        _template = plugin_template(self, "templates/interfaces_tab.mako")

        # We need to get the name of the bridge (if any)
        retcode, output = getstatusoutput('brctl show')
        if retcode != 0:
            bridges = {} # No bridge to worry about
        else:
            bridges = {}
            for line in output.split('\n'):
                if line.startswith('bridge name'):
                    pass # This is the header line
                else: # Append the bridge and devices to the list
                    bridge_name = line.split()[0]
                    bridge_interfaces = line.split()[3:]
                    bridges.update({bridge_name: bridge_interfaces})

        # Grab the current status of all interfaces
        from iproute2 import IP
        ip_obj = IP()
        interface_stats = ip_obj.interfaces

        # Get the help text for each option
        from network_help import interface_options

        network_conf = ConfigFileOpenWRT(filepath='/etc/config/network')
        validation_obj = network_conf.validation_obj()
        validation_dict = network_conf.validation_dict(validation_obj)

        return self.standard_render(
            template=_template,
            network_conf=network_conf,
            validation_dict=validation_dict,
            bridges=bridges,
            interface_stats=interface_stats,
            option_help=interface_options,
        )

    @require(has_read('Network'))
    def routes_tab(self):
        """Render the routes tab"""
        _template = plugin_template(self, "templates/routes_tab.mako")

        return self.standard_render(
            template=_template,
            network_conf=ConfigFileOpenWRT(filepath='/etc/config/network'),
        )

    @require(has_read('Network'))
    def network_raw_tab(self):
        """Render the raw network edit tab"""
        _template = plugin_template(self, "templates/network_raw_tab.mako")
        network_conf = ConfigFileOpenWRT(filepath='/etc/config/network')
        raw_network = network_conf.raw()
        del network_conf # Clean up memory

        return self.standard_render(
            template=_template,
            raw_network=raw_network
        )

    @require(has_read('Network'))
    def new_interface_tab(self):
        """Render the new interface tab"""
        _template = plugin_template(self, "templates/new_interface_tab.mako")

        # Get the help text for each option
        from network_help import interface_options

        network_conf = ConfigFileOpenWRT(filepath='/etc/config/network')
        validation_obj = network_conf.validation_obj()
        validation_dict = network_conf.validation_dict(validation_obj)

        return self.standard_render(
            template=_template,
            network_conf=network_conf,
            validation_dict=validation_dict,
            option_help=interface_options,
        )

    @require(has_read('Network'))
    def interface_options_meta(self, option=None, section=None, help=True, values=False, **data):
        """By default, return help text for the provided interface option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section.
        If 'section' is provided, attempt to obtain valid values from the specified section
        of the validation file."""
        from network_help import interface_options
        network = ConfigFileOpenWRT('/etc/config/network')
        if values:
            try:
                values_dict = {}
                for value in network.valid_options(section, option):
                    if value == '0' or value == 'off' or value == 'disabled':
                        values_dict.update({value:"disabled"})
                    elif value == '1' or value == 'on' or value == 'enabled':
                        values_dict.update({value:"enabled"})
                    else:
                        values_dict.update({value:value})
                if values_dict:
                    return json.dumps(values_dict)
                else:
                    return "{empty:true}"
            except:
                # This is so we can determine that no list values are available for this option
                # ...in JavaScript: if (!json.empty) { do stuff... }
                return "{empty:true}"

        if help:
            try:
                return '<strong>%s</strong>: %s' % (option, interface_options[option]['help'])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_write('Network'))
    def update_raw_network(self, network):
        """Save the raw network config file text as submitted"""
        try:
            network_conf = ConfigFileOpenWRT(filepath='/etc/config/network')
            network_conf.save_raw(network)
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/config/network was saved.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not save /etc/config/network. <br />%s') % e)

    @require(has_write('Network'))
    def update_interfaces(self, config=None, **kwargs):
        """Updates interfaces in /etc/config/network with the supplied options.
        Requires 'config' (i.e. 'interface lan')"""
        if not config:
            return jquery_ui_error_msg(_('Could not update network settings.  Please make sure to pass this function a "config" parameter.'))
        try:
            conf_position = None
            network_conf = ConfigFileOpenWRT('/etc/config/network')
            for index, section in enumerate(network_conf.sections):
                if "_".join(section['config']) == config:
                    conf_position = index
            if 'delete' in kwargs.keys(): # This interface has been marked for deletion
                del network_conf.sections[conf_position] # Remove this config section
                network_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> Interface has been deleted.'))
            network_conf.sections[conf_position]['options'] = {} # We'll be replacing them all
            for key, value in kwargs.items():
                #print "%s\t%s" % (key, value) # For debugging
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if key == "rename":
                        if 'interface_' in config and 'interface' not in value:
                            value = ['interface', value] # Interfaces need to use the following format: 'interface <ifname>'
                    elif key == "alias_rename":
                        if 'alias_' in config and 'alias' not in value:
                            value = ['alias', value] # Aliases need to use the following format: 'alias <aliasname>'
                        network_conf.sections[conf_position]['config'] = value
                    else:
                        network_conf.sections[conf_position]['options'].update({key: value})
            network_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Network settings have been updated.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not update network settings.<br />%s') % e)

    @require(has_write('Network'))
    def add_interface(self, config=None, alias_config=None, **kwargs):
        """Creates a new interface or alias configuration section to
        /etc/config/network depending on whether or not config or alias_config
        was passed."""
        if not config and not alias_config:
            return jquery_ui_error_msg(_('Could not add new network settings.  Please choose a name for your interface.'))
        if config:
            if 'interface' not in config:
                config = 'interface ' + config
        if alias_config:
            if 'alias' not in alias_config:
                config = 'alias ' + alias_config
        network_conf = ConfigFileOpenWRT('/etc/config/network')
        # First make sure there isn't already an interface with this name
        for section in network_conf.sections:
            if " ".join(section['config']) == config:
                return jquery_ui_error_msg(_('An interface already exists with this name.  Please choose a different name and re-submit.'))
        try: # Now try creating the new interface
            options = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if value:
                        options.update({key: value})
            config_obj = OpenWRTConfig(config=config.split(), options=options)
            network_conf.sections.append(config_obj)
            network_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Network settings have been updated.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not update network settings.<br />%s') % e)

    @require(has_read('Network'))
    def interface_status_tip(self, interface, **data):
        """Return a nice little jTip popup showing the status of the given interface"""
        from brctl import brctl
        bridge = {}
        bridge_interfaces = []
        bridge_obj = brctl()
        for key in bridge_obj.bridges.keys():
            if key == interface:
                bridge = bridge_obj.bridges[key]

        # Grab the current status of all interfaces
        from iproute2 import IP
        ip_obj = IP()

        try:
            interface_stats = ip_obj.interfaces[interface]
            _template = plugin_template(self, "templates/interface_status_tip.mako")

            return self.standard_render(
                template=_template,
                interface=interface,
                interface_stats=interface_stats,
                bridge=bridge,
            )
        except Exception, e:
            print e
            return  _('<strong>%s</strong>: No status information was found for this interface.') % interface

    @require(has_read('Network'))
    def Switch(self):
        """Render the switch configuration page"""
        _template = plugin_template(self, "templates/switch.mako")
        network_conf = ConfigFileOpenWRT(filepath='/etc/config/network')

        return self.standard_render(
            template=_template,
            help=_('<p>This interface allows you to configure the switch attached to this system (if any).</p>'),
            long_name=_('Network: Switch Configuration'),
            description=_("Configure the switch ports on this system to be on separate VLANs with or without tagging."),
            network_conf=network_conf,
        )

    @require(has_read('Network'))
    def switch_conf(self):
        """Render the (guided) switch configuration tab"""
        _template = plugin_template(self, "templates/switch_conf.mako")
        network_conf = ConfigFileOpenWRT('/etc/config/network')
        validation_obj = network_conf.validation_obj()
        validation_dict = network_conf.validation_dict(validation_obj)
        return self.standard_render(
            template=_template,
            network_conf=network_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def switch_options_meta(self, option=None, section=None, help=True, values=False, **data):
        """Return help text for the provided switch option (for jTip)."""
        network_conf = ConfigFileOpenWRT('/etc/config/network')
        if values:
            return '{empty: true}'
        if help:
            try:
                return """<strong>%s</strong>: VLANs are configured as a list of ports with the following optional prefixes:
                <ul><li>&#8217;*&#8217;: Set the default VLAN (PVID) of the Port to the current VLAN</li>
                <li>&#8217;u&#8217;: Force the port to be untagged</li>
                <li>&#8217;t&#8217;: Force the port to be tagged</li></ul>""" % option
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_write('Network'))
    def update_switch(self, config=None, **kwargs):
        """Updates a switch configuration in /etc/config/network with the supplied options.
        'config' is expected to be the interface which these settings will be applied to.
        """
        try:
            conf_position = None
            network_conf = ConfigFileOpenWRT('/etc/config/network')
            for index, section in enumerate(network_conf.sections):
                if 'switch' in section['config'] and section['config'][1] == config:
                    conf_position = index
            if 'delete' in kwargs.keys(): # This interface has been marked for deletion
                del network_conf.sections[conf_position] # Remove this config section
                network_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> Switch %s has been deleted.') % config)
            network_conf.sections[conf_position]['options'] = {} # We'll be replacing them all
            for key, value in kwargs.items():
                print "%s\t%s" % (key, value) # For debugging
                # Update target and interface options first since we've singled them out
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    network_conf.sections[conf_position]['options'].update({key: value})
            network_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Switch settings have been updated.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not update network settings.<br />%s') % e)

    @require(has_write('Network'))
    def add_switch(self, config=None, **kwargs):
        """Adds a new switch configuration section to /etc/config/network.
        'config' is expected to be the interface these settings apply to."""
        network_conf = ConfigFileOpenWRT('/etc/config/network')
        # First make sure there isn't already an switch for this interface
        for section in network_conf.sections:
            if 'switch' in section['config'] and section['config'][1] == config:
                return jquery_ui_error_msg(_('A switch configuration for %s already exists.') % config)
        try: # Now try creating the new switch configuration
            options = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if value:
                        options.update({key: value})
            config_obj = OpenWRTConfig(config=['switch', config], options=options)
            network_conf.sections.append(config_obj)
            network_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Switch configuration was added.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not add switch configuration.<br />%s') % e)

    @require(has_read('Network'))
    def StaticRoutes(self):
        """Render the static route configuration page"""
        _template = plugin_template(self, "templates/static_routes.mako")
        network_conf = ConfigFileOpenWRT(filepath='/etc/config/network')

        return self.standard_render(
            template=_template,
            help=_('<p>This interface allows you to configure IPv4 and IPv6 static routes on this system.</p>'),
            long_name=_('Network: Static Routes'),
            description=_("Configure static routes."),
            network_conf=network_conf,
        )

    @require(has_read('Network'))
    def ipv4_routes(self):
        """Render the IPv4 static routes configuration page"""
        _template = plugin_template(self, "templates/ipv4_routes.mako")
        network_conf = ConfigFileOpenWRT('/etc/config/network')
        validation_obj = network_conf.validation_obj()
        validation_dict = network_conf.validation_dict(validation_obj)
        return self.standard_render(
            template=_template,
            network_conf=network_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def route_options_meta(self, option=None, section=None, help=True, values=False, **data):
        """By default, return help text for the provided route option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section.
        If 'section' is provided, attempt to obtain valid values from the specified section
        of the validation file."""
        network_conf = ConfigFileOpenWRT('/etc/config/network')
        validation_obj = network_conf.validation_obj()
        validation_dict = network_conf.validation_dict(validation_obj)
        if values:
            try:
                values_dict = {}
                for value in network_conf.valid_options(section, option):
                    if value == '0' or value == 'off' or value == 'disabled':
                        values_dict.update({value:"disabled"})
                    elif value == '1' or value == 'on' or value == 'enabled':
                        values_dict.update({value:"enabled"})
                    else:
                        values_dict.update({value:value})
                if values_dict:
                    return json.dumps(values_dict)
                else:
                    return "{empty:true}"
            except:
                # This is so we can determine that no list values are available for this option
                # ...in JavaScript: if (!json.empty) { do stuff... }
                return "{empty:true}"
        if help:
            try:
                return '<strong>%s</strong>: %s' % (option, validation_dict[section][option]['title'])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_read('Network'))
    def validate_option(self, config=None, section=None, **kwargs):
        """Returns a json-encoded error message if the given option:value pair does not validate"""
        config_obj = ConfigFileOpenWRT('/etc/config/%s' % config)
        result = None
        # This should only ever loop once
        for key, value in kwargs.items():
            result = config_obj.validate_option(section, key, value)
        if result:
            return json.dumps(result)

    @require(has_read('Network'))
    def ipv6_routes(self):
        """Render the IPv6 static routes configuration page"""
        _template = plugin_template(self, "templates/ipv6_routes.mako")
        network_conf = ConfigFileOpenWRT('/etc/config/network')
        validation_obj = network_conf.validation_obj()
        validation_dict = network_conf.validation_dict(validation_obj)
        return self.standard_render(
            template=_template,
            network_conf=network_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def route6_options_meta(self, option=None, section=None, help=True, values=False, **data):
        """By default, return help text for the provided route6 option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section.
        If 'section' is provided, attempt to obtain valid values from the specified section
        of the validation file."""
        network_conf = ConfigFileOpenWRT('/etc/config/network')
        validation_obj = network_conf.validation_obj()
        validation_dict = network_conf.validation_dict(validation_obj)
        if values:
            try:
                values_dict = {}
                for value in network_conf.valid_options(section, option):
                    if value == '0' or value == 'off' or value == 'disabled':
                        values_dict.update({value:"disabled"})
                    elif value == '1' or value == 'on' or value == 'enabled':
                        values_dict.update({value:"enabled"})
                    else:
                        values_dict.update({value:value})
                if values_dict:
                    return json.dumps(values_dict)
                else:
                    return "{empty:true}"
            except:
                # This is so we can determine that no list values are available for this option
                # ...in JavaScript: if (!json.empty) { do stuff... }
                return "{empty:true}"

        if help:
            try:
                return '<strong>%s</strong>: %s' % (option, validation_dict[section][option]['title'])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_write('Network'))
    def update_route(self, config=None, **kwargs):
        """Updates a route in /etc/config/network with the supplied options.
        Requires that 'config' be formatted like so: '<interface>_<target>'"""
        # Convert config (i.e. 'lan_10.1.1.1') into separate variables
        interface = config.split('_')[0]
        target = config.split('_')[1]
        try:
            conf_position = None
            network_conf = ConfigFileOpenWRT('/etc/config/network')
            for index, section in enumerate(network_conf.sections):
                if 'target' in section['options'] and 'interface' in section['options']:
                    if section['options']['target'] == target and section['options']['interface'] == interface:
                        conf_position = index
            if 'delete' in kwargs.keys(): # This interface has been marked for deletion
                del network_conf.sections[conf_position] # Remove this config section
                network_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> Route has been deleted.'))
            network_conf.sections[conf_position]['options'] = {} # We'll be replacing them all
            for key, value in kwargs.items():
                #print "%s\t%s" % (key, value) # For debugging
                # Update target and interface options first since we've singled them out
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    network_conf.sections[conf_position]['options'].update({key: value})
            network_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Static Route settings have been updated.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not update network settings.<br />%s') % e)

    @require(has_write('Network'))
    def add_route(self, config=None, **kwargs):
        """Adds a new IPv4 or IPv6 route configuration section to
        /etc/config/network depending on whether or not config=route or config=route6."""
        network_conf = ConfigFileOpenWRT('/etc/config/network')
        # First make sure there isn't already an interface with this name
        for section in network_conf.sections:
            if section['config'] == config:
                for key, value in kwargs.items():
                    if key == 'interface' and value == section['options']['interface']:
                        if key == 'target' and value == section['options']['target']:
                            return jquery_ui_error_msg(_('A route already exists with this configuration.'))
        try: # Now try creating the new route
            options = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if value:
                        options.update({key: value})
            config_obj = OpenWRTConfig(config=config.split(), options=options)
            network_conf.sections.append(config_obj)
            network_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Route was added.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not add route.<br />%s') % e)

    @require(has_read('Network'))
    def Wireless(self):
        """Render the wireless configuration page"""
        wireless_template = plugin_template(self, "templates/wireless.mako")
        wireless_conf = ConfigFileOpenWRT(filepath='/etc/config/wireless')

        return self.standard_render(
            template=wireless_template,
            help=_('This is the help for the wireless configuration section.'),
            long_name=_('Wireless Configuration'),
            description=_("Configure your router's wireless settings below."),
            wireless_conf=wireless_conf,
        )

    @require(has_read('Network'))
    def wireless_options_meta(self, option=None, section=None, help=True, values=False, **data):
        """By default, return help text for the provided wireless option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section.
        If 'section' is provided, attempt to obtain valid values from the specified section
        of the validation file.
        Note:  This method has many special exceptions for various wirless settings."""
        from network_help import wifi_device_options
        wireless_conf = ConfigFileOpenWRT('/etc/config/wireless')
        validation_obj = wireless_conf.validation_obj()
        validation_dict = wireless_conf.validation_dict(validation_obj)

        if values:
            try:
                values_dict = odict()
                for value in wireless_conf.valid_options(section, option):
                    if value == '0' or value == 'off' or value == 'disabled':
                        if option == 'disabled': # disabled: disabled is confusing so I change it to True/False for this option
                            values_dict.update({value:"False - Enable this wireless device"})
                        elif option in ['channel', 'txantenna', 'rxantenna']:
                            values_dict.update({value: "%s - %s" % (value, validation_dict[section][option]['enum'][value]['title'])})
                        else:
                            values_dict.update({value:"disabled"})
                    elif value == '1' or value == 'on' or value == 'enabled':
                        if option == 'disabled':
                            values_dict.update({value:"True - Disable this wireless device"})
                        elif option in ['channel', 'txantenna', 'rxantenna']:
                            values_dict.update({value: "%s - %s" % (value, validation_dict[section][option]['enum'][value]['title'])})
                        else:
                            values_dict.update({value:"enabled"})
                    else:
                        if option in ['channel','mode', 'txantenna', 'rxantenna']:
                            if value == '': # empty (e.g. 'auto' mode is empty)
                                values_dict.update({value: validation_dict[section][option]['enum'][value]['title']})
                            else:
                                values_dict.update({value: "%s - %s" % (value, validation_dict[section][option]['enum'][value]['title'])})
                        else:
                            values_dict.update({value:value})
                if values_dict:
                    return json.dumps(values_dict)
                else:
                    return "{empty:true}"
            except Exception as e:
                # This is so we can determine that no list values are available for this option
                # ...in JavaScript: if (!json.empty) { do stuff... }
                #print e
                return "{empty:true}"

        if help:
            try:
                if section == 'wifi-device':
                    return '<strong>%s</strong>: %s' % (option, wifi_device_options[option]['help'])
                else:
                    #return '<strong>%s</strong>: %s' % (option, wifi_iface_options[option]['help'])
                    return '<strong>%s</strong>: %s' % (option, validation_dict[section][option]['title'])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_write('Network'))
    def update_raw_wireless(self, wireless):
        """Save the raw wireless config file text as submitted"""
        try:
            wireless_conf = ConfigFileOpenWRT(filepath='/etc/config/wireless')
            wireless_conf.save_raw(wireless)
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/config/wireless was saved.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not save /etc/config/wireless. <br />%s') % e)

    @require(has_read('Network'))
    def wireless_conf_tab(self):
        """Render the wireless devices/interfaces configuration form"""
        from network_help import wifi_device_options
        wireless_template = plugin_template(self, "templates/wireless_conf_tab.mako")
        wireless_conf = ConfigFileOpenWRT(filepath='/etc/config/wireless')
        validation_obj = wireless_conf.validation_obj()
        validation_dict = wireless_conf.validation_dict(validation_obj)

        return self.standard_render(
            template=wireless_template,
            wireless_conf=wireless_conf,
            wireless_device_options=self.wireless_device_options,
            wireless_interface_options=self.wireless_interface_options,
            validation_dict=validation_dict,
            wifi_device_options=wifi_device_options,
        )

    @require(has_read('Network'))
    def wireless_raw(self):
        """Render the raw /etc/conf/wireless edit form"""
        wireless_template = plugin_template(self, "templates/wireless_raw.mako")
        wireless_conf = ConfigFileOpenWRT(filepath='/etc/config/wireless')

        return self.standard_render(
            template=wireless_template,
            wireless_conf=wireless_conf,
            raw_wireless=wireless_conf.raw(),
        )

    @require(has_read('Network'))
    def new_wireless_device(self):
        """Render the new wireless device form"""
        wireless_template = plugin_template(self, "templates/wireless_new_device.mako")
        wireless_conf = ConfigFileOpenWRT(filepath='/etc/config/wireless')

        return self.standard_render(
            template=wireless_template,
            wireless_conf=wireless_conf,
            wireless_device_options=self.wireless_device_options,
        )

    @require(has_read('Network'))
    def wireless_channel_check(self, device):
        """Special method that determines which channels are valid for a given wireless 'device'.
        Returns a list of valid channels for the given device in json format.
        device - A string representing the wireless card name (e.g. 'wlan0')."""
        channel_list = ['auto']
        for line in getoutput('iwlist %s channel' % device).splitlines():
            if line.startswith(' '):
                if "Channel 00" in line:
                    pass
                elif "Current Frequency" in line:
                    pass
                else: # Convert channel to int to get rid of 0 in things like '06' then convert back to string
                    channel_list.append(str(int(line.split()[1])))
        return json.dumps(channel_list)

    @require(has_write('Network'))
    def update_wireless(self, config=None, **kwargs):
        """Updates /etc/config/wireless with the supplied options.  Requires 'config' which
        should be one of the following depending on if this is a wifi-device or wifi-iface:
            'wifi-device wlan0'
            'wifi-iface'
        The method figures out the rest from the submitted kwargs.
        """
        if not config:
            return jquery_ui_error_msg(_('Could not update wireless settings.  Please make sure to pass this function a "config" parameter.'))
        config = config.rstrip() # Get rid of trailing whitespace
        try:
            conf_position = None
            wireless_conf = ConfigFileOpenWRT('/etc/config/wireless')
            for index, section in enumerate(wireless_conf.sections):
                if config == 'wifi-iface': # There can be multiple idential 'config wifi-iface' headers in the config file
                    if 'device' not in kwargs.keys(): # A little extra validation
                        return jquery_ui_error_msg(_('You must supply a "device" option when saving a wifi-iface.'))
                    if " ".join(section['config']) == config:
                        if wireless_conf.sections[index]['options']['device'] == kwargs['device']:
                            conf_position = index
                else:
                    if " ".join(section['config']) == config:
                        conf_position = index
            if 'delete' in kwargs.keys(): # This has been marked for deletion
                del wireless_conf.sections[conf_position] # Remove this config section
                wireless_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> Device has been deleted.'))
            wireless_conf.sections[conf_position]['options'] = {} # We'll be replacing them all
            for key, value in kwargs.items():
                if key != 'options' and key != 'wifi_iface_options': # This gets submitted along with the form due to the way my dynamic form works.
                    if key == "rename":
                        if 'wifi-device' not in value:
                            value = ['wifi-device', value] # Interfaces need to use the following format: 'interface <ifname>'
                        wireless_conf.sections[conf_position]['config'] = value
                    else:
                        wireless_conf.sections[conf_position]['options'].update({key: value})
            wireless_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Wireless settings have been updated.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not update wireless settings.<br />%s') % e)

    @require(has_write('Network'))
    def add_wireless_device(self, config=None, **kwargs):
        """Creates new wireless device configuration in /etc/config/wireless"""
        if not config:
            return jquery_ui_error_msg(_('Could not add new wireless settings.  Please make sure to pass this function a "config" parameter.'))
        if 'wifi-device' not in config: # Set the config to "wifi-device <device name>"
            config = 'wifi-device ' + config
        wireless_conf = ConfigFileOpenWRT('/etc/config/wireless')
        # First make sure there isn't already an interface with this name
        for section in wireless_conf.sections:
            if " ".join(section['config']) == config:
                return jquery_ui_error_msg(_('A wireless device already exists with this name.  Please choose a different name and re-submit.'))
        try: # Now try creating the new interface
            options = {}
            for key, value in kwargs.items():
                if value:
                    options.update({key: value})
            config_obj = OpenWRTConfig(config=config.split(), options=options)
            wireless_conf.sections.append(config_obj)
            wireless_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Wireless settings have been updated.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not update wireless settings.<br />%s') % e)

    @require(has_write('Network'))
    def add_wireless_interface(self, config=None, **kwargs):
        """Creates new wireless device configuration in /etc/config/wireless"""
        if not config:
            return jquery_ui_error_msg(_('Could not add new wireless settings.  Please make sure to pass this function a "config" parameter.'))
        wireless_conf = ConfigFileOpenWRT('/etc/config/wireless')
        # First make sure there isn't already an interface with this name
        for section in wireless_conf.sections:
            if " ".join(section['config']) == config:
                if section['options']['device'] == kwargs['device']:
                    return jquery_ui_error_msg(_('A wireless interface already exists with this name.  Please choose a different name and re-submit.'))
        try: # Now try creating the new interface
            options = {}
            for key, value in kwargs.items():
                if value:
                    options.update({key: value})
            config_obj = OpenWRTConfig(config=config.split(), options=options)
            wireless_conf.sections.append(config_obj)
            wireless_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Wireless settings have been updated.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not update wireless settings.<br />%s') % e)

# TODO:  Add a "reload firewall rules" button or something that applies /etc/firewall.user
    @require(has_read('Network'))
    def Firewall(self):
        """Render the wireless configuration page"""
        firewall_template = plugin_template(self, "templates/firewall.mako")
        # TODO:  Add a secondary grid below each table grid that shows details on the 'match' portion of each firewall rule (that allows inline editing)

        return self.standard_render(
            template=firewall_template,
            help=_('This is the help for the firewall configuration section.'),
            long_name=_('Firewall Configuration'),
            description=_("Configure your router's firewall below."),
        )

    @require(has_read('Network'))
    def firewall_options_meta(self, option=None, section=None, help=True, values=False, **data):
        """By default, return help text for the provided firewall option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section.
        If 'section' is provided, attempt to obtain valid values from the specified section
        of the validation file."""
        firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
        validation_obj = firewall_conf.validation_obj()
        validation_dict = firewall_conf.validation_dict(validation_obj)
        if values:
            try:
                values_dict = {}
                for value in firewall_conf.valid_options(section, option):
                    if value == '0' or value == 'off' or value == 'disabled':
                        values_dict.update({value:"disabled"})
                    elif value == '1' or value == 'on' or value == 'enabled':
                        values_dict.update({value:"enabled"})
                    else:
                        values_dict.update({value:value})
                if values_dict:
                    return json.dumps(values_dict)
                else:
                    return "{empty:true}"
            except:
                # This is so we can determine that no list values are available for this option
                # ...in JavaScript: if (!json.empty) { do stuff... }
                return "{empty:true}"
        if help:
            try:
                return '<strong>%s</strong>: %s' % (option, validation_dict[section][option]['title'])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_read('Network'))
    def firewall_etc(self):
        """Render the raw /etc/config/firewall edit page."""
        _template = plugin_template(self, "templates/firewall_etc.mako")
        firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')

        return self.standard_render(
            template=_template,
            raw_firewall=firewall_conf.raw(),
        )

    @require(has_write('Network'))
    def update_raw_firewall(self, firewall):
        """Save the raw firewall config file text as submitted"""
        try:
            firewall_conf = ConfigFileOpenWRT(filepath='/etc/config/firewall')
            firewall_conf.save_raw(firewall)
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/config/firewall was saved.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not save /etc/config/firewall. <br />%s') % e)

    @require(has_read('Network'))
    def firewall_user_tab(self):
        """Render the raw firewall.user edit page."""
        _template = plugin_template(self, "templates/firewall_user_tab.mako")
        iptables = IPTables()

        return self.standard_render(
            template=_template,
            firewall_user_raw=iptables.raw(),
        )

    @require(has_read('Network'))
    def firewall_defaults_config(self):
        """Render the firewall defaults configuration page"""
        _template = plugin_template(self, "templates/firewall_defaults.mako")
        firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
        validation_obj = firewall_conf.validation_obj()
        validation_dict = firewall_conf.validation_dict(validation_obj)
        iptables = IPTables()

        return self.standard_render(
            template=_template,
            iptables=iptables,
            firewall_conf=firewall_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def rules_config(self):
        """Render the firewall rules configuration page"""
        _template = plugin_template(self, "templates/rules_config.mako")
        firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
        validation_obj = firewall_conf.validation_obj()
        validation_dict = firewall_conf.validation_dict(validation_obj)
        iptables = IPTables()

        return self.standard_render(
            template=_template,
            iptables=iptables,
            firewall_conf=firewall_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def forwarding_config(self):
        """Render the firewall forwarding configuration page"""
        _template = plugin_template(self, "templates/forwarding_config.mako")
        firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
        validation_obj = firewall_conf.validation_obj()
        validation_dict = firewall_conf.validation_dict(validation_obj)
        iptables = IPTables()

        return self.standard_render(
            template=_template,
            iptables=iptables,
            firewall_conf=firewall_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def port_forwarding_config(self):
        """Render the firewall port forwarding configuration page"""
        _template = plugin_template(self, "templates/port_forwarding_config.mako")
        firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
        validation_obj = firewall_conf.validation_obj()
        validation_dict = firewall_conf.validation_dict(validation_obj)
        iptables = IPTables()

        return self.standard_render(
            template=_template,
            iptables=iptables,
            firewall_conf=firewall_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def zone_config(self):
        """Render the firewall zone configuration page"""
        _template = plugin_template(self, "templates/zone_config.mako")
        firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
        validation_obj = firewall_conf.validation_obj()
        validation_dict = firewall_conf.validation_dict(validation_obj)
        iptables = IPTables()

        return self.standard_render(
            template=_template,
            iptables=iptables,
            firewall_conf=firewall_conf,
            validation_dict=validation_dict,
        )

    @require(has_write('Network'))
    def update_zone(self, zone=None, **kwargs):
        """Updates /etc/config/firewall with the supplied zone options.  Requires 'zone' (i.e. 'lan' or 'wan')"""
        if not zone:
            return jquery_ui_error_msg(_('Could not update firewall zone settings.  Please make sure to pass this function a "zone" parameter.'))
        try:
            conf_position = None
            firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
            for index, section in enumerate(firewall_conf.sections):
                if "zone" in section['config'] and zone == section['options']['name']:
                    conf_position = index
            if 'delete' in kwargs.keys(): # This interface has been marked for deletion
                del firewall_conf.sections[conf_position] # Remove this config section
                firewall_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall zone has been deleted.'))
            firewall_conf.sections[conf_position]['options'] = {}
            for key, value in kwargs.items():
                #print "%s\t%s" % (key, value) # For debugging
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    firewall_conf.sections[conf_position]['options'].update({key: value})
            firewall_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update firewall settings.<br />%s') % e)

    @require(has_write('Network'))
    def update_forwarding(self, forwarding=None, **kwargs):
        """Updates /etc/config/firewall with the supplied forwarding options.  Requires 'forwarding'"""
        if not forwarding:
            return jquery_ui_error_msg(_('Could not update firewall TCP/IP forwarding settings.  Please make sure to pass this function a "forwarding" parameter.'))
        # 'forwarding': '1_lan_wan'
        try:
            conf_position = None
            firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
            for index, section in enumerate(firewall_conf.sections):
                sorted_values = section['options'].values()
                sorted_values.sort()
                if "_".join(sorted_values) == forwarding:
                    conf_position = index
            if 'delete' in kwargs.keys(): # This interface has been marked for deletion
                del firewall_conf.sections[conf_position] # Remove this config section
                firewall_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall TCP/IP forwarding section has been deleted.'))
            firewall_conf.sections[conf_position]['options'] = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    firewall_conf.sections[conf_position]['options'].update({key: value})
            firewall_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update firewall settings.<br />%s') % e)

    @require(has_write('Network'))
    def update_rule(self, rule=None, **kwargs):
        """Updates /etc/config/firewall with the supplied firewall rule options.  Requires 'rule'"""
        if not rule:
            return jquery_ui_error_msg(_('Could not update firewall rule.  Please make sure to pass this function a "rule" parameter.'))
        try:
            conf_position = None
            firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
            for index, section in enumerate(firewall_conf.sections):
                sorted_values = section['options'].values()
                sorted_values.sort()
                sorted_values = "_".join(sorted_values)
                sorted_values = sorted_values.replace('.', '_')
                sorted_values = sorted_values.replace(':', '_')
                if sorted_values == rule:
                    conf_position = index
            if 'delete' in kwargs.keys(): # This interface has been marked for deletion
                del firewall_conf.sections[conf_position] # Remove this config section
                firewall_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall rule has been deleted.'))
            firewall_conf.sections[conf_position]['options'] = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    firewall_conf.sections[conf_position]['options'].update({key: value})
            firewall_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update firewall settings.<br />%s') % e)

    @require(has_write('Network'))
    def update_port_forwarding(self, redirect=None, **kwargs):
        """Updates /etc/config/firewall with the supplied firewall port forwarding options.  Requires 'redirect'"""
        if not redirect:
            return jquery_ui_error_msg(_('Could not update firewall port forwarding settings.  Please make sure to pass this function a "redirect" parameter.'))
        try:
            conf_position = None
            firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
            for index, section in enumerate(firewall_conf.sections):
                sorted_values = section['options'].values()
                sorted_values.sort()
                sorted_values = "_".join(sorted_values)
                sorted_values = sorted_values.replace('.', '_')
                sorted_values = sorted_values.replace(':', '_')
                if sorted_values == redirect:
                    conf_position = index
            if 'delete' in kwargs.keys(): # This interface has been marked for deletion
                del firewall_conf.sections[conf_position] # Remove this config section
                firewall_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall rule section has been deleted.'))
            firewall_conf.sections[conf_position]['options'] = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    firewall_conf.sections[conf_position]['options'].update({key: value})
            firewall_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update firewall settings.<br />%s') % e)

    @require(has_write('Network'))
    def update_firewall_defaults(self, **kwargs):
        """Updates /etc/config/firewall with the supplied defaults options."""
        try:
            conf_position = None
            firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
            for index, section in enumerate(firewall_conf.sections):
                if "defaults" in section['config']:
                    conf_position = index
            firewall_conf.sections[conf_position]['options'] = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    firewall_conf.sections[conf_position]['options'].update({key: value})
            firewall_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update firewall settings.<br />%s') % e)

    @require(has_write('Network'))
    def update_raw_firewall_user(self, firewall_user):
        """Save the raw passwd file text as submitted"""
        try:
            f = open('/etc/firewall.user', 'w')
            f.write(firewall_user)
            f.close()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/firewall.user was saved.'))
        except:
            return jquery_ui_error_msg(_('Could not save /etc/firewall.user.  Please make sure PyCI is running as root.'))

    @require(has_write('Network'))
    def add_firewall_zone(self, **kwargs):
        """Creates new firewall zone in /etc/config/firewall"""
        firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
        # First make sure there isn't already a zone with this name
        for section in firewall_conf.sections:
            if 'zone' in section['config'] and section['options']['name'] == kwargs['name']:
                return jquery_ui_error_msg(_('A firewall zone already exists with this name.  Please choose a different name and re-submit.'))
        try: # Now try creating the new zone
            options = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if value:
                        options.update({key: value})
            # Add the interface
            config_obj = OpenWRTConfig(config=['zone'], options=options)
            firewall_conf.sections.append(config_obj)
            firewall_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update firewall settings.<br />%s') % e)

    @require(has_write('Network'))
    def add_firewall_rule(self, **kwargs):
        """Creates new firewall rule in /etc/config/firewall"""
        firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
        # First make sure there isn't already a zone with this name
        sorted_kwargs = kwargs.values()
        sorted_kwargs.sort()
        try:
            sorted_kwargs = "_".join(sorted_kwargs)
        except TypeError:
            return jquery_ui_error_msg(_("You can't submit more than one of the same option."))
        sorted_kwargs = sorted_kwargs.replace('.', '_')
        sorted_kwargs = sorted_kwargs.replace(':', '_')
        for section in firewall_conf.sections:
            if 'rule' in section['config']:
                sorted_values = section['options'].values()
                sorted_values.sort()
                sorted_values = "_".join(sorted_values)
                sorted_values = sorted_values.replace('.', '_')
                sorted_values = sorted_values.replace(':', '_')
                if sorted_values == sorted_kwargs:
                    return jquery_ui_error_msg(_('A firewall rule exactly like this one already exists.  Please choose a different name and re-submit.'))
        try: # Now try creating the new zone
            options = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if value:
                        options.update({key: value})
            # Add the interface
            config_obj = OpenWRTConfig(config=['rule'], options=options)
            firewall_conf.sections.append(config_obj)
            firewall_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update firewall settings.<br />%s') % e)

    @require(has_write('Network'))
    def add_redirect(self, **kwargs):
        """Creates new firewall redirect (port forwarding) rule in /etc/config/firewall"""
        firewall_conf = ConfigFileOpenWRT('/etc/config/firewall')
        # First make sure there isn't already a zone with this name
        sorted_kwargs = kwargs.values()
        sorted_kwargs.sort()
        try:
            sorted_kwargs = "_".join(sorted_kwargs)
        except TypeError:
            return jquery_ui_error_msg(_("You can't submit more than one of the same option."))
        sorted_kwargs = sorted_kwargs.replace('.', '_')
        sorted_kwargs = sorted_kwargs.replace(':', '_')
        for section in firewall_conf.sections:
            if 'redirect' in section['config']:
                sorted_values = section['options'].values()
                sorted_values.sort()
                sorted_values = "_".join(sorted_values)
                sorted_values = sorted_values.replace('.', '_')
                sorted_values = sorted_values.replace(':', '_')
                if sorted_values == sorted_kwargs:
                    return jquery_ui_error_msg(_('A firewall rule exactly like this one already exists.  Please choose a different name and re-submit.'))
        try: # Now try creating the new zone
            options = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if value:
                        options.update({key: value})
            # Add the interface
            config_obj = OpenWRTConfig(config=['redirect'], options=options)
            firewall_conf.sections.append(config_obj)
            firewall_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Firewall settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update firewall settings.<br />%s') % e)

# Note:  This is a work in progress
    #@require(has_write('Network'))
    #def update_iptables(self, **kwargs):
        #"""Updates the permissions table with the provided operation and data..."""
        ## {'opt': '--', 'destination': '0.0.0.0/0', 'target': 'MASQUERADE', 'prot': 'all', 'source': '127.0.0.1/255.0.0.0', 'in': '*', 'id': '1', 'match': '', 'out': 'wlan0'}
        #print cherrypy.request.params
        #return _("Success")

    @require(has_read('Network'))
    def QOS(self):
        """Render the Quality of Service (QOS: traffic shaping) configuration page"""
        _template = plugin_template(self, "templates/qos.mako")

        return self.standard_render(
            template=_template,
            help=_("""<p>This section allows you to configure this host's Quality of Service (QOS) settings.  QOS, also called "traffic shaping" is the mechanism that determines which kinds of network traffic get prioritized or de-prioritized.  For instance, if you wanted to ensure that your web surfing didn't interrupt or degrade the quality of your Voice-over-IP calls you could prioritize VoIP protocols (SIP, IAX, Skype, etc) over everything else.</p>
            <p>The six tabs below represent the six sections of OpenWRT's QOS scripts:
            <dl>
            <dt>Classification Defaults</dt><dd>This section is for specifying the default classification of various types of network traffic based on their packet size and port number.  All packets are classified based on these settings before any other classifications are applied.</dd>
            <dt>Classifications</dt><dd>Here you can classify network traffic based on layer7 protocols, ipp2p (special peer-to-peer traffic detection module), ports, layer 4 protocol (tcp or udp), or TCP flags (SYN, ACK, etc).  These classifications are applied after the default classifications.</dd>
            <dt>Re-Classifications</dt><dd>These settings are for classifying packets after they've been classified under one of the aforementioned sections.  For example, even though a packet or connection may be classified as "Bulk" you probably want to keep TCP SYN and ACK packets classified as "Priority" under all circumstances.</dd>
            <dt>Interfaces</dt><dd>This tab allows you to specify which network interfaces will have QOS applied along with related settings.  <strong>Note:</strong> If you configure connbytes for a second QOS interface make sure you also configure it in the sister QOS interface.  For example, if you configured a new "wan_boost" interface with connbytes set to "0-9437184" make sure you set connbytes to "9437184-" on "wan".</dd>
            <dt>Classes</dt><dd>Here you can configure the classes that packets will ultimately be categorized into.  Unless you really know what you're doing it is best to leave these settings at their default values.</dd>
            <dt>Classgroups</dt><dd>This is where interface classgroups can be configured or created.  A classgroup contains the various classification categories packets will be categorized into (e.g. "Priority", "Bulk", etc).  Unless you really know what you're doing it is best to leave these settings at their default values.</dd>
            <dt>/etc/config/qos</dt><dd>If you so desire you can edit the qos configuration file directly under this tab.</dd>
            </dl>
            </p>"""),
            long_name=_('QOS Configuration'),
            description=_("Configure the Quality of Service (aka traffic shaping) settings on this host."),
        )

    @require(has_read('Network'))
    def qos_interfaces(self):
        """Render the QOS interfaces configuration page."""
        _template = plugin_template(self, "templates/qos_interfaces.mako")
        qos_conf = ConfigFileOpenWRT(filepath='/etc/config/qos')
        validation_obj = qos_conf.validation_obj()
        validation_dict = qos_conf.validation_dict(validation_obj)

        return self.standard_render(
            template=_template,
            qos_conf=qos_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def qos_classifications(self):
        """Render the QOS classifications configuration page."""
        _template = plugin_template(self, "templates/qos_classifications.mako")
        qos_conf = ConfigFileOpenWRT(filepath='/etc/config/qos')
        validation_obj = qos_conf.validation_obj()
        validation_dict = qos_conf.validation_dict(validation_obj)

        return self.standard_render(
            template=_template,
            qos_conf=qos_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def qos_reclassifications(self):
        """Render the QOS reclassifications configuration page."""
        _template = plugin_template(self, "templates/qos_reclassify.mako")
        qos_conf = ConfigFileOpenWRT(filepath='/etc/config/qos')
        validation_obj = qos_conf.validation_obj()
        validation_dict = qos_conf.validation_dict(validation_obj)

        return self.standard_render(
            template=_template,
            qos_conf=qos_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def qos_classes(self):
        """Render the QOS classes configuration page."""
        _template = plugin_template(self, "templates/qos_classes.mako")
        qos_conf = ConfigFileOpenWRT(filepath='/etc/config/qos')
        validation_obj = qos_conf.validation_obj()
        validation_dict = qos_conf.validation_dict(validation_obj)

        return self.standard_render(
            template=_template,
            qos_conf=qos_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def qos_classgroups(self):
        """Render the QOS classgroups configuration page."""
        _template = plugin_template(self, "templates/qos_classgroups.mako")
        qos_conf = ConfigFileOpenWRT(filepath='/etc/config/qos')
        validation_obj = qos_conf.validation_obj()
        validation_dict = qos_conf.validation_dict(validation_obj)

        return self.standard_render(
            template=_template,
            qos_conf=qos_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def qos_defaults(self):
        """Render the QOS packet classification defaults configuration page."""
        _template = plugin_template(self, "templates/qos_defaults.mako")
        qos_conf = ConfigFileOpenWRT(filepath='/etc/config/qos')
        validation_obj = qos_conf.validation_obj()
        validation_dict = qos_conf.validation_dict(validation_obj)

        return self.standard_render(
            template=_template,
            qos_conf=qos_conf,
            validation_dict=validation_dict,
        )

    @require(has_read('Network'))
    def qos_raw(self):
        """Render the raw qos configuration file edit page."""
        _template = plugin_template(self, "templates/qos_raw.mako")
        qos_conf = ConfigFileOpenWRT('/etc/config/qos')

        return self.standard_render(
            template=_template,
            raw_qos=qos_conf.raw(),
        )

    @require(has_write('Network'))
    def update_raw_qos(self, qos):
        """Save the raw qos config file text as submitted"""
        try:
            qos_conf = ConfigFileOpenWRT(filepath='/etc/config/qos')
            qos_conf.save_raw(qos)
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/config/qos was saved.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not save /etc/config/qos. <br />%s') % e)

    @require(has_read('Network'))
    def qos_options_meta(self, option=None, section=None, help=True, values=False, **data):
        """By default, return help text for the provided QOS option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section.
        If 'section' is provided, attempt to obtain valid values from the specified section
        of the validation file."""
        qos_conf = ConfigFileOpenWRT('/etc/config/qos')
        validation_obj = qos_conf.validation_obj()
        validation_dict = qos_conf.validation_dict(validation_obj)
        if values:
            try:
                values_dict = {}
                for value in qos_conf.valid_options(section, option):
                    if value == '0' or value == 'off' or value == 'disabled':
                        values_dict.update({value:"disabled"})
                    elif value == '1' or value == 'on' or value == 'enabled':
                        values_dict.update({value:"enabled"})
                    else:
                        values_dict.update({value:value})
                if values_dict:
                    return json.dumps(values_dict)
                else:
                    return "{empty:true}"
            except:
                # This is so we can determine that no list values are available for this option
                # ...in JavaScript: if (!json.empty) { do stuff... }
                return "{empty:true}"

        if help:
            try: # Note: this is only temporary until I write complete i18n-aware help for each option
                return '<strong>%s</strong>: %s' % (option, validation_dict[section][option]['title'])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_write('Network'))
    def update_qos(self, config=None, **kwargs):
        """Updates /etc/config/qos with the supplied options.  Requires 'config'"""
        if not config:
            return jquery_ui_error_msg(_('Could not update QOS settings.  Please make sure to pass this function a "config" parameter.'))
        try:
            config_type = config.split()[0]
            config_param = config.split()[1]
            conf_position = None
            qos_conf = ConfigFileOpenWRT('/etc/config/qos')
            for index, section in enumerate(qos_conf.sections):
                if " ".join(section['config']) == config:
                    conf_position = index
                elif len(section['config']) > 1 and section['config'][1] == config:
                    conf_position = index
                elif len(config.split()) > 2: # The 2nd and 3rd words are an option: value pair to match
                    val = config.split()[2]
                    if section['config'][0] == config_type:
                        if config_param in section['options'] and section['options'][config_param] == val:
                            conf_position = index
            if 'delete' in kwargs.keys(): # This interface has been marked for deletion
                del qos_conf.sections[conf_position] # Remove this config section
                qos_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> QOS Config section has been deleted.'))
            qos_conf.sections[conf_position]['options'] = {} # Empty it out before we re-fill it
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if key == "rename":
                        if config_type not in value:
                            value = [config_type, value]
                        qos_conf.sections[conf_position]['config'] = value
                    else:
                        qos_conf.sections[conf_position]['options'].update({key: value})
            qos_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> QOS settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update QOS settings.<br />%s') % e)

    @require(has_write('Network'))
    def add_qos_config(self, config=None, **kwargs):
        """Creates new qos configuration in /etc/config/qos"""
        if not config:
            return jquery_ui_error_msg(_('Could not add new QOS configuration.  Please provide a config option.'))
        qos_conf = ConfigFileOpenWRT('/etc/config/qos')
        if isinstance(config, list):
            config_type = config[0]
            config_param = config[1]
            config = " ".join(config)
        for section in qos_conf.sections:
            if config == " ".join(section['config']):
                if config not in ['classify', 'reclassify', 'default']:
                    return jquery_ui_error_msg(_('A QOS class already exists with this name.  Please choose a different interface and re-submit.'))
        try: # Now try creating the new interface
            options = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if value:
                        options.update({key: value})
            if len(config.split()) == 1:
                config_obj = OpenWRTConfig(config=[config], options=options)
            else:
                config_obj = OpenWRTConfig(config=[config_type,config_param], options=options)
            qos_conf.sections.append(config_obj)
            qos_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> QOS settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update QOS settings.<br />%s') % e)

    @require(has_read('Network'))
    def DHCPServer(self):
        """Render the DHCP Server (dnsmasq) configuration page"""
        _template = plugin_template(self, "templates/dhcp.mako")

        return self.standard_render(
            template=_template,
            help=_('This page lets you configure the dnsmasq daemon on your host via the /etc/config/dhcp file (not /etc/dnsmasq.conf).'),
            long_name=_('DHCP Configuration'),
            description=_("Configure the dnsmasq DHCP daemon on OpenWRT."),
        )

    @require(has_write('Network'))
    def update_raw_dhcp(self, dhcp):
        """Save the raw dhcp config file text as submitted"""
        try:
            dhcp_conf = ConfigFileOpenWRT(filepath='/etc/config/dhcp')
            dhcp_conf.save_raw(dhcp)
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/config/dhcp was saved.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not save /etc/config/dhcp. <br />%s') % e)

    @require(has_write('Network'))
    def update_dhcp(self, config=None, **kwargs):
        """Updates /etc/config/dhcp with the supplied options.  Requires 'config' (i.e. 'dhcp lan' or 'dnsmasq')"""
        if not config:
            return jquery_ui_error_msg(_('Could not update dhcp settings.  Please make sure to pass this function a "config" parameter.'))
        try:
            conf_position = None
            dhcp_conf = ConfigFileOpenWRT('/etc/config/dhcp')
            for index, section in enumerate(dhcp_conf.sections):
                if " ".join(section['config']) == config:
                    conf_position = index
            if 'delete' in kwargs.keys(): # This interface has been marked for deletion
                del dhcp_conf.sections[conf_position] # Remove this config section
                dhcp_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> DHCP Config section has been deleted.'))
            dhcp_conf.sections[conf_position]['options'] = {} # Empty it out before we refill it (this makes sure any deleted options are removed)
            for key, value in kwargs.items():
                #print "%s\t%s" % (key, value) # For debugging
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if key == "rename":
                        if 'dhcp' not in value:
                            value = ['interface', value] # Interface DHCP configurations need to use the following format: 'dhcp <ifname>'
                        dhcp_conf.sections[conf_position]['config'] = value
                    else:
                        dhcp_conf.sections[conf_position]['options'].update({key: value})
            dhcp_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> DHCP settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update DHCP settings.<br />%s') % e)

    @require(has_write('Network'))
    def add_dhcp_interface(self, interface=None, **kwargs):
        """Creates new dhcp interface configuration elements in /etc/config/dhcp"""
        if not interface:
            return jquery_ui_error_msg(_('Could not add new DHCP interface.  Please provide an interface option.'))
        dhcp_conf = ConfigFileOpenWRT('/etc/config/dhcp')
        # First make sure there isn't already a dhcp interface with this name
        for section in dhcp_conf.sections:
            if 'interface' in section['config'] and section['config'][1] == interface:
                return jquery_ui_error_msg(_('A DHCP interface already exists with this name.  Please choose a different interface and re-submit.'))
        try: # Now try creating the new interface
            options = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if value:
                        options.update({key: value})
            # Add the interface
            options.update({'interface': interface})
            config_obj = OpenWRTConfig(config=['dhcp',interface], options=options)
            dhcp_conf.sections.append(config_obj)
            dhcp_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> DHCP settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update DHCP settings.<br />%s') % e)

    @require(has_read('Network'))
    def dhcp_raw(self):
        """Render the raw dhcp configuration file edit page."""
        _template = plugin_template(self, "templates/dhcp_raw.mako")
        dhcp_conf = ConfigFileOpenWRT('/etc/config/dhcp')

        return self.standard_render(
            template=_template,
            raw_dhcp=dhcp_conf.raw(),
        )

    @require(has_read('Network'))
    def ethers(self):
        """Render the raw /etc/ethers configuration file edit page."""
        _template = plugin_template(self, "templates/ethers.mako")
        # OpenWRT actually overwrites /etc/ethers so local changes are meant to go in /etc/ethers.local
        if os.path.exists('/etc/ethers'):
            ethers = open('/etc/ethers').read()
        else:
            ethers = "# There is no /etc/ethers file on this system.  You may use this form to create one."

        return self.standard_render(
            template=_template,
            ethers=ethers,
        )

    @require(has_read('Network'))
    def ethers_json(
        self,
        rows=None,
        sidx=None,
        _search=None,
        searchField=None,
        searchOper=None,
        searchString=None,
        page=None,
        sord=None,
        nd=None):
        """Render the /etc/ethers file in jqGrid json format"""
        if os.path.exists('/etc/ethers'):
            ethers = open('/etc/ethers')
        else:
            ethers = "\n"
        header = [_('mac'), _('ip')]
        ethers_list = []
        for line in ethers:
            if line: # filter out blank lines
                line = line.split()
                ethers_list.append(line)
        return jqgrid_json(
            self,
            ethers_list,
            header,
            rows=rows,
            sidx=sidx,
            _search=_search,
            searchField=searchField,
            searchOper=searchOper,
            searchString=searchString,
            page=page,
            sord=sord
        )

    @require(has_write('Network'))
    def update_ethers(self, oper=None, ip=None, id=None, mac=None):
        """
        Updates /etc/ethers with the given parameters.
        """
        out_ethers = ""
        ethers = "/etc/ethers"
        if oper == "del": # We're' deleting a line
            if os.path.exists(ethers):
                # Note: This preserves comments =)
                for line in open(ethers):
                    if line.lstrip().startswith(id): # id == mac in this case
                        pass # We don't want this one
                    else:
                        out_ethers += line
                etc_ethers = open(ethers, 'w')
                etc_ethers.write(out_ethers)
                etc_ethers.close()
        elif oper == "add": # Adding a new ip/mac
            if os.path.exists(ethers):
                # First we double-check that /etc/hosts ends in a newline character:
                for line in open(ethers):
                    last_line = line
                if not last_line.endswith('\n'): # No newline?!? Fix it
                    # 'with .." ensures the file is closed after write() is complete
                    with open(ethers, 'a') as f:
                        f.write('\n') # Write that missing newline
            etc_ethers = open(ethers, 'a') # Append mode, love it!
            etc_ethers.write("%s %s\n" % (mac, ip))
            etc_ethers.close()
        else: # We're doing an update
            if os.path.exists(ethers):
                for line in open(ethers):
                    if line.lstrip().startswith(id): # Note: that's 'id' not 'ip'
                        # Replace with our modified version
                        out_ethers += "%s %s\n" % (ip, mac)
                    else: # Leave it alone
                        out_ethers += line
                etc_ethers = open(ethers, 'w')
                etc_ethers.write(out_ethers)
                etc_ethers.close()

    @require(has_write('Network'))
    def update_raw_ethers(self, ethers):
        """Save the raw /etc/ethers file as submitted"""
        try:
            f = open('/etc/ethers', 'w')
            f.write(ethers)
            f.close()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/ethers was saved.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not save /etc/ethers.  Please make sure PyCI is running as root.</br>%s') % e)

    @require(has_read('Network'))
    def get_ethers_spreadsheet(self, table=None, chain=None):
        """Exports /etc/ethers as a spreadsheet"""
        spreadsheet_template = lookup_template('html', 'spreadsheet.mako')
        cherrypy.response.headers['Content-Type'] = 'application/vnd.ms-excel; charset=utf-8'
        headers = ['MAC Address', 'IP Address']
        ethers = open('/etc/ethers')
        ethers_list = []
        for line in ethers:
            if line: # filter out blank lines
                line = line.split()
                ethers_list.append(line)
        return spreadsheet_template.render(headers=headers, rows=ethers_list)

    @require(has_read('Network'))
    def dhcp_config(self):
        """Render the dhcp configuration tab"""
        _template = plugin_template(self, "templates/dhcp_config.mako")

        dhcp_conf = ConfigFileOpenWRT(filepath='/etc/config/dhcp')
        validation_obj = dhcp_conf.validation_obj()
        validation_dict = dhcp_conf.validation_dict(validation_obj)

        return self.standard_render(
            template=_template,
            dhcp_conf=dhcp_conf,
            dhcp_interface_options=self.dhcp_interface_options,
            validation_dict=validation_dict,
            dnsmasq_options=self.dnsmasq_options,
        )

    @require(has_read('Network'))
    def new_dhcp_interface(self):
        """Render the new dhcp interface tab"""
        _template = plugin_template(self, "templates/new_dhcp_interface.mako")

        return self.standard_render(
            template=_template,
            dhcp_conf=ConfigFileOpenWRT(filepath='/etc/config/dhcp'),
            dhcp_interface_options=self.dhcp_interface_options,
        )

    @require(has_read('Network'))
    def dnsmasq_options_help(self, option=None, dump=False, **data):
        """Return help text for the provided option (for jTip)"""
        # TODO:  Make sure we have definitions for all options
        options_dict = {
            "interface": _('Yes, this is redundant:  The interface with which this configuration pertains.'),
            "dhcp_option": _('Use this setting for the various dnsmasq options.  See "dnsmasq --help" for details.'),
            "dynamicdhcp": _('I have no idea what this option is for.  This product is open source--care to share?'),
            "force": _('Use this setting for the various dnsmasq options.  See "dnsmasq --help" for details.'),
            # TODO:  Finish these
            #"ignore",
            #"leasetime",
            #"limit",
            #"name",
            #"netmask",
            #"start"
        }

        if dump:
            return options_dict
        try:
            return '<strong>%s</strong>: %s' % (option, options_dict[option])
        except:
            return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_read('Network'))
    def dhcp_interface_options_meta(self, option=None, section=None, help=True, values=False, **data):
        """By default, return help text for the provided dhcp interface option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section.
        If 'section' is provided, attempt to obtain valid values from the specified section
        of the validation file."""
        dhcp_conf = ConfigFileOpenWRT('/etc/config/dhcp')
        validation_obj = dhcp_conf.validation_obj()
        validation_dict = dhcp_conf.validation_dict(validation_obj)
        if values:
            try:
                values_dict = {}
                for value in dhcp_conf.valid_options(section, option):
                    if value == '0' or value == 'off' or value == 'disabled':
                        values_dict.update({value:"disabled"})
                    elif value == '1' or value == 'on' or value == 'enabled':
                        values_dict.update({value:"enabled"})
                    else:
                        values_dict.update({value:value})
                if values_dict:
                    return json.dumps(values_dict)
                else:
                    return "{empty:true}"
            except:
                # This is so we can determine that no list values are available for this option
                # ...in JavaScript: if (!json.empty) { do stuff... }
                return "{empty:true}"

        if help:
            try: # Note: this is only temporary until I write complete i18n-aware help for each option
                return '<strong>%s</strong>: %s' % (option, validation_dict[section][option]['title'])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    #@require(has_read('Network'))
    #def dhcp_interface_options_meta(self, option=None, dump=False, **data):
        #"""Return help text for the provided option (for jTip)"""
        ## TODO:  Make sure we have definitions for all options
        #options_dict = {
            #"addnhosts": _('Path to an additional host file.'),
            #"authoritative": _('Set this to 1 if this will be the only DNS server on the local network.'),
            #"boguspriv": _('If set to "1", dnsmasq will not forward reverse lookups for local networks.'),
            #"dhcp_boot": _('DHCP boot image location.'),
            #"dhcpleasemax": _('Maximum number of allowed DHCP leases.'),
            #"dnsforwardmax": _('Maximum number of allowed concurrent queries.'),
            #"domain": _('Name of the local network domain (e.g. "lan").'),
            #"domainneeded": _('If set to "1" dnsmasq will not forward domain lookups that do not contain an FQDN.'),
            ##"ednspacket_max": _('No idea what this means.'),
            #"enable_tftp": _('If set to "1", enable the built-in tftp server (for network boot environments).'),
            #"expandhosts": _('If set to "1", append the domain name set in the "domain" option to hostnames in the hosts file (e.g. "server" becomes "server.lan".'),
            #"filterwin2k": _('If set to "1", filter out some of Microsoft\'s old/abandoned attempts at a protocol monopoly.'),
            #"leasefile": _('Path to the file where lease information will be stored.'),
            ## TODO:  Finish these
            ##"local":
            ##"localise_queries",
            ##"logqueries",
            ##"nohosts",
            ##"nonegcache",
            ##"noresolv",
            ##"port",
            ##"queryport",
            ##"readethers",
            ##"resolvfile",
            ##"strictorder",
            ##"tftp_root"
        #}

        #if dump:
            #return options_dict
        #try:
            #return '<strong>%s</strong>: %s' % (option, options_dict[option])
        #except:
            #return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_read('Network'))
    def dhcp_status(self):
        """Render the dhcp status tab (whether or not it is running, current leases, etc)"""
        _template = plugin_template(self, "templates/dhcp_status.mako")

        return self.standard_render(
            template=_template,
        )

    @require(has_read('Network'))
    def dhcp_leases_json(
        self,
        rows=None,
        sidx=None,
        _search=None,
        searchField=None,
        searchOper=None,
        searchString=None,
        page=None,
        sord=None,
        nd=None):
        """Render the dnsmasq DHCP lease table in jqGrid json format"""
        import time
        dhcp_conf = ConfigFileOpenWRT(filepath='/etc/config/dhcp')
        dnsmasq_config = [a for a in dhcp_conf.sections if 'dnsmasq' in a['config']][0] # get the dnsmasq config section
        lease_filepath = dnsmasq_config['options']['leasefile'] # ...so we can get the lease file path
        # Don't need these anymore
        del dnsmasq_config
        del dhcp_conf
        header = [_('Expiration'), _('MAC Address'), _('IP Address'), _('Computer Name'), _('Client ID')]
        f = open(lease_filepath, 'r')
        # This is an example line from the lease file:
        # 1245374230 52:54:00:12:34:56 10.111.111.139 * 01:52:54:00:12:34:56
        lease_list = []
        for line in f:
            if line: # filter out blank lines
                line = line.split()
                lease_list.append(line)
        # Note: I'm using 'Client ID' as the jqGrid ID field (via
        #       id_field_index=4) so that it gets passed to update_dhcp_lease
        #       properly
        return jqgrid_json(
            self,
            lease_list,
            header,
            id_field_index=4,
            rows=rows,
            sidx=sidx,
            _search=_search,
            searchField=searchField,
            searchOper=searchOper,
            searchString=searchString,
            page=page,
            sord=sord
        )

    @require(has_write('Network'))
    def dhcp_service_control(self, operation):
        """Based on the 'operation' parameter, stop, start, or restart the dhcp server."""
        # TODO:  Finish this
        try:
            dhcp_conf=ConfigFileOpenWRT(filepath='/etc/config/dhcp')
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/config/dhcp was saved.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not save /etc/config/dhcp. <br />%s') % e)

    @require(has_write('Network'))
    def update_dhcp_lease(self, oper=None, id=None, permission=None, user=None, plugin=None):
        """Performs the specified operation (oper) on the specified dhcp lease(s).
        Currently only supports deleting but I don't see any reason why I can't revisit this in the future to add more fancy stuff."""
        try:
            dhcp_conf = ConfigFileOpenWRT(filepath='/etc/config/dhcp')
            dnsmasq_config = [a for a in dhcp_conf.sections if 'dnsmasq' in a['config']][0] # get the dnsmasq config section
            lease_filepath = dnsmasq_config['options']['leasefile'] # ...so we can get the lease file path
            # Don't need these anymore
            del dnsmasq_config
            del dhcp_conf
            lease_file = open(lease_filepath, 'r')
            if oper == "del":
                if isinstance(id, list):
                    for item in id.split(','):
                        delete_dhcp_lease(id)
                else:
                    delete_dhcp_lease(id)
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> DHCP lease, "%s" has been deleted.') % id)
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Got an exception trying to run delete_dhcp_lease():<br />%s') % e)

    @require(has_read('Network'))
    def DDNS(self):
        """Render the dynamic DNS configuration page"""
        _template = plugin_template(self, "templates/ddns.mako")

        return self.standard_render(
            template=_template,
            help=_('This section allows you to configure Dynamic DNS settings on this host.'),
            long_name=_('Dynamic DNS Configuration'),
            description=_("Configure Dynamic DNS settings."),
        )

    @require(has_read('Network'))
    def ddns_config(self):
        """Render the Dynamic DNS configuration tab"""
        _template = plugin_template(self, "templates/ddns_config.mako")

        ddns_conf = ConfigFileOpenWRT(filepath='/etc/config/ddns')
        validation_obj = ddns_conf.validation_obj()
        validation_dict = ddns_conf.validation_dict(validation_obj)

        return self.standard_render(
            template=_template,
            ddns_conf=ddns_conf,
            validation_dict=validation_dict,
        )

    @require(has_write('Network'))
    def update_raw_ddns(self, ddns):
        """Save the raw ddns config file text as submitted"""
        try:
            ddns_conf = ConfigFileOpenWRT(filepath='/etc/config/ddns')
            ddns_conf.save_raw(ddns)
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/config/ddns was saved.'))
        except Exception as e:
            e = str(e).replace('\n', '<br />')
            return jquery_ui_error_msg(_('Could not save /etc/config/ddns. <br />%s') % e)

    @require(has_write('Network'))
    def update_ddns(self, config=None, **kwargs):
        """Updates /etc/config/ddns with the supplied options.  Requires 'config' which should be represent
        the original config line (e.g. 'config "mydns"').  It is used to find the original entry in the
        event that the user is changing it.
        Note: The hostname appears to be the only unique identifier for each of these ddns entries."""
        if not config:
            return jquery_ui_error_msg(_('Could not update Dynamic DNS settings.  Please make sure to pass this function a "config" parameter.'))
        try:
            conf_position = None
            ddns_conf = ConfigFileOpenWRT(filepath='/etc/config/ddns')
            for index, section in enumerate(ddns_conf.sections):
                if config in section['config']:
                    conf_position = index
            if 'delete' in kwargs.keys(): # This DDNS entry has been marked for deletion
                del ddns_conf.sections[conf_position] # Remove this config section
                ddns_conf.save()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> Dynamic DNS config for %s has been deleted.') % `config`)
            ddns_conf.sections[conf_position]['options'] = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if key == "rename":
                        ddns_conf.sections[conf_position]['hostname'] = value
                    else:
                        ddns_conf.sections[conf_position]['options'].update({key: value})
            ddns_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Dynamic DNS settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update Dynamic DNS settings.<br />%s') % e)

    @require(has_write('Network'))
    def add_ddns(self, config=None, **kwargs):
        """Creates a new Dynamic DNS configuration in /etc/config/ddns.  Requires 'config'."""
        if not config:
            return jquery_ui_error_msg(_('Could not add new Dynamic DNS interface.  Please provide a "config" parameter.'))
        ddns_conf = ConfigFileOpenWRT(filepath='/etc/config/ddns')
        # First make sure there isn't already a ddns config with this name
        for section in ddns_conf.sections:
            if config in section['config']:
                return jquery_ui_error_msg(_('A Dynamic DNS configuration already exists with this name.  Please choose a different name and re-submit.'))
        try:
            options = {}
            for key, value in kwargs.items():
                if key != 'options': # This gets submitted along with the form due to the way my dynamic form works.
                    if value:
                        options.update({key: value})
            config_obj = OpenWRTConfig(config=['service', config], options=options)
            ddns_conf.sections.append(config_obj)
            ddns_conf.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Dynamic DNS settings have been updated.'))
        except Exception, e:
            print e
            return jquery_ui_error_msg(_('Could not update Dynamic DNS settings.<br />%s') % e)

    @require(has_read('Network'))
    def ddns_raw(self):
        """Render the raw /etc/config/ddns file edit page."""
        _template = plugin_template(self, "templates/ddns_raw.mako")
        ddns_conf = ConfigFileOpenWRT(filepath='/etc/config/ddns')

        return self.standard_render(
            template=_template,
            raw_ddns=ddns_conf.raw(),
        )

    @require(has_read('Network'))
    def ddns_options_meta(self, option=None, section=None, help=True, values=False, **data):
        """By default, return help text for the provided Dynamic DNS option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section.
        If 'section' is provided, attempt to obtain valid values from the specified section
        of the validation file."""
        ddns_conf = ConfigFileOpenWRT(filepath='/etc/config/ddns')
        validation_obj = ddns_conf.validation_obj()
        validation_dict = ddns_conf.validation_dict(validation_obj)
        if values:
            try:
                values_dict = {}
                for value in ddns_conf.valid_options(section, option):
                    if value == '0' or value == 'off' or value == 'disabled':
                        values_dict.update({value:"disabled"})
                    elif value == '1' or value == 'on' or value == 'enabled':
                        values_dict.update({value:"enabled"})
                    else:
                        values_dict.update({value:value})
                if values_dict:
                    return json.dumps(values_dict)
                else:
                    return "{empty:true}"
            except:
                # This is so we can determine that no list values are available for this option
                # ...in JavaScript: if (!json.empty) { do stuff... }
                return "{empty:true}"

        if help:
            try: # Note: this is only temporary until I write complete i18n-aware help for each option
                return '<strong>%s</strong>: %s' % (option, validation_dict[section][option]['title'])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_read('Network'))
    def index(self):
        """Return the index page"""
        return self.Interfaces()