# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

"""
openwrt.py: A collection of classes to simplify the reading and writing of
OpenWRT's /etc/config/* files.
"""

import os, re, gettext
from utils import format_columns
from configfile import ConfigFile
from odict import odict

locale_dir = os.path.join(os.getcwd(), 'i18n')
def get_translation():
    try:
        trans = gettext.translation('PyCI', locale_dir)
    except:
        trans = gettext.NullTranslations()
    return trans
trans = get_translation()
_ = lambda s: trans.ugettext(s).encode('utf-8')

def surround_with_quotes(_list):
    """
    Takes a list of items and returns a string with each list item surrounded
    by single quotes (for UCI stuff)
    """
    out_string = ""
    for item in _list:
        out_string += "'" + item + "'"
        out_string += " "
    return out_string

class OpenWRTConfig(object):
    """
    Class to store config sections from /etc/config/* files.

    OpenWRTConfig object parameters can be accessed as a dictionary...
        >>> config = ConfigFileOpenWRT(filepath='/etc/config/network')
        >>> test = config.configs() # test is an OpenWRTConfig object
        >>> print test[0]['config']
        interface lan
    ...or directly as attributes:
        >>> print test[0].config
        interface lan

    Parameters can be set the same way:
        >>> test[0]['config'] = 'rule'
        >>> test[0].config = 'rule'
    """
    def __init__(self, config, options, list_items=[]):
        """Sets up a dictionary of options in this config section.

        'config' is expected to be a list such as ["interface", "lan"].
        'options' is expected to be a dictionary of key:value pairs such as, {'ifname': 'eth1'}
        'list_items' is for luci-style dependency/validation files."""
        self.config = config
        self.options = options
        self.list_items = list_items

    def asList(self):
        """Returns our parameters as a list"""
        config_list = [self.config]
        for key, value in self.options.items():
            if isinstance(value, list):
                if value:
                    config_list.append("%s %s" % (key, value))
            else:
                config_list.append(" ".join([key, value]))
        for item in self.list_items:
            config_list.append(item)
        return config_list

    def __str__(self):
        """Returns the config as it would appear in an /etc/config/* file."""
        out_string = "config " + surround_with_quotes(self.config) + "\n"
        option_list = []
        list_items_list = []
        for option, value in self.options.items():
            option_list.append("'" + option + "'")
            option_list.append("'" + value + "'")
        for list_item in self.list_items:
            list_items_list.append('   depends')
            list_items_list.append("'" + list_item + "'")

        for line in format_columns(option_list, 2): # format the options into evenly-spaced columns
            out_string += "\toption %s\n" % line.rstrip() # format_columns puts unnecessary whitespace at the end of things
        for line in format_columns(list_items_list, 2):
            out_string += "\tlist %s\n" % line.rstrip()
        return out_string.rstrip('\n') # Get rid of the trailing newline

    def d_str(self):
        """
        Returns the config file in dependency/validation file format which means...
        No quotes for 2nd field (option/list_item).
        No quotes for config (variable, section, or package).

        Example:
            config package
                    option title 'Network configuration'
            config section
                    option named    'true'
                    option required 'true'
                    option name     'interface'
                    option package  'network'
                    option title    'Network interface'
                    list   depends  'proto=none'
        """
        out_string = "config " + " ".join(self.config) + "\n"
        option_list = []
        list_items_list = []
        # Validation/dependency files don't use quotes for the 2nd field (option/list_item)
        for option, value in self.options.items():
            option_list.append(option)
            option_list.append("'" + value + "'")
        for list_item in self.list_items:
            list_items_list.append('depends')
            list_items_list.append("'" + list_item + "'")

        options_count = (len(option_list) / 2)
        # Format the options into evenly-spaced columns:
        for index, line in enumerate(format_columns(option_list+list_items_list, 2)):
            if index < options_count:
                # format_columns puts unnecessary whitespace at the end of things
                out_string += "\toption %s\n" % line.rstrip()
            else:
                out_string += "\tlist   %s\n" % line.rstrip()

        return out_string.rstrip('\n') # Get rid of the trailing newline

    def __repr__(self):
        """Returns a string representation of a dictionary made up of self.config and self.options.

        Example:
            repr(test_config[0])
            {'config': ['interface', 'lan'], 'options': {'netmask': '255.255.255.0', 'dns': '192.168.1.1', 'proto': 'static', 'ifname': 'eth0.0', 'ipaddr': '192.168.1.2', 'gateway': '192.168.1.1'}}"""
        return str({'config': self.config, 'options': self.options, 'list_items': self.list_items})

    def __iter__(self):
        return self.asList().__iter__()

    def __len__(self):
        return len(self.asList())

    def __contains__(self, val):
        for item in self.asList():
            if item == val:
                return True
        else:
            return False

    def __getitem__(self, val):
        """This lets you access OpenWRTConfig objects as a dictionary or a list."""
        try:
            return self.__dict__[val]
        except:
            try:
                return self.asList()[val]
            except:
                return None

    def __setitem__(self, key, value):
        """
        This lets you set/assign attributes to OpenWRTConfig objects like a dictionary.
        Note: This is really only necessary for setting for 'config' values.

        Example:
            >>> config_obj['config'] = 'rule'
        """
        self.__dict__[key] = value

class ValidationException(Exception): pass

# I might want to use these validation objects in the future
class ValidationDependency(object):
    """Class to store 'list depends whatever=foo, bar' objects from luci-style
    validation files"""
    def __init__(self, option_requirement, required_option):
        # option_requirement is for things like 'list depends encryption=wpa'
        self.option_requirement = option_requirement
        # required_option is for things like 'list depends ipaddr'
        self.required_option = required_option

class ValidationEnumerator(object):
    """Class to store enumerators from a luci-style validation file"""
    def __init__(self, variable, value, title):
        self.variable = variable
        self.value = value
        self.title = title

class ValidationSection(object):
    """Class to store the various requirements and variables that are associated with
    any given 'section' in a luci-style validation file"""
    def __init__(self, name, named=False, required=False, title=None, package=None, depends=[], variables=[]):
        self.name = name
        self.named = named
        self.required = required
        self.title = title
        self.package = package
        self.depends = depends
        self.variables = variables

class ValidationVariable(object):
    """Class to store various requirements associated with any given 'variable' in a
    luci-style validation file."""
    def __init__(self, name, section, title=None, variable_type=None, required=False, depends=[], valueof=None):
        self.name = name
        self.section = section
        self.title = title
        self.variable_type = variable_type
        self.required = required
        self.depends = depends
        self.valueof = valueof

class ConfigFileOpenWRT(ConfigFile):
    """
    Class for accessing/saving config files in /etc/config.
    When initialized with a filepath this class will contain a list of
    OpenWRTConfig objects representing the contents of filepath.
    """

    def __init__(self, filepath=None):
        """
        Read a UCI-formatted (/etc/config/*) file into a list of
        OpenWRTConfig objects each with it's own options.

        Example usage:
            >>> network = ConfigFileOpenWRT('/etc/config/network')
            >>> print network.sections[0]
            config interface lan
                option 'netmask' '255.255.255.0'
                option 'dns'     '192.168.1.1'
                option 'proto'   'static'
                option 'ifname'  'eth0.0'
                option 'ipaddr'  '192.168.1.2'
                option 'gateway' '192.168.1.1'
            >>> network.sections[0]['options']['dns'] = '4.2.2.1'
            >>> network.save()
            >>> print network.sections[0].options['dns']
            4.2.2.1
        """
        self.filepath = filepath
        if self.filepath:
            self.sections = self.read_config_file()

    def read_config_file(self, filepath=None):
        """Reads self.filepath and returns a list of OpenWRTConfig objects representing the file."""
        # Note: If this method seems wacky it is because I added the list_item stuff *after* the config/options part was done.
        #       It works well regardless =)
        if not filepath:
            filepath = self.filepath
        configs_list = []
        configs = []
        list_items_list = []
        count = 0
        quoted_regex = re.compile('\'(.*?)\'', re.IGNORECASE) # This matches things 'in quotes'
        double_quoted_regex = re.compile('"(.*?)"', re.IGNORECASE) # This matches things "in quotes"
        if os.path.exists(filepath):
            config_file = open(filepath, 'r')
            for line in config_file:
                line = line.rstrip('\n').rstrip() # Get rid of trailing whitespace and newline characters
                if line.rstrip().startswith('#'): # This line is a comment
                    pass # TODO: Figure out a way to preserve comments (and their position)
                elif line.startswith('config'):
                    if "'" in line: # This config uses single quotes
                        config = quoted_regex.split(line.lstrip('config ')) # The word 'config' is redundant
                        config = [a.replace('\t', ' ') for a in config] # replace tabs with spaces
                        config = [a.strip() for a in config] # Get rid of leading and trailing whitespace in the list items
                        config = [a for a in config if a != ''] # Get rid of empty items (typically at the end)
                    elif '"' in line: # This config uses double quotes
                        config = double_quoted_regex.split(line.lstrip('config')) # The word 'config' is redundant
                        config = [a.replace('\t', ' ') for a in config] # replace tabs with spaces
                        config = [a.strip() for a in config] # Get rid of leading and trailing whitespace in the list items
                        config = [a for a in config if a != ''] # Get rid of empty items (typically at the end)
                    else:
                        config = line.split()[1:] # The word 'config' is redundant
                    configs_list.append({'config': config})
                    list_items_list.append([])
                    count += 1
                elif line.lstrip().startswith('option'):
                    if "'" in line: # These options are in single quotes
                        options = quoted_regex.split(line.lstrip().lstrip('option').lstrip()) # The word 'option' is redundant
                        options = [a.replace('\t', ' ') for a in options] # replace tabs with spaces
                        options = [a.strip() for a in options] # Get rid of leading and trailing whitespace in the list items
                        options = [a for a in options if a != ''] # Get rid of empty items (typically at the end)
                    elif '"' in line: # These options are in double quotes
                        options = double_quoted_regex.split(line.lstrip().lstrip('option').lstrip()) # The word 'option' is redundant
                        options = [a.replace('\t', ' ') for a in options] # replace tabs with spaces
                        options = [a.strip() for a in options] # Get rid of leading and trailing whitespace in the list items
                        options = [a for a in options if a != ''] # Get rid of empty items (typically at the end)
                    else: # They're just separated by spaces
                        options = line.split()[1:] # The word 'option' is redundant
                    # Append each option to the dictionary as-is (the 'config' section gets separated out below)
                    if len(options) == 1: # The second option is empty (i.e. '')
                        options.append('') # Add an empty string to the end of the list so we don't get IndexError exeptions
                    configs_list[count-1].update({options[0]: options[1]})
                elif line.lstrip().startswith('list'): # This is a validation/dependency file
                    if "'" in line: # These options are in quotes
                        list_item = quoted_regex.split(line.lstrip().lstrip('list').lstrip())
                        list_item = [a.replace('\t', ' ') for a in list_item] # replace tabs with spaces
                        list_item = [a.strip() for a in list_item] # Get rid of leading and trailing whitespace in the list items
                        list_item = [a for a in list_item if a != ''] # Get rid of empty items (typically at the end)
                    else: # They're just separated by spaces
                        list_item = line.split()[1:]
                    # Append each option to the dictionary as-is (the 'config' section gets separated out below)
                    if len(list_item) == 1: # The second option is empty (i.e. '')
                        list_item.append('') # Add an empty string to the end of the list so we don't get IndexError exeptions
                    list_items_list[count-1].append(list_item)

            for config, list_items in zip(configs_list, list_items_list):
                config_list = config['config']
                del config['config'] # Take it out so it doesn't get mixed up in the options
                #del list_items['config']
                temp = OpenWRTConfig(config=config_list, options=config, list_items=list_items) # Assemble a config object
                configs.append(temp) # Append it to the configs list
        else:
            configs = self.empty_config()
        return configs

    def empty_config(self):
        """Returns an empty list/dict object in the event we're working with an empty or non-existant file."""
        empty_conf_object = [{'config': '', 'options': {}}]
        return empty_conf_object

    def save(self):
        """
        Saves the config file

        The standard OpenWRT config file format is like so:
        config section option
        <tab>option someoption someoption ...
        <tab>option someoption someoption ...
        """
        self.validate()
        self.make_backup(self.filepath)
        #self.make_init() # TODO:  Implement this capability
        f = open(self.filepath, 'w')
        f.write(str(self))
        f.close()
        # Touch the '/tmp/pyci_changes' file so the UI knows a reboot is needed
        self.made_changes()

    def save_raw(self, text):
        """Saves the raw (plaintext) version of the config file using 'text'.
        This overrides the default save_raw function in order to provide validation."""
        self.validate()
        self.make_backup(self.filepath)
        f = open(self.filepath, 'w')
        f.write(text)
        f.close()
        self.made_changes()

    def __str__(self):
        """Returns a string representing how the config file will look in /etc/config/"""
        out_string = ""
        list_items = False
        for config in self.sections:
            if config.list_items:
                list_items = True
        if list_items:
            for config in self.sections:
                out_string += config.d_str() + '\n'
        else:
            for config in self.sections:
                out_string += str(config) + '\n'
        return out_string

    def __repr__(self):
        """Returns a string representing how the sections list is laid out"""
        return str(self.sections)

    def validation_obj(self, validation_file=None):
        """Returns a validation object by reading 'validation_file' using self.read_config_file."""
        if validation_file:
            validation_obj = self.read_config_file(validation_file)
        else:
            (basedir, filename) = os.path.split(self.filepath)
            validation_file = os.path.join(os.getcwd(), 'metadata/%s' % filename)
            if os.path.exists(validation_file):
                validation_obj = self.read_config_file(validation_file)
            else:
                raise IOError('Could not locate a validation file')
        return validation_obj

    def validation_dict(self, validation_obj):
        """Returns a ordered dictionary of values representing the required options, dependencies,
        and such from a given validation object (read_config_file(validation_file) of a file in
        <PyCI dir>/metadata/."""

        # These loops split each section of a given validation object into a dictionary.
        # The reason for this is because some options with the same name have different
        # requirements depending on which section they're in.
        validation_dict = odict()
        validation_sections = []
        for config in validation_obj:
            if config['config'][0] == 'section':
                validation_dict.update({config['options']['name']: odict([('list_items', [])])})

        # Add each sectional variable to the validation_dict
        for config in validation_obj:
            if config['config'][0] == 'variable':
                section = config['options']['section'].split('.')[1]
                if section not in validation_dict:
                    validation_dict[section] = odict()
                if section not in validation_sections:
                    validation_sections.append(section)
                if 'type' in config['options'] and config['options']['type'] == 'enum':
                    # Add it as an empty list so we can append to it below
                    config['options']['values'] = []
                validation_dict[section].update({config['options']['name']: config['options']})

        # Attach enumerators to their parent section options
        # Note: We re-loop here in the event that sections/options/enumerators are out of order
        for config in validation_obj:
            if config['config'][0] == 'enum':
                section = config['options']['variable'].split('.')[1]
                option = config['options']['variable'].split('.')[2]
                validation_dict[section][option]['values'].append(config['options']['value'])
                if 'enum' not in validation_dict[section][option]:
                    validation_dict[section][option]['enum'] = odict()
                validation_dict[section][option]['enum'][config['options']['value']] = config['options']
            # Now add list items (dependencies)
            if config['config'][0] == 'section':
                section = config['options']['name']
                # Create a sub-dict with the same name as the section just for list items
                validation_dict[section]['list_items'] = config['list_items']
            if config['config'][0] == 'variable':
                #print config['options']['section']
                section = config['options']['section'].split('.')[1]
                option = config['options']['name']
                validation_dict[section][option]['list_items'] = config['list_items']
        # Sort the options in each section so they show up nice in select boxes
        for section in validation_sections:
            validation_dict[section].sort(key=lambda x: x[0].lower())
        return validation_dict

    def validate(self, validation_file=None):
        """
        Validates the current configs against the definitions specified in the given validation file.
        Returns None if everything validates and raises a ValidationException if there's a problem.
        If no validation file is given it tries to be smart about it by looking in <PyCI dir>/metadata
        for a file that matches the filename in self.filepath.
        """
        import ipaddr
        mac_regex = re.compile('([a-fA-F0-9]{2}:?){6}')
        validation_obj = self.validation_obj(validation_file)
        validation_dict = self.validation_dict(validation_obj)

        # Now perform validation...
        invalid_entries = []
        for config in self.sections:
            for option, value in config['options'].items():
                section = config['config'][0]
                if option not in validation_dict[section]:
                    if option.startswith('vlan'): # vlanX is OK
                        pass
                    else:
                        invalid_entries.append(
                            _('"%s": "%s" is not a valid option within "%s"') %
                            (" ".join(config['config']), option, config['config'][0])
                        )
                elif 'datatype' in validation_dict[section][option]:
                    # Perform value validations (i.e. ip addresses, MACs, etc are in the proper format)
                    if value and validation_dict[section][option]['datatype'] == 'ip4addr':
                        try:
                            ipaddr.IPv4(value)
                        except ipaddr.IPv4IpValidationError:
                            invalid_entries.append(
                            _('"%s", %s: "%s" is not a valid IPv4 address') %
                            (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'ip6addr':
                        try:
                            ipaddr.IPv6(value)
                        except ipaddr.IPv6IpValidationError:
                            invalid_entries.append(
                            _('"%s", %s: "%s" is not a valid IPv6 address') %
                            (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'macaddr':
                        if not mac_regex.match(value):
                            invalid_entries.append(
                            _('"%s", %s: "%s" is not a valid MAC address (format is 00:00:00:00:00:00)') %
                            (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'file':
                        if not os.path.exists(value):
                            invalid_entries.append(
                            _('"%s", %s: "%s" does not exist') % (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'device':
                        # 'device' is just the path to a file in /dev
                        if not os.path.exists(value):
                            invalid_entries.append(
                            _('"%s", %s: "%s" does not exist') % (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'directory':
                        if not os.path.isdir(value):
                            invalid_entries.append(
                            _('"%s", %s: "%s" is not a directory or does not exist') % (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'boolean':
                        if value not in ['0', '1', 'on', 'off', 'enabled', 'disabled']:
                            invalid_entries.append(
                            _('"%s", %s: "%s" is not a valid boolean value (must be "1", "0", "on", "off", "enabled", or "disabled")') %
                            (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'float':
                        try:
                            float(value)
                        except ValueError: # Couldn't turn value into a float
                            invalid_entries.append(
                            _('"%s", %s: "%s" is not a floating point value') %
                            (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'uint':
                        try:
                            int(value)
                        except ValueError: # Couldn't turn value into an int
                            invalid_entries.append(
                            _('"%s", %s: "%s" is not an integer') %
                            (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'integer':
                        try:
                            int(value)
                        except ValueError: # Couldn't turn value into an int
                            invalid_entries.append(
                            _('"%s", %s: "%s" is not an integer') %
                            (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'host':
                        if not re.match("^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z]|[A-Za-z][A-Za-z0-9\-]*[A-Za-z0-9])$",
                                        value):
                            invalid_entries.append(
                            _('"%s", %s: "%s" is not a valid hostname') %
                            (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'hostname':
                        if not re.match("^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z]|[A-Za-z][A-Za-z0-9\-]*[A-Za-z0-9])$",
                                        value):
                            invalid_entries.append(
                            _('"%s", %s: "%s" is not a valid hostname') %
                            (" ".join(config['config']), option, value)
                            )
                    elif value and validation_dict[section][option]['datatype'] == 'port':
                        try:
                            if int(value) > 65534 or int(value) < 0:
                                invalid_entries.append(
                                _('"%s", %s: "%s" is not a valid port number (must be 0-65534)') %
                                (" ".join(config['config']), option, value)
                                )
                        except ValueError: # Couldn't turn value into an int
                            invalid_entries.append(
                            _('"%s", %s: "%s" is not a valid port number (must be 0-65534)') %
                            (" ".join(config['config']), option, value)
                            )

        # Now we perform structural validation (required options are set, dependencies are met, etc)
        for section, options in validation_dict.items():
            # Check that required options are set
            # TODO:  Check minlength and maxlength
            #print "section: %s" % (section)
            for option, values in options.items():
                #print "option: %s" % option
                if 'required' in values:
                    required_option = option
                    for config in self.sections:
                        if config['config'][0] == section:
                            if not required_option in config['options'].keys():
                                invalid_entries.append(
                                _('"%s" is missing option "%s"') %
                                (" ".join(config['config']), required_option)
                                )
                # Check that enum values match valid options
                if 'type' in values and values['type'] == 'enum':
                    for config in self.sections:
                        if config['config'][0] == section and option in config['options']:
                            if config['options'][option] not in values['values']:
                                if config['options'][option] == '' and 'none' in values['values']:
                                    # An empty string is the same as 'none' as far as UCI is concerned
                                    pass
                                else:
                                    print "found a bad entry with %s" % option
                                    invalid_entries.append(
                                    _('"%s", %s: "%s" is invalid.  Must be one of: %s') %
                                        (" ".join(config['config']), option, config['options'][option],
                                        ", ".join(values['values']))
                                    )
                # Check list items (dependencies)
                if 'list_items' in values and values['list_items']:
                    #print "There's list items: %s" % values['list_items']
                    requirements = []
                    dependency = ""
                    for list_item in values['list_items']:
                        # I'm using an ordered dictionary here since the order of the key=value pairs
                        # matters.  Besides, ordered dictionaries are cool!  I can't wait for Python 3.1
                        requirements_dict = odict()
                        #print "list_item: %s " % list_item
                        if list_item[0] == 'depends':
                            dependency = list_item[1]
                            option_requirements = dependency.split(',')
                            #print option_requirements
                            for option_requirement in option_requirements:
                                #print "option_requirement: %s" % option_requirement
                                if '=' in option_requirement:
                                    requirement_option = option_requirement.split('=')[0]
                                    requirement = option_requirement.split('=')[1]
                                    requirements_dict.update({requirement_option: requirement})
                                else:
                                    requirements_dict.update({option_requirement: ''})
                                #if requirement_option not in requirements:
                                    #requirements.update({requirement_option: []})
                                #requirements[requirement_option].append(requirement)
                            requirements.append(requirements_dict)
                    for config in self.sections:
                        success = False
                        set_items = []
                        missing_items = []
                        if option in config['options']:
                            for requirements_dict in requirements:
                                #print requirements_dict.byindex(0)
                                requirements_count = len(requirements_dict.keys())
                                match_count = 0
                                relevant_option = requirements_dict.byindex(0)[0]
                                relevant_value = requirements_dict.byindex(0)[1]
                                if relevant_option in config['options'] and config['options'][relevant_option] == relevant_value:
                                    # Check to see if we meet all the requirements of this combo
                                    for relevant_option, relevant_value in requirements_dict.items():
                                        if relevant_option == requirements_dict.byindex(0)[0]:
                                            match_count += 1 # We already matched this one
                                            set_items.append('%s=%s' % (relevant_option, relevant_value))
                                        else:
                                            if relevant_option in config['options'] and config['options'][relevant_option] == relevant_value:
                                                match_count += 1
                                                set_items.append('%s=%s' % (relevant_option, relevant_value))
                                            elif relevant_option in config['options'] and relevant_value == '':
                                                # This means that the value can be anything
                                                match_count += 1
                                                set_items.append('%s=%s' % (relevant_option, config['options'][relevant_option]))
                                            else:
                                                missing_items.append("%s=%s" % (relevant_option, relevant_value))
                                    if match_count == requirements_count:
                                        # We matched at least one full required structure for this option
                                        success = True
                        if missing_items:
                            invalid_entries.append(
                            _('"%s": If you set "%s" you must also set "%s"') %
                            (" ".join(config['config']), " ".join(set_items), " ".join(missing_items))
                            )
            # Now check whole section list items (dependencies)
            if validation_dict[section]['list_items']:
                #print "validation_dict[section]['list_items']: %s" % validation_dict[section]['list_items']
                requirement_dict = {}
                dependency = ""
                for list_item in validation_dict[section]['list_items']:
                    #print "list_item: %s" % list_item
                    if list_item[0] == 'depends':
                        dependency = list_item[1]
                        #print "dependency: %s" % dependency
                        if '=' in dependency:
                            option_requirement = dependency.split('=')[0]
                            requirement = dependency.split('=')[1]
                            requirement_struct = requirement.split(',')
                            requirement_struct = [a.strip() for a in requirement_struct]
                            requirement_option = requirement_struct[0]
                            requirement_values = requirement_struct[1:]
                            #print "requirement_struct: %s" % requirement_struct
                            #print "requirement_option: %s" % requirement_option
                            #print "requirement_values: %s" % requirement_values
                            # requirement_dict['proto']['static'] = [['ipaddr', 'netmask'], ['ip6addr']]
                            if option_requirement in requirement_dict:
                                if requirement_option in requirement_dict[option_requirement]:
                                    requirement_dict[option_requirement][requirement_option].append(requirement_values)
                                else:
                                    requirement_dict[option_requirement][requirement_option] = [requirement_values]
                            else:
                                requirement_dict[option_requirement] = {}
                                requirement_dict[option_requirement][requirement_option] = [requirement_values]
                        else: # Just add it to the dict as an empty string (we'll use the key)
                            requirement_dict.update({dependency: ''})
                for config in self.sections:
                    if config['config'][0] == section:
                        for required_option, required_values in requirement_dict.items():
                            if required_option in config['options']:
                                matched_structures = 0
                                # We've got to validate this option's required structure
                                for required_value, value_structure in required_values.items():
                                    if config['options'][required_option] == required_value:
                                        structure_text = ''
                                        # Make sure that the structure config['options'] matches one of required_value's valid structures
                                        for structure in value_structure:
                                            # structure could be something like ['ipaddr', 'netmask'] or just ['ip6addr']
                                            struct_len = len(structure)
                                            count = 0
                                            missing_items = []
                                            set_items = []
                                            for item in structure:
                                                structure_text += '%s and ' % item
                                                if item in config['options']:
                                                    count += 1
                                                    set_items.append(item)
                                                else:
                                                    missing_items.append(item)
                                            structure_text = structure_text.rstrip('and ')
                                            structure_text += ' or '
                                            if count == struct_len:
                                                matched_structures += 1
                                            elif count:
                                                # We matched only a part of a structural element
                                                # This means we should invalidate based on the assumption that only complete structural
                                                # elements are acceptable.  For example, if you have 'proto=static' and 'ipaddr' with 'ip6addr'
                                                # but not 'netmask' you'd pass the 'ip6addr' check but not this check (which would get caught
                                                # since we have 'ipaddr' but no 'netmask')

                                                invalid_entries.append(
                                                _('"%s": You cannot set %s without %s') %
                                                (" ".join(config['config']), set_items, missing_items)
                                                )
                                        if matched_structures > 0:
                                            # At least one required structure was matched: Valid
                                            pass
                                        else:
                                            # No structures matched: Invalid
                                            structure_text = structure_text.rstrip().rstrip('or')
                                            invalid_entries.append(
                                            _('"%s": If "%s" is set to "%s" you need at least one of the following options or sets of options:\n%s') %
                                            (" ".join(config['config']), required_option, required_value, structure_text)
                                            )
        if invalid_entries: # Note: convert newlines to <br /> if outputting to HTML
            raise ValidationException("Encountered validation errors: \n    %s" % "\n    ".join(invalid_entries))

    def valid_options(self, section, option, validation_file=None):
        """Return a list of valid values for any given option contained within 'section'
        of 'validation_file'.  If no validation_file is provided, this method will look
        for a matching file in <PyCI dir>/metadata/"""
        validation_obj = self.validation_obj(validation_file)
        validation_dict = self.validation_dict(validation_obj)
        option_dict = validation_dict[section][option]
        valid_options = []
        valueof_option_sub = None
        if 'values' in option_dict:
            # This is a list of all possible values for this option
            return option_dict['values']
        elif 'datatype' in option_dict and option_dict['datatype'] == 'boolean':
            return ['0', '1']
        elif 'valueof' in option_dict:
            config = option_dict['valueof'].split('.')[0]
            valueof_option = option_dict['valueof'].split('.')[1]
            # Sometimes valueof points to a specific option and not just the second field of a config...
            if len(option_dict['valueof'].split('.')) == 3:
                valueof_option_sub = option_dict['valueof'].split('.')[2]
                #print "config: %s" % config
                #print "valueof_option: %s" % valueof_option
                #print "valueof_option_sub: %s" % valueof_option_sub
            config_obj = self.read_config_file("/etc/config/%s" % config)
            for section in config_obj:
                if section['config'][0] == valueof_option:
                    if valueof_option_sub:
                        valid_options.append(section['options'][valueof_option_sub])
                    else:
                        valid_options.append(section['config'][1])
            return valid_options
        # Sometimes valueof shows up as a list item:
        elif option_dict['list_items']:
            for list_item in option_dict['list_items']:
                if 'valueof' in list_item:
                    config = list_item[1].split('.')[0]
                    valueof_option = list_item[1].split('.')[1]
                    config_obj = self.read_config_file("/etc/config/%s" % config)
                    for section in config_obj:
                        if section['config'][0] == valueof_option:
                            valid_options.append(section['config'][1])
            return valid_options

    def validate_option(self, section, option, value, validation_file=None):
        """Checks if the given section[option] is valid based on the rules specified
        in validation_file.  If no validation_file is given it will use the default
        (<PyCI dir>/metadata/<config file name>).
        Returns None if the option is valid.
        Returns a detailed error message if invalid."""
        import ipaddr
        mac_regex = re.compile('([a-fA-F0-9]{2}:?){6}$')
        validation_obj = self.validation_obj(validation_file)
        validation_dict = self.validation_dict(validation_obj)
        option_dict = validation_dict[section][option]
        valid_options = []
        if 'values' in option_dict:
            # This is a list of all possible values for this option
            if value not in option_dict['values']:
                return _("'%s' is not a valid value for %s.  Must be one of: %s") % (
                value, option, option_dict['values'])
        elif 'datatype' in option_dict and option_dict['datatype'] == 'boolean':
            if value not in ['0', '1', 'on', 'off', 'enabled', 'disabled']:
                return _("%s must be 1 (true) or 0 (false)") % option
        elif 'datatype' in option_dict and option_dict['datatype'] == 'ip4addr':
            try:
                ipaddr.IPv4(value)
            except ipaddr.IPv4IpValidationError:
                return _("'%s' is not a valid IPv4 address") % value
            except ValueError:
                return _("'%s' is not a valid IPv4 address") % value
        elif 'datatype' in option_dict and option_dict['datatype'] == 'ip6addr':
            try:
                ipaddr.IPv6(value)
            except ipaddr.IPv6IpValidationError:
                return _("'%s' is not a valid IPv6 address") % value
            except ValueError:
                return _("'%s' is not a valid IPv6 address") % value
        elif 'datatype' in option_dict and option_dict['datatype'] == 'macaddr':
            if not mac_regex.match(value):
                return _("'%s' is not a valid MAC address") % value
        elif 'datatype' in option_dict and option_dict['datatype'] == 'file':
            if not os.path.exists(value):
                return _("'%s' does not exist") % value
        elif 'datatype' in option_dict and option_dict['datatype'] == 'device':
            # 'device' is just the path to a file in /dev
            if not os.path.exists(value):
                return _("Device '%s' does not exist") % value
        elif 'datatype' in option_dict and option_dict['datatype'] == 'directory':
            if not os.path.isdir(value):
                return _("'%s' does not exist") % value
        elif 'datatype' in option_dict and option_dict['datatype'] == 'uint':
            try:
                int(value)
            except ValueError: # Couldn't turn value into an int
                return _("'%s' must be an integer") % option
        elif 'datatype' in option_dict and option_dict['datatype'] == 'float':
            try:
                float(value)
            except ValueError: # Couldn't turn value into an int
                return _("'%s' must be a floating point value") % option
        elif 'datatype' in option_dict and option_dict['datatype'] == 'integer':
            try:
                int(value)
            except ValueError: # Couldn't turn value into an int
                return _("'%s' must be an integer") % option
        elif 'datatype' in option_dict and option_dict['datatype'] == 'host':
            if not re.match(
            "^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z]|[A-Za-z][A-Za-z0-9\-]*[A-Za-z0-9])$",
                            value):
                return _("'%s' is not a valid hostname") % value
        elif 'datatype' in option_dict and option_dict['datatype'] == 'hostname':
            if not re.match(
            "^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z]|[A-Za-z][A-Za-z0-9\-]*[A-Za-z0-9])$",
                            value):
                return _("'%s' is not a valid hostname") % value
        elif 'datatype' in option_dict and option_dict['datatype'] == 'port':
            try:
                if int(value) > 65535 or int(value) < 0:
                    return _('"%s" is not a valid port number (must be 0-65535)') % value
            except ValueError: # Couldn't turn value into an int
                return _('"%s" is not a valid port number (must be 0-65535)') % value
        elif 'datatype' in option_dict and option_dict['datatype'] == 'portrange':
            try:
                if '-' in value: # A port range as opposed to just a portrange
                    for val in value.split('-'):
                        if int(val) > 65535 or int(val) < 0:
                            return _('"%s" is not a valid port range (must be "[0-65535]-[0-65535]")') % val
                else:
                    if int(value) > 65535 or int(value) < 0:
                        return _('"%s" is not a valid port number (must be 0-65535)') % value
            except ValueError: # Couldn't turn value into an int
                return _('"%s" is not a valid port number (must be 0-65535)') % value
        elif 'valueof' in option_dict and option_dict['valueof'] == 'network.interface':
            for section in self.sections:
                if section['config'][0] == 'interface':
                    valid_options.append(section['config'][1])
            if value not in valid_options:
                return _("'%s' is not a valid value for %s.  Must be one of: %s") % (
                value, option, valid_options)

    def option_datatype(self, section, option, validation_file=None):
        """Returns the datatype value for a given option in a given section.
        Will be one of:
        uint, string, host, device, boolean, file, ip4addr, ip6addr, port,
        directory, integer, macaddr, hostname"""
        validation_obj = self.validation_obj(validation_file)
        validation_dict = self.validation_dict(validation_obj)
        option_dict = validation_dict[section][option]
        if 'datatype' in option_dict:
            # This is a list of all possible values for this option
            return option_dict['datatype']

# The following three functions are specific to /etc/config/firewall
def fw_optional(option_dict, option):
    """Returns the value of 'option' in option_dict if 'option' is actually present.
    Returns None if 'option' isn't in the dict."""
    try:
        return option_dict[option]
    except:
        return None

def new_fw_rule(**kwargs):
    """Returns a list of one or two Rule() objects from the given keyword arguments.
    If proto=='tcp' or proto=='udp', return one Rule().
    if proto=='tcpudp', return two Rule() objects (one for each of 'tcp' and 'udp').
    If proto==None, return one Rule() based on the given keyword arguments."""
    from iptables import Rule
    try:
        if kwargs['protocol'] == 'tcpudp':
            tcp_rule = Rule(**kwargs)
            udp_rule = Rule(**kwargs)
            tcp_rule.protocol = 'tcp'
            udp_rule.protocol = 'udp'
            return [tcp_rule, udp_rule]
        else:
            return [Rule(**kwargs)]
    except Exception as e:
        print e
        return [Rule(**kwargs)]

def fw_rule(firewall_conf_rule, rules_output=False):
    """This is a python port of the fw_rule function inside OpenWRT's uci_firewall.sh script.
    The purpose is to output an equivalent iptables command line from a standard UCI firewall
    config section.  It takes a single 'rule' section from a ConfigFileOpenWRT object,
    'firewall_conf_rule' as the only argument and sources it to return a list of one or two
    (if proto=='tcpudp') Rule() objects which can be used to generate the iptables lines.
    if 'rules_output' is True, return Rule() objects instead of just strings."""

    # This is a smaller version of the mako template in iptables.py
    # I made it smaller because it was taking WAY too long to load up the Rules tab
    rule_template = """
<%!
def s(text):
    return repr(unicode(text).encode('utf-8'))

def optional(switch, value):
    if value is not None:
        if value == True:
            return switch
        else:
            return '%s %s' % (switch, s(value))
    return ''
%>
% for rule in rules:
iptables -t filter \\
-I ${rule.chain|s} \\
% if isinstance(match, list):
 % for item in rule.match:
${optional('-m', item)} \\
 % endfor
% else:
${optional('-m', rule.match)} \\
% endif
${optional('-s', rule.source)} \\
${optional('-d', rule.destination)} \\
${optional('-p', rule.protocol)} \\
${optional('-sport', rule.source_port)} \\
${optional('-dport', rule.destination_port)} \\
${optional('--icmp-type', rule.icmp_type)} \\
${optional('--mac-source', rule.mac_source)} \\
-j ${rule.target|s}
%endfor"""
    chain = fw_optional(firewall_conf_rule, 'src')
    source = fw_optional(firewall_conf_rule, 'src_ip')
    mac_source = fw_optional(firewall_conf_rule, 'src_mac')
    if mac_source: # uci_firewall.sh auto-sets match like this:
        match = 'mac'
    else:
        match = None
    source_port = fw_optional(firewall_conf_rule, 'src_port')
    if source_port and '-' in source_port: # UCI configs use a dash (-) but iptables uses a colon (:)
        source_port = source_port.replace('-', ':')
    dest = fw_optional(firewall_conf_rule, 'dest') # zone_$src or zone_${src}_forward or zone_${dest}_$target
    destination = fw_optional(firewall_conf_rule, 'dest_ip')
    destination_port = fw_optional(firewall_conf_rule, 'src_dport')
    if destination_port and '-' in destination_port:
        destination_port = destination_port.replace('-', ':')
    protocol = fw_optional(firewall_conf_rule, 'proto')
    icmp_type = fw_optional(firewall_conf_rule, 'icmp_type')
    target = fw_optional(firewall_conf_rule, 'target')
    if not target:
        target = 'DROP'
    # Now we need to handle the target chain exactly the way UCI does...
    # If 'src' is set but not 'dest':
    if chain and not dest:
        # chain is zone_$src
        chain = 'zone_%s' % chain
    elif chain and dest:
        # chain is zone_${src}_forward
        chain = 'zone_%s_forward' % chain
    elif dest and not chain:
        # chain is zone_${dest}_$target
        chain = 'zone_%s_%s' % (dest, target)
    rules = new_fw_rule(
        table='filter', # These are all filter rules
        chain=chain,
        source=source,
        match=match,
        mac_source=mac_source,
        source_port=source_port,
        destination=destination,
        destination_port=destination_port,
        protocol=protocol,
        icmp_type=icmp_type,
        target=target
    )
    out_list = []
    if not rules_output:
        for rule in rules:
            out_list.append(rule.iptables_line(rule_template))
        return out_list
    else:
        for rule in rules:
            out_list.append(rule)
        return out_list

def fw_redirect(firewall_conf_rule, rules_output=False):
    """This is a python port of the fw_redirect function inside OpenWRT's uci_firewall.sh script.
    The purpose is to output an equivalent iptables command line from a standard UCI firewall
    config section.  It takes a single 'rule' section from a ConfigFileOpenWRT object,
    'firewall_conf_rule' as the only argument and sources it to return a list of one or two
    (if proto=='tcpudp') Rule() objects which can be used to generate the iptables lines.
    if 'rules_output' is True, return Rule() objects instead of just strings."""

    # This is a smaller version of the mako template in iptables.py just like in fw_rule()
    # but this version is specific to port forwarding.
    rule_template = """
<%!
def s(text):
    return repr(unicode(text).encode('utf-8'))

def optional(switch, value):
    if value is not None:
        if value == True:
            return switch
        else:
            return '%s %s' % (switch, s(value))
    return ''
%>
% for rule in rules:
iptables -t filter \\
% if 'prerouting' in rule.chain:
-A ${rule.chain|s} \\
% else:
-I ${rule.chain|s} \\
% endif
% if isinstance(match, list):
 % for item in rule.match:
${optional('-m', item)} \\
 % endfor
% else:
${optional('-m', rule.match)} \\
% endif
${optional('-s', rule.source)} \\
${optional('-d', rule.destination)} \\
${optional('-p', rule.protocol)} \\
${optional('-sport', rule.source_port)} \\
${optional('-dport', rule.destination_port)} \\
${optional('--icmp-type', rule.icmp_type)} \\
${optional('--mac-source', rule.mac_source)} \\
-j ${rule.target|s} \\
${optional('--to-destination', rule.to_destination)} \\
%endfor"""
    chain = fw_optional(firewall_conf_rule, 'src')
    source = fw_optional(firewall_conf_rule, 'src_ip')
    mac_source = fw_optional(firewall_conf_rule, 'src_mac')
    if mac_source: # uci_firewall.sh auto-sets match like this:
        match = 'mac'
    else:
        match = None
    source_port = fw_optional(firewall_conf_rule, 'src_port')
    if source_port and '-' in source_port: # UCI configs use a dash (-) but iptables uses a colon (:)
        source_port = source_port.replace('-', ':')
    dest = fw_optional(firewall_conf_rule, 'dest') # zone_$src or zone_${src}_forward or zone_${dest}_$target
    destination = fw_optional(firewall_conf_rule, 'dest_ip')
    destination_port = fw_optional(firewall_conf_rule, 'src_dport')
    if destination_port and '-' in destination_port:
        destination_port = destination_port.replace('-', ':')
    to_destination = "%s:%s" % (fw_optional(firewall_conf_rule, 'dest_ip'),fw_optional(firewall_conf_rule, 'dest_port'))
    protocol = fw_optional(firewall_conf_rule, 'proto')
    icmp_type = fw_optional(firewall_conf_rule, 'icmp_type')
    target = fw_optional(firewall_conf_rule, 'target')
    # Now we need to handle the target chain exactly the way UCI does...
    # If 'src' is set but not 'dest':
    filter_chain = "zone_%s_forwarding" % chain
    nat_chain = "zone_%s_prerouting" % chain
    # fw_redirect() makes two rules:
    #   One in the filter table to ACCEPT
    #   One in nat to do the actual port forwarding
    filter_rules = new_fw_rule(
        table='filter',
        chain=filter_chain,
        source=source,
        match=match,
        mac_source=mac_source,
        source_port=source_port,
        destination=destination,
        destination_port=destination_port,
        protocol=protocol,
        icmp_type=icmp_type,
        target='ACCEPT'
    )
    nat_rules = new_fw_rule(
        table='nat',
        chain=nat_chain,
        source=source,
        match=match,
        mac_source=mac_source,
        source_port=source_port,
        to_destination=to_destination,
        destination_port=destination_port,
        protocol=protocol,
        icmp_type=icmp_type,
        target='DNAT'
    )
    if not rules_output:
        out_list = []
        for rule in nat_rules:
            out_list.append(rule.iptables_line(rule_template))
        for rule in filter_rules:
            out_list.append(rule.iptables_line(rule_template))
        return out_list
    else:
        out_list = []
        for rule in nat_rules:
            out_list.append(rule)
        for rule in filter_rules:
            out_list.append(rule)
        return out_list