# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

# Built-in modules
import sys, os, re, gettext

# 3rd party modules
import cherrypy

# PyCI modules and methods
from auth import check_permission

# Internationalization
locale_dir = os.path.join(os.path.dirname(__file__), 'i18n')
try:
    trans = gettext.translation('PyCI', locale_dir)
except:
    trans = gettext.NullTranslations()
_ = lambda s: trans.ugettext(s)

"""This module contains classes and methods required and used by all PyCI plugins."""

def detect_mime_type(filepath):
    """Attempts to detect the mime type of the given file and sets the appropriate cherrypy response header"""
    f, extension = os.path.splitext(filepath)
    if extension == '.css': # It seems CherryPy can't automatically figure out what mime types are inside tarballs
        cherrypy.response.headers['Content-Type'] = "text/css"
    elif extension == '.js':
        cherrypy.response.headers['Content-Type'] = "text/javascript"
    elif extension == '.png':
        cherrypy.response.headers['Content-Type'] = "image/png"
    elif extension == '.gif':
        cherrypy.response.headers['Content-Type'] = "image/gif"
    elif extension == '.jpeg' or extension == '.jpeg':
        cherrypy.response.headers['Content-Type'] = "image/jpeg"

class Plugin(object):
    """
    The base class from which all plugins are derived.  Contains functions
    that are both required and just plain useful when writing plugins.

    Note: If you're going to override the __init__() function of Plugin make sure
    to include 'super(YourPlugin, self).__init__()' in your plugin's __init__().
    Otherwise your plugin will likely have problems with internationalization.
    """

    _cp_config = {
        'tools.auth.on': True
    }

    sort_priority = 1
    plugin_dict = {
        'help': 'No help',
        'long_name': 'No name',
        'description': 'No description',
    }
    navigable_functions = {}

    def __init__(self):
        """Initializes _() for internationalization.  If you override this method make sure to include this code"""
        locale_dir = os.path.join(os.getcwd(), 'plugins_enabled/%s/i18n' % self.__module__)
        try:
            trans = gettext.translation(self.__module__, locale_dir)
        except:
            trans = gettext.NullTranslations()
        self._ = lambda s: trans.ugettext(s)

    def standard_render(self, template, **kwargs):
        """Renders mako 'template' with the standard variables"""
        from mako import exceptions
        plugin_dict = self.plugin_dict
        plugin_dict.update(**kwargs)
        try:
            language = os.environ['LANG']
        except KeyError: # LANG isn't set
            language = u'en_US' # Fall back/default to US English
        try:
            return template.render_unicode(plugin_name=self.__class__.__name__, plugin_object=Plugin, plugin_dict=plugin_dict, language=language, gettext_domain=self.__module__, _=self._)
        except:
            return exceptions.html_error_template().render()

    def _static(self, filepath, **kwargs):
        """Return static content contained within the calling plugin's
        directory or zip file as specified by 'filepath'.

        To use this function you can use URLs like so:
        <script type="text/javascript" src="/Admin/_static?filepath=static/js/admin.js"></script>

        Note:  This function will work with zipped plugins."""
        local_filepath = "plugins_enabled/%s/%s" % (self.__module__, filepath)
        if os.path.exists(local_filepath):
            detect_mime_type(local_filepath)
            return open(local_filepath, 'r').read()
        else: # Lets check inside the plugin's zip file
            try: # This uses the zipimport __loader__ built-in to try to extract the file from the plugin's zip archive
                detect_mime_type(filepath)
                return sys.modules[self.__module__].__loader__.get_data(filepath)
            except:
                raise cherrypy.HTTPError(404)

    def debug_helper(self, *args, **kwargs):
        """Return a page with all the arguments we were passed (for debugging)"""
        html = ""
        for arg in args:
            html += _("You submitted the following arguments:<br />\n")
            html += "%s<br />\n" % arg
        for key, value in kwargs.iteritems():
            html += _("You submitted the following keyword arguments:<br />\n")
            html += "%s: %s<br />\n" % (key, value)
        return html

    def _files(self):
        """Return a list of files inside the plugin and whether or not they're inside a zip archive."""
        files_dict = {}
        files = []
        try: # If we're inside a zip our static files will be inside
            for filepath in sys.modules[self.__module__].__loader__._files.itervalues():
                files.append(filepath[0])
            files_dict.update({'zipped': True, 'files': files})
        except:
            os.chdir('plugins_enabled')
            for root, dirs, _files_ in os.walk(self.__module__):
                if len(_files_) > 0:
                    for filename in _files_:
                        filepath = os.path.join(root, filename)
                        files.append(filepath)
            files_dict.update({'zipped': True, 'files': files})
            os.chdir('..')
        return files_dict

    #def locale(self, string):
        #"""Renders a javascript page containing an i18n object representing all translatable strings for this plugin."""
        #plugin_dict = self.plugin_dict
        #plugin_dict.update(**kwargs)
        #i18n = _(string)
        #return template.render(plugin_name=self.__class__.__name__, plugin_object=Plugin, plugin_dict=plugin_dict, i18n=i18n)

def get_translation(plugin_file):
    """Returns a gettext translation object with the specified domain and locale_dir.
    @plugin_file: The calling plugin's __file__ variable."""
    # The rstrip('.zip') is necessary because __file__ returns <path to>/plugin_name.zip when the plugin is zipimported
    plugin_path = os.path.dirname(os.path.abspath(plugin_file)).rstrip('.zip')
    #print "plugin_path: %s" % plugin_path
    plugin_name = plugin_path.split('/').pop()
    locale_dir = os.path.join(plugin_path, 'i18n')
    #print "get_translation() locale_dir: %s" % locale_dir
    if os.path.exists(locale_dir):
        #print "translation locale path exists for %s" % plugin_name
        try:
            trans = gettext.translation(plugin_name, locale_dir)
            return trans
        except:
            #print "Failed to get our translation!"
            trans = gettext.NullTranslations()
            return trans
    else: # Fall back
        #print "Had to fallback!"
        locale_dir = os.path.join(os.getcwd(), 'i18n')
        try:
            trans = gettext.translation(domain='PyCI', localedir=locale_dir)
            return trans
        except: # Further fall back
            trans = gettext.NullTranslations()
            return trans

def get_plugins(plugin_dir):
    """Adds plugins to sys.path and returns them as a list"""

    #registered_plugins
    registered_plugins = []

    #check to see if a plugins directory exists and add any found plugins (even if they're zipped)
    if os.path.exists(plugin_dir):
        plugins = os.listdir(plugin_dir)
        pattern = ".py$"
        for plugin in plugins:
            plugin_path = os.path.join(plugin_dir, plugin)
            if os.path.splitext(plugin)[1] == ".zip":
                sys.path.append(plugin_path)
                (plugin, ext) = os.path.splitext(plugin) # Get rid of the .zip extension
                registered_plugins.append(plugin)
            elif plugin != "__init__.py":
                if re.search(pattern, plugin):
                    (shortname, ext) = os.path.splitext(plugin)
                    registered_plugins.append(shortname)
            if os.path.isdir(plugin_path):
                plugins = os.listdir(plugin_path)
                for plugin in plugins:
                    if plugin != "__init__.py":
                        if re.search(pattern, plugin):
                            (shortname, ext) = os.path.splitext(plugin)
                            sys.path.append(plugin_path)
                            registered_plugins.append(shortname)
    return registered_plugins

def get_display_plugins():
    """Returns a list of plugins that have the display variable set"""
    result = []
    for plugin in Plugin.__subclasses__():
        if plugin.display:
            result.append(plugin)
    return result

def load_plugins(plugins):
    """Imports all plugins given a list.
    Note:  Assumes they're all in sys.path."""
    for plugin in plugins:
        __import__(plugin, None, None, [''])
        if plugin not in Plugin.__subclasses__():
            __import__(plugin, None, None, [plugin]) # This takes care of importing zipped plugins

def init_plugin_system(cfg):
    """Initializes the plugin system by appending all plugins into sys.path and then using load_plugins() to import them.

    @cfg - A dictionary with two keys:
        @plugin_path - path to the plugin directory (e.g. 'plugins_enabled')
        @plugins - List of plugin names to import (e.g. ['admin', 'network', 'status']."""
    if not cfg['plugin_path'] in sys.path:
        sys.path.insert(0, cfg['plugin_path'])
    load_plugins(cfg['plugins'])

def find_plugins():
    """Returns a list containing all active plugin objects"""
    return Plugin.__subclasses__()

def get_user_plugins(user):
    """Returns a list of plugins with the display variable set that the given user has at least read access to.
    In other words, it will only lists plugins the user is supposed to see."""
    plugins = get_display_plugins()
    out_list = []
    for plugin in plugins:
        if check_permission(user, 'r', plugin.__name__):
            out_list.append(plugin)
    return out_list