# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

import os
# 3rd party modules
import cherrypy
# PyCI modules
from plugin_utils import Plugin, get_translation, find_plugins
from auth import require, has_read, has_write
from pyci_db import connect_db, disconnect_db, check_user_permission, check_group_permission, delete_user_permission_by_id, delete_group_permission_by_id, change_user_permission, change_group_permission
from utils import lookup_template, jqgrid_select_options, plugin_template, jquery_ui_error_msg, jquery_ui_highlight_msg, jqgrid_json
from passwd import list_etc_passwd
from group import list_etc_group
from odict import odict
from cherrypy_config import CherryPyConfig

"""Admin Interface Plugin (for configuring things like user/group permissions, themes, etc)"""

# Take care of internationalization
trans = get_translation(__file__)
_ = lambda s: trans.ugettext(s)

def get_language_setting():
    """Returns the language setting in the PyCI DB (if there is one)"""
    pyci_db = connect_db('pyci.db')
    dbcursor = pyci_db.cursor()
    try: # This will raise an exception if there's no record
        record = dbcursor.execute('SELECT setting FROM pyci_settings WHERE option="language"').fetchone()[0]
        disconnect_db(pyci_db)
        return record
    except:
        return None

def set_language_preference(language):
    """Sets a user's language preference."""
    # TODO:  Change things like this to be logged to syslog as well (or instead)
    #print _("Setting language to: %s") % language
    pyci_db = connect_db('pyci.db')
    dbcursor = pyci_db.cursor()
    if not language: # if it is None, remove the language preference altogether
        record = dbcursor.execute('DELETE FROM pyci_settings WHERE option="language"')
        try:
            del(os.environ['LANG']) # This unsets the LANG variable for the current running CherryPy process/environment
        except: # If it is already unset we don't need to raise an exception
            pass
        return True
    record = dbcursor.execute('SELECT setting FROM pyci_settings WHERE option="language"').fetchone()
    if record:
        dbcursor.execute("UPDATE pyci_settings SET setting=(?) WHERE option='language'", (language,))
        os.environ['LANG'] = language
    else:
        dbcursor.execute("INSERT INTO pyci_settings VALUES (null, 'language', ?)", (language,))
        os.environ['LANG'] = language
    disconnect_db(pyci_db)
    return True

class PyCISettings(Plugin):
    """A plugin for managing PyCI's configuration."""
    display = True
    display_name = _("PyCI Settings")
    sort_priority = 0
    navigable_functions = odict([
        (_('Language'), 'Language'),
        (_('Manage Plugins'), 'ManagePlugins'),
        (_('Plugin Permissions'), 'Permissions'),
        (_('Web Server Settings'), 'WebServerSettings'),
    ])

    @require(has_read((('PyCISettings'))))
    def Language(self):
        """Render the preferences configuration page"""
        language_template = plugin_template(self, "templates/language.mako")
        valid_languages = []
        for item in os.listdir('./i18n'):
            if os.path.isdir('i18n/%s' % item) and not item.startswith("."):
                valid_languages.append(item)
        return self.standard_render(
            template=language_template,
            valid_languages=valid_languages,
            preferred_language=get_language_setting(),
            help=_("This form allows you to permanently set the language used by PyCI.  Only languages that are present in PyCI's i18n directory will be listed (even though plugin authors may have more language options available)."),
            long_name=_('PyCI Administration: Preferred Language'),
            description=_("Select your preferred language so that PyCI may use it if available.  Only available languages will be listed.<br /><strong>Note:</strong> It may take a few seconds for this form to respond."),
        )

    @require(has_write((('PyCISettings'))))
    def set_language(self, language):
        """Saves the user's language preference in the PyCI DB and changes the current LANG environment variable"""
        try:
            set_language_preference(language)
            yield jquery_ui_highlight_msg(_('<strong>Success:</strong> Language setting has been updated.'))
            cherrypy.engine.restart()
        except Exception as e:
            print e
            yield jquery_ui_error_msg(_('Could not save your language setting.  Please make sure PyCI is running as root.<br />%s') % e)

    @require(has_read((('PyCISettings'))))
    def Permissions(self):
        """Render the user permissions configuration page"""
        permissions_template = plugin_template(self, "templates/permissions.mako")
        group_names = []
        groups_list = list_etc_group()
        for group in groups_list: # This builds a list of group names
            group_names.append(group[0])
        jqggroups = jqgrid_select_options(group_names)
        user_names = []
        etc_passwd = list_etc_passwd()
        for user in etc_passwd: # This builds a list of user names
            user_names.append(user[0])
        jqgusers = jqgrid_select_options(user_names)
        plugin_list = ["all",]
        plugins = find_plugins()
        for plugin in plugins:
            plugin_list.append(plugin.__name__)
        jqgplugins = jqgrid_select_options(plugin_list)
        return self.standard_render(
            template=permissions_template,
            help=_('<p>The permissions interface allows you to control which users and groups are allowed to access PyCI as well as which plugins they have access to and whether or not they can make changes using those plugins.</p><p>Users and groups can be granted access to individual plugins or "all" plugins.  To create a new permission just click on the plus (+) sign in the lower-left corner of the grid.  "r" grants read-only permission while "rw" grants read-write permission.</p><p>In the event that you lock yourself out of PyCI you can reset the permissions to allow the root user read/write access to all plugins:  Just remove or rename the pyci.db file in the PyCI directory and restart PyCI.</p>'),
            long_name=_('PyCI Administration: Permissions'),
            description=_("This plugin lets you configure user and group permissions within PyCI (who has access to what)."),
            groups=group_names,
            users=user_names,
            plugins=plugin_list,
            jqggroups=jqggroups,
            jqgusers=jqgusers,
            jqgplugins=jqgplugins
        )

    @require(has_read((('PyCISettings'))))
    def WebServerSettings(self):
        """Render a page where the user can configure the PyCI webserver."""
        _template = plugin_template(self, "templates/pyci_settings.mako")

        return self.standard_render(
            template=_template,
            help=_("""<p>This interface allows you to configure the PyCI web server.</p>
            <ul>
            <li><strong>Listen Address:</strong> The IP address PyCI will listen for HTTP connections.  Use '0.0.0.0' to listen on all IP addresses.</li>
            <li><strong>Port:</strong> The web server port PyCI will listen for HTTP connections.</li>
            <li><strong>Number of Threads:</strong> The amount of threads (integer) PyCI will use to serve web pages.  Higher values will deliver web pages faster.  Lower values will conserve memory.</li>
            <li><strong>Session Timeout:</strong> The length of time (minutes) to wait before an inactive user is automatically logged out of PyCI.</li>
            <li><strong>Log to Screen:</strong> This setting controls whether or not PyCI will output the access and error logs to stdout (on the console--useful for debugging).</li>
            <li><strong>Access Log:</strong> The path where PyCI will save the access log.</li>
            <li><strong>Error Log:</strong> The path where PyCI will save the error log.  Note: Besides errors, this log also contains authentication information (e.g. 'User X authenticated successfully')</li>
            <li><strong>SSL Certificate:</strong> The path to the SSL certificate PyCI should use for HTTPS connections (e.g. '/usr/PyCI/pyci.crt').</li>
            <li><strong>SSL Private Key:</strong> The path to the private SSL key PyCI should use for HTTPS connections (e.g. '/usr/PyCI/pyci.key').</li>
            </ul>"""),
            long_name=_('PyCI Administration: PyCI Settings'),
            description=_("This interface allows you to configure the PyCI webserver."),
        )

    @require(has_read((('PyCISettings'))))
    def ManagePlugins(self):
        """Render a page where the user can enable or disable plugins."""
        _template = plugin_template(self, "templates/manage_plugins.mako")

        return self.standard_render(
            template=_template,
            help=_('<p>This interface allows you to enable or disable plugins within PyCI.  Use the context menu (right-click) to enable or disable plugins.</p> <p>Note: PyCI will need to be restarted by hand (or just reboot) after these settings are changed.</p>'),
            long_name=_('PyCI Administration: Manage Plugins'),
            description=_("This interface allows you to enable or disable plugins within PyCI.  Once you're done changing which plugins are enabled or disabled you'll need to restart PyCI in order for the changes to take effect."),
            # TODO:  Write something like the Language method(s) that automatically restarts PyCI after the change.
            # TODO:  ...or even better, get plugin_management_json doing an __import__ after plugins are enabled.
        )

    @require(has_read((('PyCISettings'))))
    def pyci_settings_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns the editable CherryPy settings from cherrypy.conf in a json format that jqGrid likes."""
        cherrypy_config = CherryPyConfig()
        header = ['Option', 'Value']
        return jqgrid_json(self, cherrypy_config.asList(), header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)

    @require(has_write((('PyCISettings'))))
    def update_pyci_settings(self, Option=None, Value=None, **kwargs):
        """Enables or disables a given plugin (meant to be used by a context menu in jqGrid)"""
        cherrypy_config = CherryPyConfig()
        if Option == _('Listen Address'):
            cherrypy_config.listen_address = Value
        elif Option == _('Port'):
            cherrypy_config.port = Value
        elif Option == _('Number of Threads'):
            cherrypy_config.num_threads = Value
        elif Option == _('Session Timeout'):
            cherrypy_config.sessions_timeout = Value
        elif Option == _('Log to Screen'):
            cherrypy_config.log_to_screen = Value
        elif Option == _('Access Log'):
            cherrypy_config.access_log = Value
        elif Option == _('Error Log'):
            cherrypy_config.error_log = Value
        elif Option == _('SSL Certificate'):
            cherrypy_config.ssl_certificate = Value
        elif Option == _('SSL Private Key'):
            cherrypy_config.ssl_private_key = Value
        cherrypy_config.save()
        return _("Success")

    @require(has_read((('PyCISettings'))))
    def plugin_management_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns all plugins in plugins_enabled and plugins_disabled in a json format that's compatible with jqGrid."""
        plugin_list = []
        plugins_enabled = os.listdir('plugins_enabled')
        for plugin in plugins_enabled:
            if not plugin.startswith('.'): # Ignore hidden directories (like .svn =)
                if plugin.endswith('.zip'):
                    pass # There will still be a plugin directory of the same name containing the i18n stuff
                else:
                    plugin_list.append([plugin, True])
        plugins_disabled = os.listdir('plugins_disabled')
        for plugin in plugins_disabled:
            if not plugin.startswith('.'):
                if plugin.endswith('.zip'):
                    pass # There will still be a plugin directory of the same name containing the i18n stuff
                else:
                    plugin_list.append([plugin, False])
        header = ['Plugin', 'Enabled']
        return jqgrid_json(self, plugin_list, header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)

    @require(has_write((('PyCISettings'))))
    def update_plugins(self, plugin=None, enable=None, disable=None, **kwargs):
        """Enables or disables a given plugin (meant to be used by a context menu in jqGrid)"""
        import shutil
        if enable:
            plugin_path = 'plugins_disabled/%s' % plugin
            destination_path = 'plugins_enabled/'
            if os.path.exists(plugin_path):
                shutil.move(plugin_path, destination_path)
            if os.path.exists(plugin_path + '.zip'): # Move the zip file too
                shutil.move(plugin_path + '.zip', destination_path)
        elif disable:
            plugin_path = 'plugins_enabled/%s' % plugin
            destination_path = 'plugins_disabled/'
            if os.path.exists(plugin_path):
                shutil.move(plugin_path, destination_path)
            if os.path.exists(plugin_path + '.zip'): # Move the zip file too
                shutil.move(plugin_path + '.zip', destination_path)
        return _("Success")

    @require(has_read(('PyCISettings')))
    def get_user_permissions_spreadsheet(self):
        """Exports the user permissions table as a spreadsheet"""
        spreadsheet_template = lookup_template('html', 'spreadsheet.mako')
        cherrypy.response.headers['Content-Type'] = 'application/vnd.ms-excel; charset=utf-8'
        headers = [_("Permission ID"), _("User"), _("Plugin"), _("Permission")]
        rows = check_user_permission()
        return spreadsheet_template.render(headers=headers, rows=rows)

    @require(has_read(('PyCISettings')))
    def get_group_permissions_spreadsheet(self):
        """Exports the user permissions table as a spreadsheet"""
        spreadsheet_template = lookup_template('html', 'spreadsheet.mako')
        cherrypy.response.headers['Content-Type'] = 'application/vnd.ms-excel; charset=utf-8'
        headers = [_("Permission ID"), _("Group"), _("Plugin"), _("Permission")]
        rows = check_group_permission()
        return spreadsheet_template.render(headers=headers, rows=rows)

    @require(has_read(('PyCISettings')))
    def get_plugins_spreadsheet(self):
        """Exports the list of enabled/disabled plugins as a spreadsheet"""
        cherrypy_config = CherryPyConfig()
        spreadsheet_template = lookup_template('html', 'spreadsheet.mako')
        cherrypy.response.headers['Content-Type'] = 'application/vnd.ms-excel; charset=utf-8'
        headers = [_('Option'), _('Value')]
        rows = cherrypy_config.asList()
        return spreadsheet_template.render(headers=headers, rows=rows)

    @require(has_write(('PyCISettings')))
    def update_user_permissions(self, oper=None, id=None, permission=None, user=None, plugin=None):
        """Updates the permissions table with the provided operation and data..."""
        if oper == "del":
            if isinstance(id, list):
                for item in id.split(','):
                    if item:
                        delete_user_permission_by_id(item)
            else:
                delete_user_permission_by_id(id)
        if oper == "add":
            change_user_permission(user=user, permission=permission, plugin=plugin)
        if oper == "edit":
            change_user_permission(user=user, permission=permission, plugin=plugin, id=id)
        return _("Success")

    @require(has_write(('PyCISettings')))
    def update_group_permissions(self, oper=None, id=None, permission=None, group=None, plugin=None):
        """Updates the permissions table with the provided operation and data..."""
        if oper == "del":
            if isinstance(id, list):
                for item in id.split(','):
                    if item:
                        delete_group_permission_by_id(item)
            else:
                delete_group_permission_by_id(id)
        if oper == "add":
            change_group_permission(group=group, permission=permission, plugin=plugin)
        if oper == "edit":
            change_group_permission(group=group, permission=permission, plugin=plugin, id=id)
        return _("Success")

    @require(has_write(('PyCISettings')))
    def change_user_permission(self, user, plugin, permission):
        """Updates the permissions table with the data from users_json.
        This is just here to expose the function in case someone wants to use it as an API"""
        change_user_permission(user, plugin, permission)

    @require(has_read(('PyCISettings')))
    def user_permissions_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns all user permissions from the permissions table in pyci.db"""
        header = ["id", "user", "plugin", "permission"]
        permissions_list = check_user_permission()
        return jqgrid_json(self, permissions_list, header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)

    @require(has_read(('PyCISettings')))
    def group_permissions_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns all group permissions from the permissions table in pyci.db"""
        header = ["id", "group", "plugin", "permission"]
        permissions_list = check_group_permission()
        return jqgrid_json(self, permissions_list, header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)

    @require(has_read((('PyCISettings'))))
    def index(self):
        """Return the index page"""
        return self.WebServerSettings()