# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

import threading, socket, os, sys, time, string, random
from optparse import OptionParser

__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'
__doc__ = """\
speedtester.py:  An application and an importable module that can be used
to test bandwidth throughput on any given TCP/IP network.  This module contains two threads:

    1. SpeedTestServer(address="0.0.0.0", port=8081, verbose=False):  Listens on 'address'
       and 'port' for incoming UDP packets (should be 1kb chunks but it probably doesn't
       matter).  Every 1000 8kb chunks it receives it writes an average bandwidth speed to
       /tmp/average_speed (in KB/sec).  If 'verbose' is set it will print the average speed
       to stdout every 1000 packets (will constantly scroll your terminal).
    2. SpeedTestClient(host, port=8081, packets="random"):  Sends 1-kilobyte UDP packets to
       'host':'port' as fast as it can.  By default packets are constructed from os.urandom
       to ensure that any enabled network compression tools don't inflate the reported speed.
       If you would like to use a repeating sequence of characters (to test compression) just
       pass the sequence as a string to the 'packets' variable.  For example, 'packets="0"'
       would have the client sending nothing but zeros to the speed test server.
       'packets="1234567890"' would send packets like so: '12345678901234567890...'
"""

class SpeedTestServer(threading.Thread):
    """
    This thread...
    """
    def __init__(self, address="0.0.0.0", port=8081, verbose=False):
        threading.Thread.__init__(self, name="SpeedTestServer")
        self.address = address
        self.port = port
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.s.bind((address, port))
        self.pkt_size = 8192
        self.quitting = False
        self.verbose = verbose
        # Zero out any existing /tmp/speed_test_speed file
        f = open('/tmp/speed_test_speed', 'w')
        f.write("0")
        f.close()

    def quit(self):
        self.quitting = True

    def run(self):
        print "SpeedTestServer now listening for a connection..."
        if time.clock() != time.clock():
            time.time = time.clock
        self.s.listen(1)
        conn, addr = self.s.accept()
        data = os.urandom(self.pkt_size) # Collect a random string to send as our packet
        conn.sendall('Start') # This kicks off the client to start receiving
        while not self.quitting:
            start = time.time()
            total_data = 0
            for i in range(0,100):
                conn.sendall(data)
                total_data += len(data)
            speed = round(total_data / (time.time() - start) / 1024 / 1024, 2)
            # Record the speed to a temporary file that other apps can read from
            f = open('/tmp/speed_test_speed', 'w')
            f.write(str(speed))
            f.close()
            if self.verbose:
                print 'Send Speed: %sMB/sec' % speed
        print "SpeedTestServer shutting down..."
        self.s.close()

class SpeedTestClient(threading.Thread):
    """
    This thread sends 8k UDP packets to 'host':'port' as fast as it can.
    If 'packets' is something other than 'random', use 'packets' to generate
    the UDP packets (will repeat the 'packets' until it reaches 1024 bytes).
    """
    def __init__(self, host="localhost", port=8081, listen_port=8082):
        threading.Thread.__init__(self, name="SpeedTestClient")
        self.host = host
        self.port = port
        self.listen_port = listen_port
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.bind(('', self.listen_port))
        self.s.connect((self.host, self.port))
        self.pkt_size = 8192
        self.quitting = False

    def quit(self):
        self.quitting = True

    def shutdown_server(self):
        """Emulates a client trying to get data to force it into the while loop so it can quit."""
        data = self.s.recv(5)
        data = self.s.recv(self.pkt_size)

    def run(self):
        if time.clock() != time.clock():
            time.time = time.clock
        print "Waiting for server..."
        data = self.s.recv(5)
        print "Got '%s' from server..." % data
        print "Starting speed test with %s..." % self.host
        while not self.quitting:
            data = ""
            start = time.time()
            total_data = 0
            while total_data < self.pkt_size * 100:
                data = self.s.recv(self.pkt_size)
                total_data += len(data)
            speed = round(total_data / (time.time() - start) / 1024 / 1024, 2)
            print 'Receive Speed: %sMB/sec' % speed

def start_speedtest_server(address, port=8081, verbose=False):
    """Starts up the SpeedTestServer thread"""
    speed_tester = SpeedTestServer(address=address, port=port, verbose=verbose)
    speed_tester.setName('SpeedTestServer')
    speed_tester.start()

def stop_speedtest_server():
    """Stops up the SpeedTestServer thread"""
    for t in threading.enumerate():
        if t.getName() == 'SpeedTestServer':
            t.quit()
    # Now try to connect to it after 2 seconds to make sure it shuts down.
    # If it didn't shut down it is likely still stuck waiting for a client to connect
    time.sleep(2)
    speed_tester = SpeedTestClient(host='127.0.0.1', port=8081)
    speed_tester.setName('SpeedTestClient')
    speed_tester.quit()
    speed_tester.run()

def start_speedtest_client(host, port=8081):
    """Starts up the SpeedTestClient thread"""
    speed_tester = SpeedTestClient(host=host, port=port)
    speed_tester.setName('SpeedTestClient')
    speed_tester.start()

def stop_speedtest_client():
    """Stops up the SpeedTestClient thread"""
    for t in threading.enumerate():
        if t.getName() == 'SpeedTestClient':
            t.quit()

def speedtest_server_status():
    """Returns True if the SpeedTestServer thread is running.  None if not."""
    for t in threading.enumerate():
        if t.getName() == 'SpeedTestServer':
            return True

def speedtest_client_status():
    """Returns True if the SpeedTestClient thread is running.  None if not."""
    for t in threading.enumerate():
        if t.getName() == 'SpeedTestClient':
            return True

def localhost_test():
    """Starts up a SpeedTestServer() thread and a SpeedTestClient() thread on
    localhost then proceeds to test the throughput between them.
    Will quit cleanly if presented with a KeyboardInterrupt.
    Only meant to be used for testing."""
    start_speedtest_server('localhost', verbose=True)
    start_speedtest_client('localhost')
    try:
        while True:
            pass
    except KeyboardInterrupt:
        stop_speedtest_client()
        stop_speedtest_server()
        sys.exit(0)

if __name__ == '__main__':
    # Grab command line arguments and the command to run (if any)
    usage = 'usage: %prog [options]'
    parser = OptionParser(usage=usage)
    #parser.disable_interspersed_args()
    parser.add_option("-c", "--client", dest="client", default=None, help="Run in client mode connecting to the given host.", metavar="<host>")
    parser.add_option("-s", "--server", dest="server", default=None, help="Run in server mode listening on the given address.  Use '0.0.0.0' to listen on all addresses.", metavar="<address>")
    (options, args) = parser.parse_args()

    # Check to make sure we were passed at least one command line argument
    try:
        sys.argv[1]
    except:
        print "\nError:  At a minimum you must supply '-c <host>' (client mode connecting to <host>) or '-s <address>' (server mode)"
        parser.print_help()
        sys.exit(2)

    try:
        if options.server: # Run in server mode
            start_speedtest_server(options.server, verbose=True)
            while True:
                time.sleep(1)
                server_running = False
                for t in threading.enumerate():
                    if t.getName() == 'SpeedTestServer':
                        server_running = True
                if not server_running:
                    print "Client disconnected.  Quitting..."
                    sys.exit(0)
        elif options.client: # Run in client mode
            start_speedtest_client(options.client)
            while True:
                time.sleep(1)

    except KeyboardInterrupt:
        if options.client:
            stop_speedtest_client()
        elif options.server:
            stop_speedtest_server()
        time.sleep(1)
        sys.exit(0)