#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Description: System plugin

import os, json, threading
import cherrypy
from plugin_utils import Plugin, get_translation
from auth import require, has_read, has_write
from utils import lookup_template, jqgrid_select_options, jqgrid_json, plugin_template, jquery_ui_error_msg, jquery_ui_highlight_msg
from passwd import PasswdFile, User, list_etc_passwd
from group import GroupFile, Group, list_etc_group
from openwrt import ConfigFileOpenWRT
from odict import odict
import eftw

# Internationalization
trans = get_translation(__file__)
_ = lambda s: trans.ugettext(s)

def valid_shells():
    """Returns a list of all shells in /etc/shells"""
    shells = open('/etc/shells').readlines()
    shell_list = []
    for line in shells:
        line = line.rstrip('\n')
        if len(line.split('#')[0]) > 2:
            shell_list.append(line.split('#')[0])
    return shell_list

def gen_unique_uid(range_start=None, range_end=None):
    """Returns a unique UID (one not already used in /etc/passwd) starting at 1000 and ending at 100,000 by default.
    Optional: Return a UID in the provided range (range_start/range_end).
    ...raises an exception if there are no free IDs in the given range."""
    class OutOfUIDsException(Exception):
        """Just a little custom exception for reporting we're out of UIDs"""
        def __init__(self, value):
            self.parameter = value
        def __str__(self):
            return repr(self.parameter)
    passwd_list = list_etc_passwd()
    uids = []
    for line in passwd_list: # Make a list of all UIDs in /etc/passwd
        uids.append(line[2])
    uids.sort()
    if range_start == None or range_end == None:
        unique_uid = gen_unique_uid(1000, 100000) # Gotta love recursion
    else:
        for i in range(range_start, range_end):
            if i not in uids:
                unique_uid = i
                break
    try:
        return unique_uid
    except NameError:
        raise OutOfUIDsException(_("Ran out of unique UIDs.  Please specify a larger range."))

def gen_unique_gid(range_start=None, range_end=None):
    """Returns a unique GID (one not already used in /etc/group) starting at 1000 and ending at 100,000 by default.
    Optional: Return a GID in the provided range (range_start/range_end).
    ...raises an exception if there are no free IDs in the given range."""
    class OutOfGIDsException(Exception):
        """Just a little custom exception for reporting we're out of GIDs"""
        def __init__(self, value):
            self.parameter = value
        def __str__(self):
            return repr(self.parameter)
    groups = GroupFile()
    gids = groups.gids()
    gids.sort()
    if range_start == None or range_end == None:
        unique_gid = gen_unique_gid(1000, 100000) # Gotta love recursion
    else:
        for i in range(range_start, range_end):
            if str(i) not in gids:
                unique_gid = str(i)
                break
    try:
        return unique_gid
    except NameError:
        raise OutOfUIDsException(_("Ran out of unique UIDs.  Please specify a larger range."))

def list_etc_hosts():
    """
    Returns /etc/hosts as a list of lists.
    Suitable for things like jqgrid and spreadsheet output
    """
    hostname_list = []
    # Assemble a list of lists of ips and hostnames...
    etc_hosts = open('/etc/hosts')
    for line in etc_hosts:
        if line:
            if line.startswith('#'): # This is a comment
                pass # Skip it
            elif len(line.split()) > 0:
                line = line.split()
                ip = line[0]
                hostnames = line[1:]
                hostname_list.append([ip, " ".join(hostnames)])
    return hostname_list

def write_etc_hosts(text):
    """Overwrites /etc/hosts with 'text' ...to save duplicated code"""
    etc_hosts = open('/etc/hosts', 'w')
    etc_hosts.write(text)
    etc_hosts.close()

def logread(date_format=None):
    """
    Parses the output of the 'logread' command and returns it as a list of
    lists in the following format:
        [
            [datetime_obj, hostname, facility, priority, process, pid, message],
        ]
    If a date_format is given, log dates will be converted to a string before
    being added to the output like so:  datetime_obj.strftime(date_format)
    """
    from commands import getoutput
    from datetime import datetime
    now = datetime.now()
    out_list = []
    for line in getoutput('logread').split('\n'):
        line = line.split()
        # Before we do anything else we need to convert the date into a Python datetime object.
        # Since the dates from logread don't include years we need to figure out if a
        # given date was from this year or last year...
        date_and_time = " ".join([line[0], line[1], line[2]])
        # We assume that if the month in the log occurs later this year that means
        # it is really a date from last year.
        thismonth = int(datetime.strftime(now, '%m'))
        datetime_obj = datetime.strptime(date_and_time, '%b %d %H:%M:%S')
        linemonth = int(datetime_obj.strftime('%m'))
        # A negative number below means this line's month has yet to occur this year
        if thismonth - linemonth < 0: # This date was from last year
            year = (now.year - datetime_obj.year) + datetime_obj.year
        else: # This date occurred this year
            year = now.year
        # Now reassemble the datetime_obj with the correct year
        date_and_time = "%s %s" % (year, date_and_time)
        datetime_obj = datetime.strptime(date_and_time, '%Y %b %d %H:%M:%S')
        # Now we can put the rest of our variables together...
        hostname = line[3]
        facility = line[4].split('.')[0]
        priority = line[4].split('.')[1]
        process = line[5].rstrip(':')
        # There can be a PID at the end of a process in brackets...
        if process.endswith(']') and not process.startswith('['):
            pid = process.split('[')[1].rstrip(']')
            process = process.split('[')[0]
        else:
            pid = None
        message = " ".join(line[6:]) # Everything after the process/pid is the message
        if date_format: # Convert to string using the given format
            datetime_obj = datetime_obj.strftime(date_format)
        out_list.append([
            datetime_obj,
            hostname,
            facility,
            priority,
            process,
            pid,
            message
        ])
    return out_list

# TODO:  Turn this into a thread (deferred) that can be cancelled
def delayed_remove_file(filepath, delay=300):
    """Waits 'delay' seconds before removing the file at 'filepath'.
    Default delay is 300 seconds (5 minutes)"""
    from time import sleep
    sleep(delay)
    try:
        os.remove(filepath)
    except:
        pass # Don't really care--it might have been removed already

def list_init_scripts():
    """
    Returns a list of lists of all init scripts in the following format (all strings):
    [
        [
            <init.d script name>,
            <rc.d link name(s)>,
            <enabled: true/false>,
            <valid/invalid>,
            <START value>
        ],
    ]
    """
    import re
    start_regex = re.compile('.*START=(.*)')
    scripts_list = []
    scripts_dict = {}
    enabled_scripts = []
    # First we assemble a dictionary of <init script name>: <rc.d link>
    links = os.listdir('/etc/rc.d/')
    for link in links:
        link_path = "/etc/rc.d/%s" % link
        linked_script = os.path.split(os.path.realpath(link_path))[1]
        if linked_script == 'rcS':
            pass # rcS is executed from /etc/inittab so we shouldn't be messing with it here
        else:
            if linked_script not in scripts_dict.keys():
                scripts_dict.update({linked_script: link})
            else: # We need to add a second link for this script
                scripts_dict[linked_script] = "%s, %s" % (scripts_dict[linked_script], link)
            # If the script is linked and executable it is enabled:
            if os.access(os.path.realpath(link_path), os.X_OK): # os.X_OK == executable
                if linked_script not in enabled_scripts:
                    enabled_scripts.append(linked_script)
    # Next we assemble the output, checking for scripts in /etc/init.d that aren't enabled
    scripts = os.listdir('/etc/init.d/')
    for script in scripts:
        if script == 'rcS':
            pass # We don't want to include rcS
        elif script in enabled_scripts:
            scripts_list.append([script, scripts_dict[script], 'true'])
        else:
            scripts_list.append([script, '', 'false'])
    # Lastly we mark any files in /etc/init.d that aren't valid init scripts
    # Valid init scripts (on OpenWRT) require the following:
    # A shebang line like so:
    #    #!/bin/sh /etc/rc.common
    # If its link in /etc/rc.d starts with S##:
    #    A START variable/integer: START=50
    # TODO:  Also check for start() and stop() functions
    for index, script_list in enumerate(scripts_list):
        valid_shebang = False
        start_value = None
        for line in open("/etc/init.d/%s" % script_list[0]):
            if line.startswith('#!/bin/sh /etc/rc.common'):
                valid_shebang = True
            if start_regex.match(line):
                start_value = start_regex.split(line)[1].strip()
            if valid_shebang and start_value:
                scripts_list[index].extend(['valid', start_value])
                break # No need to continue reading lines
        if not valid_shebang or not start_value:
            if script_list[0] == 'umount':
                # umount doesn't have a START value since it only runs at rebooot/halt
                scripts_list[index].extend(['valid', start_value])
            else:
                scripts_list[index].extend(['invalid', start_value])
    return scripts_list

def disable_initscript(script):
    """Disables the given init script by removing its links from /etc/rc.d"""
    scripts_list = list_init_scripts()
    for script_list in scripts_list:
        if script_list[0] == script:
            link = script_list[1]
            break # No need to continue
    if ',' in link: # There's an S *and* a K script
        link_list = link.split(',')
        for _link in link_list:
            _link = _link.strip()
            os.remove('/etc/rc.d/%s' % _link)
    elif link:
        os.remove('/etc/rc.d/%s' % link)

def html_escape(text):
    """Produce entities within text."""
    html_escape_table = {
        "&": "&amp;",
        '"': "&quot;",
        "'": "&apos;",
        ">": "&gt;",
        "<": "&lt;",
    }
    L=[]
    for c in text:
        L.append(html_escape_table.get(c,c))
    return "".join(L)

def isrunning(pid):
    """Returns True if the process associated with pid is still alive"""
    return os.path.exists('/proc/%s' % pid)

class System(Plugin):
    """A plugin for managing your system."""
    display = True
    sort_priority = 1
    navigable_functions = odict([
        (_('Backup And Restore'), 'BackupAndRestore'),
        (_('Hostname'), 'Hostname'),
        (_('Hosts'), 'Hosts'),
        (_('Init Scripts'), 'InitScripts'),
        (_('Reboot'), 'Reboot'),
        (_('System Logs'), 'SystemLogs'),
        (_('Timezone'), 'Timezone'),
        (_('Users &amp; Groups'), 'UsersAndGroups'),
        ])

    @require(has_read('System'))
    def UsersAndGroups(self):
        """Render the users and groups administration page"""
        users_template = plugin_template(self, "templates/users_and_groups.mako")
        group_names = []
        for group in list_etc_group(): # This builds a list of group names
            group_names.append(group[0])

        user_list = []
        for user in list_etc_passwd(): # This builds a list of user names
            user_list.append(user[0])

        shell_list = valid_shells()
        shells = jqgrid_select_options(shell_list)

        passwd = PasswdFile()
        try:
            shadow = open('/etc/shadow', 'r').read()
        except:
            shadow = _("/etc/shadow is not in use on this system")
        groupfile = GroupFile()
        help=_("<p>This interface lets you manage users and groups on your system.</p> <p>Under the Users tab you can add new users or edit the properties of existing ones.  User's passwords can be changed, disabled, or locked via the context menu (right-click).</p><p>The Groups tab lets you add or edit groups.  It also sports a context menu.</p><p> Both the Users and Groups tabs allow you to export their respective grids as spreadsheets.  When you click on the 'Export to spreadsheet' button your web browser should pop up a dialog that can open the document in your spreadsheet program automatically.</p><p>If you like, you may edit the /etc/passwd, /etc/shadow, and /etc/group files directly by clicking on their respective buttons in the interface.</p>")

        return self.standard_render(
            template=users_template,
            help=help,
            long_name=_('PyCI Administration Interface: Users &amp; Groups'),
            description=_("User &amp; Group Administration:  Add/edit users and groups on your system."),
            groups=group_names,
            shells=shells,
            shell_list=shell_list,
            user_list=user_list,
            raw_passwd=passwd.raw(),
            raw_shadow=shadow,
            raw_group=groupfile.raw(),
        )

    @require(has_read('System'))
    def Reboot(self):
        """Render a page where the user can reboot the system.

        Note the capital R in 'Reboot'.  Not to be confused with 'reboot' which immediately reboots the system."""
        reboot_template = plugin_template(self, "templates/reboot.mako")

        return self.standard_render(
            template=reboot_template,
            help=_('<p>When you click that "Reboot" button it will restart this system--applying any changes you have made to system configuration files.</p><p>Rebooting typically takes anywhere from 10-120 seconds depending on the system and how it is configured.</p>'),
            long_name=_('PyCI Administration Interface: Reboot'),
            description=_("Reboot:  Reboot this host."),
        )

    @require(has_write('System'))
    def reboot(self):
        """Reboots the system"""
        # We only reboot if we're POST'd to (to save people from accidentally typing 'reboot' instead of "Reboot"
        if cherrypy.request.method == 'POST':
            from platform import uname
            hostname = uname()[1]
            try:
                os.system('reboot')
                #import sys
                #sys.exit(0) # Use this for testing.
                return jquery_ui_highlight_msg(_('%s is now rebooting...') % hostname)
            except Exception as e:
                return jquery_ui_error_msg(_('An error was encountered when trying to reboot the system.<br />%s') % e)
        else:
            return _("This method will only reboot the host if it gets a POST.  Simply doing a GET on this URL does nothing.")

    @require(has_read('System'))
    def Hostname(self):
        """Render a page where the user can change the system's hostname."""
        _template = plugin_template(self, "templates/hostname.mako")
        import platform
        hostname = platform.uname()[1]

        return self.standard_render(
            template=_template,
            help=_('<p>Use the form below to change the (short) hostname of this system.  For changing the fully-qualifieid domain name (FQDN) of this system you can use the <a href="/System/Hosts">Hosts</a> configuration page.</p>'),
            long_name=_('System Configuration: Hostname'),
            description=_("Hostname:  Change the hostname of this host."),
            hostname=hostname,
        )

    @require(has_write('System'))
    def update_hostname(self, hostname):
        """
        Updates the system hostname with the given hostname.

        That means writing the hostname to /proc/sys/kernel/hostname
        ...and if it exists, update /etc/hostname
        ...and if this is OpenWRT, update /etc/config/system too
        """
        try:
            sys_conf = '/etc/config/system'
            if os.path.exists(sys_conf): # This is an OpenWRT host
                system_config = ConfigFileOpenWRT(sys_conf)
                for index, section in enumerate(system_config.sections):
                    if section['config'] == 'system':
                        break # We just want the index
                system_config.sections[index]['options']['hostname'] = hostname
                system_config.save()
            if os.path.exists('/etc/hostname'): # Write it here too
                hostname_file = open('/etc/hostname', 'w')
                hostname_file.write(hostname)
                hostname_file.close()
            # Now update the kernel with the new hostname
            kernel_hostname = open('/proc/sys/kernel/hostname', 'w')
            kernel_hostname.write(hostname)
            kernel_hostname.close()

            return jquery_ui_highlight_msg(_('Hostname changed to "%s" successfully.') % hostname)
        except Exception as e:
            return jquery_ui_error_msg(_('Unable to change the hostname.  Please make sure PyCI is running as root (what else could it be?)<br />%s') % e)

    @require(has_read('System'))
    def Hosts(self):
        """Render a page where the user can edit /etc/hosts."""
        _template = plugin_template(self, "templates/hosts.mako")

        return self.standard_render(
            template=_template,
            help=_('<p>Use the grid edit hostnames on this system.  You should be able to edit items in-line simply by clicking on them.  Hosts can be deleted by selecting them and clicking on the trash can icon in the lower left-hand corner of the grid.  Hosts can be added by clicking on the plus sign (+) icon.</p>'),
            long_name=_('System Configuration: Hosts'),
            description=_("Edit /etc/hosts."),
        )

    @require(has_read('System'))
    def hosts_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns /etc/hosts in a json format that's compatible with jqGrid."""
        header = ["ip", "hostnames"]
        return jqgrid_json(self, list_etc_hosts(), header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)

    @require(has_write('System'))
    def update_hosts(self, oper=None, ip=None, hostnames=None, id=None):
        """
        Updates /etc/hosts with the given parameters.
        """
        out_hosts = ""
        if oper == "del": # We're' deleting an ip
            # Note: This preserves comments =)
            for line in open('/etc/hosts'):
                if line.lstrip().startswith(id):
                    pass # We don't want this one
                else:
                    out_hosts += line
            write_etc_hosts(out_hosts)
        elif oper == "add": # Adding a new ip/hostname(s)
            # First we double-check that /etc/hosts ends in a newline character:
            for line in open('/etc/hosts'):
                last_line = line
            if not last_line.endswith('\n'): # No newline?!? Fix it
                # 'with .." ensures the file is closed after write() is complete
                with open('/etc/hosts', 'a') as f:
                    f.write('\n') # Write that missing newline
            etc_hosts = open('/etc/hosts', 'a') # Append mode, love it!
            etc_hosts.write("%s\t%s\n" % (ip, hostnames))
            etc_hosts.close()
        else: # We're doing an update
            for line in open('/etc/hosts'):
                if line.lstrip().startswith(id): # Note that's 'id' and not 'ip'
                    # Replace with our modified version
                    out_hosts += "%s\t%s\n" % (ip, hostnames)
                else: # Leave it alone
                    out_hosts += line
            write_etc_hosts(out_hosts)

    @require(has_write('System'))
    def add_group(self, group, gid=None, users=None, **kwargs):
        """Add a group to /etc/group.
        Requires a group name, 'group' and optionally accepts a gid and list of users.
        'users' can either be a Python list or a comma-separated string."""
        if not gid:
            gid = gen_unique_gid()
        if users and isinstance(users, str):
            users = users.split(',')
        try:
            new_group = Group(group, 'x', gid, users)
            new_group.save()
            return jquery_ui_highlight_msg(_('Group %s added successfully.') % group)
        except Exception as e:
            return jquery_ui_error_msg(_('Group %s could not be saved.<br />%s') % (group, e))

    @require(has_write('System'))
    def add_user(self, username, shell=None, password=None, uid=None, gid=None, gecos=None, groups=None, homedir=None, **kwargs):
        """Add a user"""
        # Note:  If only one group is passed, 'groups' does not appear as a list
        # TODO:  Add the ability to have a preference for default values
        # {'username': 'testuser', 'shell': '/bin/csh', 'uid': '', 'asmSelect0': 'Select groups', 'gid': '100', 'gecos': '', 'groups': 'pyci', 'homedir': ''}
        password_disabled = False
        passwd = PasswdFile()
        if passwd.user(username):
            return jquery_ui_error_msg(_('User %s already exists.  Please try a different username.') % username)
        if not password:
            password = '!' # The User class will automatically save '!' as a "Disabled" password
            password_disabled = True
        if not shell:
            shell = valid_shells()[0] # Default to the first valid shell in /etc/shells
            for valid_shell in valid_shells():
                if 'bash' in valid_shell: # ...unless we have bash.  Use that instead =)
                    shell = valid_shell # TODO:  Change this so it can be configured via preference
        if not uid:
            uid = str(gen_unique_uid())
        if not gid:
            group = GroupFile()
            if not group.group(username): # No group already exists with this name so we create one
                gid = str(gen_unique_gid())
                new_group = Group(username, 'x', gid)
                new_group.save()
            else: # Group already exists, use it
                gid = str(group.group(username).gid)
        if not homedir:
            homedir = '/home/%s' % username
        try:
            userobj = User(username, password, uid, gid, gecos, homedir, shell)
            if os.path.exists('/etc/shadow'):
                userobj.construct_shadow_entry() # Make an ShadowEntry shell object before saving
            if not password_disabled:
                userobj.hash_password() # Hash the password before saving (doesn't do it automatically)
            userobj.save()
            # Now add the user to the supplied supplemental groups (if any)
            if groups:
                groupfile = GroupFile()
                if isinstance(groups, str): # i.e. not a list
                    groupobj = groupfile.group(groups)
                    groupobj.users.append(username)
                    groupobj.save()
                else:
                    for group in groups:
                        groupobj = groupfile.group(group)
                        groupobj.users.append(username)
                        groupobj.save()
            return jquery_ui_highlight_msg(_('User %s added successfully.') % username)
        except Exception as e:
            return jquery_ui_error_msg(_('User %s could not be saved.<br />%s') % (username, e))

    @require(has_read('System'))
    def get_hosts_spreadsheet(self):
        """Exports /etc/hosts as a spreadsheet"""
        spreadsheet_template = lookup_template('html', 'spreadsheet.mako')
        cherrypy.response.headers['Content-Type'] = 'application/vnd.ms-excel; charset=utf-8'
        headers = [_("IP Address"), _("Hostnames")]
        rows = list_etc_hosts()
        return spreadsheet_template.render(headers=headers, rows=rows)

    @require(has_read('System'))
    def get_users_spreadsheet(self):
        """Exports the user permissions table as a spreadsheet"""
        spreadsheet_template = lookup_template('html', 'spreadsheet.mako')
        cherrypy.response.headers['Content-Type'] = 'application/vnd.ms-excel; charset=utf-8'
        headers = [_("Username"), _("Password"), _("UID"), _("GID"), _("GECOS"), _("Home Directory"), _("Shell")]
        rows = list_etc_passwd()
        return spreadsheet_template.render(headers=headers, rows=rows)

    @require(has_read('System'))
    def get_syslog_spreadsheet(self):
        """Exports the syslog grid as a spreadsheet"""
        spreadsheet_template = lookup_template('html', 'spreadsheet.mako')
        cherrypy.response.headers['Content-Type'] = 'application/vnd.ms-excel; charset=utf-8'
        headers = [_("Date"), _("Hostname"), _("Facility"), _("Priority"), _("Process"), _("PID"), _("Message")]
        syslog = logread('%m-%d %H:%M:%S')
        return spreadsheet_template.render(headers=headers, rows=syslog)

    @require(has_read('System'))
    def get_groups_spreadsheet(self):
        """Exports the group permissions table as a spreadsheet"""
        spreadsheet_template = lookup_template('html', 'spreadsheet.mako')
        cherrypy.response.headers['Content-Type'] = 'application/vnd.ms-excel; charset=utf-8'
        headers = [_("Group"), _("GID"), _("Users")]
        rows = list_etc_group()
        return spreadsheet_template.render(headers=headers, rows=rows)

    @require(has_read('System'))
    def get_initscripts_spreadsheet(self):
        """Exports the init scripts grid as a spreadsheet"""
        spreadsheet_template = lookup_template('html', 'spreadsheet.mako')
        cherrypy.response.headers['Content-Type'] = 'application/vnd.ms-excel; charset=utf-8'
        headers = [_('Script'), '/etc/rc.d', _('Enabled'), _('Valid'), 'START', _('Action')]
        rows = list_init_scripts()
        return spreadsheet_template.render(headers=headers, rows=rows)

    @require(has_write('System'))
    def update_groups(self, oper=None, id=None, group=None, gid=None, users=None):
        """jqGrid function:  Perform the specified operation on the specified user."""
        print cherrypy.request.params
        if oper == "del":
            try:
                self.delete_group(id)
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> %s has been deleted.') % id)
            except Exception as e:
                return jquery_ui_error_msg(_('Could not delete %s.  Please make sure PyCI is running as root.<br />%s') % (id,e))
        # Default to updating since jqGrid doesn't pass anything like "oper=update" when users perform an edit-in-place
        try:
            groupfile = GroupFile()
            group = groupfile.group(group)
            group.groupname = id
            group.users = users
            group.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> %s has been updated.') % id)
        except Exception as e:
            return jquery_ui_error_msg(_('Could not update group %s.  Please make sure PyCI is running as root.<br />%s') % (id,e))

    @require(has_write('System'))
    def delete_group(self, groupname):
        """Permanentely deletes the given user"""
        groups = GroupFile()
        group = groups.group(groupname)
        group.delete()

    @require(has_read('System'))
    def users_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns all users from /etc/passwd in a json format that's compatible with jqGrid."""
        header = ["user", "password", "uid", "gid", "gecos", "homedir", "shell"]
        passwd_list = list_etc_passwd()
        return jqgrid_json(self, passwd_list, header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)

    @require(has_write('System'))
    def update_users(self, oper=None, id=None, shell=None, gecos=None, uid=None, gid=None, homedir=None, user=None):
        """jqGrid function:  Perform the specified operation on the specified user."""
        if oper == "del":
            try:
                self.delete_user(id)
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> %s has been deleted.') % id)
            except Exception as e:
                return jquery_ui_error_msg(_('Could not delete %s.  Please make sure PyCI is running as root.<br />%s') % (id,e))

        # Default to updating since jqGrid doesn't pass anything like "oper=update" when users perform an edit-in-place
        try:
            passwd = PasswdFile()
            user = passwd.user(user)
            user.shell = shell
            user.gecos = gecos
            user.homedir = homedir
            user.uid = uid
            user.gid = gid
            user.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> %s has been updated.') % id)
        except Exception as e:
            return jquery_ui_error_msg(_('Could not update user %s.  Please make sure PyCI is running as root.<br />%s') % (id,e))

    @require(has_write('System'))
    def change_user_attribute(self, username, **attributes):
        """Changes a given user's attribute in /etc/passwd and /etc/shadow (if present).  Valid attributes are:
        username, password, uid, gid, gecos, homedir, shell, disable (password), and locked (password)."""
        user_password = None
        confirm_password = None
        print cherrypy.request.params
        passwd = PasswdFile()
        try:
            user = passwd.user(username)
            for attribute, value in attributes.iteritems():
                if attribute == "username":
                    pass # Figure out a way to implement user name changes
                elif attribute == "password": # Don't set it yet
                    user_password = value
                elif attribute == "confirm_password": # Make sure the passwords match
                    confirm_password = value
                elif attribute == "uid":
                    user.uid = value
                elif attribute == "gid":
                    user.uid = value
                elif attribute == "gecos":
                    user.gecos = value
                elif attribute == "homedir":
                    user.homedir = value
                elif attribute == "shell":
                    user.shell = value
                elif attribute == "disable":
                    if value:
                        user.disable()
                elif attribute == "locked":
                    if value == 'true':
                        user.lock()
                    elif value == 'false':
                        user.unlock()
            if user_password:
                if user_password == confirm_password:
                    if os.path.exists('/etc/shadow'):
                        user.shadow.password = user_password
                        user.hash_password()
                    else:
                        user.password = user_password
                        user.hash_password()
                else:
                    return jquery_ui_error_msg(_('Could not change password for %s.  Passwords do not match.') % username)
            user.save()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> %s updated.') % username)
        except:
            return jquery_ui_error_msg(_('Could not change password for %s.  Please make sure PyCI is running as root.') % username)

    @require(has_write('System'))
    def delete_user(self, username):
        """Permanentely deletes the given user and removes that user from any groups it resides in."""
        passwd = PasswdFile()
        userobj = passwd.user(username)
        userobj.delete()
        del(passwd)
        # TODO:  Finish writing the bit that removes the user from the various groups
        groupfile = GroupFile()

    @require(has_read('System'))
    def get_raw_passwd(self):
        """Returns the raw passwd file"""
        return open('/etc/passwd', 'r').read()

    @require(has_write('System'))
    def update_raw_passwd(self, passwd):
        """Save the raw passwd file text as submitted"""
        try:
            f = open('/etc/passwd', 'w')
            f.write(passwd)
            f.close()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/passwd was saved.'))
        except:
            return jquery_ui_error_msg(_('Could not save /etc/passwd.  Please make sure PyCI is running as root.'))

    @require(has_write('System'))
    def update_raw_shadow(self, shadow):
        """Save the raw passwd file text as submitted"""
        try:
            f = open('/etc/shadow', 'w')
            f.write(shadow)
            f.close()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/shadow was saved.'))
        except:
            return jquery_ui_error_msg(_('Could not save /etc/shadow.  Please make sure PyCI is running as root.'))

    @require(has_write('System'))
    def update_raw_group(self, group):
        """Save the raw /etc/group file text as submitted"""
        try:
            f = open('/etc/group', 'w')
            f.write(group)
            f.close()
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/group was saved.'))
        except:
            return jquery_ui_error_msg(_('Could not save /etc/group.  Please make sure PyCI is running as root.'))

    @require(has_read('System'))
    def groups_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns all groups from /etc/group in a json format that's compatible with jqGrid."""
        header = ["group", "gid", "users"]
        group_list = list_etc_group()
        return jqgrid_json(self, group_list, header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)

    @require(has_read('System'))
    def SystemLogs(self):
        """Render the main system log (syslog) and kernel log (dmesg) page.
        Default view is the syslog."""
        _template = plugin_template(self, "templates/system_logs.mako")

        return self.standard_render(
            template=_template,
            help=_("<p>This interface allows you to view the system logs.</p>"),
            long_name=_('System: Logs'),
            description=_("View this host's system and kernel logs."),
        )

    @require(has_read('System'))
    def syslog_tab(self):
        """Render the syslog view/sort/search tab."""
        _template = plugin_template(self, "templates/syslog.mako")
        from commands import getstatusoutput
        retcode, raw_syslog = getstatusoutput('logread')
        if retcode != 0:
            try:
                raw_syslog = open('/var/log/messages').read()
            except:
                raw_syslog = "Could not open system log"

        return self.standard_render(
            template=_template,
            raw_syslog=raw_syslog,
        )

    @require(has_read('System'))
    def dmesg_tab(self):
        """Render the kernel log view tab."""
        _template = plugin_template(self, "templates/dmesg.mako")
        from commands import getoutput
        dmesg = getoutput('dmesg')

        return self.standard_render(
            template=_template,
            dmesg=unicode(dmesg, errors='ignore'),
        )

    @require(has_read('System'))
    def syslog_conf(self):
        """Render the syslog configuration tab."""
        _template = plugin_template(self, "templates/syslog_conf.mako")
        system_conf = ConfigFileOpenWRT(filepath='/etc/config/system')
        if 'log_size' in system_conf.sections[0]['options']:
            log_size = system_conf.sections[0]['options']['log_size']
        else:
            log_size = ''
        if 'log_ip' in system_conf.sections[0]['options']:
            log_ip = system_conf.sections[0]['options']['log_ip']
        else:
            log_ip = ''

        return self.standard_render(
            template=_template,
            log_size=log_size,
            log_ip=log_ip,
        )

    @require(has_read('System'))
    def syslog_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns the syslog (via 'logread') in a json format that's compatible with jqGrid."""
        header = [_("Date"), _("Hostname"), _("Facility"), _("Priority"), _("Process"), _("PID"), _("Message")]
        syslog = logread('%m-%d %H:%M:%S')
        return jqgrid_json(self, syslog, header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)

    @require(has_read('System'))
    def syslog_options_meta(self, option=None, section=None, help=True, values=False, **data):
        """By default, return help text for the provided system option (for jTip).
        If 'values' is True, return a json-formatted list of possible values for the given option/section.
        If 'section' is provided, attempt to obtain valid values from the specified section
        of the validation file."""
        system_conf = ConfigFileOpenWRT('/etc/config/system')
        validation_obj = system_conf.validation_obj()
        validation_dict = system_conf.validation_dict(validation_obj)
        if values:
            try:
                values_dict = {}
                for value in system_conf.valid_options(section, option):
                    if value == '0' or value == 'off' or value == 'disabled':
                        values_dict.update({value:"disabled"})
                    elif value == '1' or value == 'on' or value == 'enabled':
                        values_dict.update({value:"enabled"})
                    else:
                        values_dict.update({value:value})
                if values_dict:
                    return json.dumps(values_dict)
                else:
                    return "{empty:true}"
            except:
                # This is so we can determine that no list values are available for this option
                # ...in JavaScript: if (!json.empty) { do stuff... }
                return "{empty:true}"

        if help:
            try:
                return '<strong>%s</strong>: %s' % (option, validation_dict[section][option]['title'])
            except:
                return  _('<strong>%s</strong>: No help text was found for this option.') % option

    @require(has_write('System'))
    def update_syslog_conf(self, log_ip, log_size):
        """
        Updates the system log configuration with the supplied options.
        """
        try:
            sys_conf = '/etc/config/system'
            system_config = ConfigFileOpenWRT(sys_conf)
            for index, section in enumerate(system_config.sections):
                if section['config'] == 'system':
                    break # We just want the index
            system_config.sections[index]['options']['log_ip'] = log_ip
            system_config.sections[index]['options']['log_size'] = log_size
            system_config.save()

            return jquery_ui_highlight_msg(_('<strong>Success:</strong> System log configuration updated successfully.'))
        except Exception as e:
            return jquery_ui_error_msg(_('Unable to update the system log configuration. <br />%s') % e)

    @require(has_read('System'))
    def Timezone(self):
        """Render the timezone modification page."""
        from timezones import timezones
        _template = plugin_template(self, "templates/timezone.mako")
        system_config = ConfigFileOpenWRT('/etc/config/system')
        # /etc/config/system only has one section
        if 'zonename' in system_config.sections[0]['options']:
            current_timezone = system_config.sections[0]['options']['zonename']
        else:
            current_timezone = system_config.sections[0]['options']['timezone']

        return self.standard_render(
            template=_template,
            help=_("<p>This interface allows you to change this host's timezone setting.  Choose the location nearest to you or select the specific time zone you wish to use.</p>"),
            long_name=_('System: Timezone'),
            description=_("Change this host's timezone setting."),
            timezones=timezones,
            current_timezone=current_timezone,
        )

    @require(has_write('System'))
    def update_timezone(self, timezone):
        """
        Updates the system timezone and zonename with the given timezone
        """
        from timezones import timezones
        try:
            sys_conf = '/etc/config/system'
            system_config = ConfigFileOpenWRT(sys_conf)
            for index, section in enumerate(system_config.sections):
                if section['config'] == 'system':
                    break # We just want the index
            system_config.sections[index]['options']['zonename'] = timezone
            system_config.sections[index]['options']['timezone'] = timezones[timezone]
            system_config.save()

            return jquery_ui_highlight_msg(_('Timezone changed to "%s" successfully.') % timezone)
        except Exception as e:
            return jquery_ui_error_msg(_('Unable to change the timezone. <br />%s') % e)

    @require(has_read('System'))
    def BackupAndRestore(self):
        """Render the backup and restore page where users can download/upload a tarball containing
        the contents of /etc."""
        _template = plugin_template(self, "templates/backup_and_restore.mako")

        return self.standard_render(
            template=_template,
            help=_("<p>This interface allows you to backup and restore the contents of the /etc directory on your host.  The /etc directory is where all your system configuration files is stored.  This includes things like users and groups, hosts, network settings, etc.</p><p>Backup files are stored in a standard compressed tarball format (*.tar.gz) that can be extracted and viewed using widely available free software.</p>"),
            long_name=_('System: Backup and Restore'),
            description=_("Backup or restore your system's configuration."),
        )

    @require(has_read('System'))
    def backup(self):
        """Returns a compressed tarball containing the contents of /etc that can be used by self.restore."""
        import tarfile, datetime
        from platform import uname
        from cherrypy.lib.static import serve_file
        import thread

        now = datetime.datetime.now()
        hostname = uname()[1]
        filepath = "/tmp/%s_etc-%s.tar.gz" % (hostname, now.strftime('%Y%m%d_%H%M%S'))
        tar = tarfile.open(filepath, 'w:gz')
        cwd = os.getcwd() # Save our current working directory so we can go back
        os.chdir('/')
        tar.add('/etc') # Add everything in /etc to our tarball
        tar.close()
        os.chdir(cwd) # Go back to where we were before
        # Kick off a delayed removal of the tarball (no need to leave it hanging around)
        thread.start_new_thread(delayed_remove_file, (filepath, 300))
        # Now serve up the file to the client
        return serve_file(filepath, "application/x-download", "attachment")

    @require(has_write('System'))
    def restore(self, backup):
        """Restores /etc using the given 'backup_file' which should be a tarball created using self.backup()"""
        import tarfile
        combined_data = ''
        cwd = os.getcwd() # Save our current working directory so we can go back
        os.chdir('/')
        uploaded_file = open(backup.filename, 'wb')
        try:
            while True: # Read the uploaded file in chunks (in case it is large--you never know!)
                data = backup.file.read(8192)
                if not data:
                    break
                uploaded_file.write(data)
            uploaded_file.close()
            # Now extract the tarball overwriting everything in /etc
            tar = tarfile.open(backup.filename)
            tar.extractall('/')
            # Change back to our previous working directory
            os.chdir(cwd)
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> %s was restored successfully.') % backup.filename)
        except Exception as e:
            return jquery_ui_error_msg(_('Could extract the uploaded file. <br />%s') % e)

    @require(has_read('System'))
    def InitScripts(self):
        """Render a page where the user can stop, start, restart, and configure system init scripts
        (i.e. the scripts in /etc/init.d/ and /etc/rc.d)."""
        _template = plugin_template(self, "templates/initscripts.mako")

        return self.standard_render(
            template=_template,
            help=_("""<p>This plugin allows you to stop, start, restart, enable, disable, and edit init scripts on this system.  Init scripts are the scripts in /etc/init.d that are run when the system boots or shuts down/restarts.</p>
            <p>When you enable or disable an init script a symbolic link is either added or removed from the /etc/rc.d directory.  These links start with S## or K## where ## is a number from 00 through 99.  Scripts that start with "S" will be run at boot (they will be excuted as, "&lt;script&gt; start") and scripts that start with "K" will be run at shutdown/reboot ("&lt;script&gt; stop").</p>
            <p>When you enable or disable an init script via this interface a symbolic link will be created or deleted from the /etc/rc.d directory.  To specify the order of scripts just change their START value.</p>
            <p><strong>Note:</strong> When you enable a script PyCI will honor START variables.  That means if a script contains 'START=50' PyCI will create the symbolic link as /etc/rc.d/S50scriptname when you enable it.</p>"""),
            long_name=_('PyCI Administration Interface: Init Scripts'),
            description=_("Start, stop, restart, enable, disable, or edit init scripts on this system.<br /> <strong>Tip:</strong> You can right-click on a row to perform various actions."),
        )

    @require(has_read('System'))
    def initscripts_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns /etc/hosts in a json format that's compatible with jqGrid.
        Note: Appends an empty list item to each script's list as a placeholder for the Action column."""
        header = ["script", "link", "enabled", "valid", "start", "action"]
        scripts_list = list_init_scripts()
        for script_list in scripts_list:
            script_list.append('')
        return jqgrid_json(self, scripts_list, header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)

    @require(has_read('System'))
    def get_initscript(self, script):
        """Returns the text of the given script as a string"""
        try:
            return open('/etc/init.d/%s' % script).read()
        except:
            return _("Could not open /etc/init.d/%s for reading.  Freaky.") % script

    @require(has_write('System'))
    def update_initscript(self, script=None, oper=None, script_text=None, id=None, start=None, enabled=None):
        """
        Updates the given init.d script with the provided script_text or performs the given operation (oper)
        """
        # TODO:  Refactor this WHOLE method.  I'm sure I can simplify it considerably.
        import re
        start_regex = re.compile('.*START=(.*)')
        script_rename = False
        if id:
            if script and script != id:
                script_rename = script # Used later in rename operations
            script = id # jqGrid passes the script name in as 'id'
        scripts_list = list_init_scripts()
        if oper == 'save':
            try:
                script_path = '/etc/init.d/%s' % script
                print "Trying to write %s" % script_path
                f = open(script_path, 'w')
                f.write(script_text)
                f.close()
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/init.d/%s was saved.') % script)
            except Exception as e:
                return jquery_ui_error_msg(_('Could not save /etc/init.d/%s.  Please make sure PyCI is running as root.<br /> %s') % (script, e))
        if oper == 'enable_boot' or oper == 'enable_halt':
            try:
                link = None
                # Check to make sure there isn't already a boot link in /etc/rc.d
                for script_list in scripts_list:
                    if script_list[0] == script:
                        link = script_list[1]
                        start_value = script_list[4]
                        break # No need to continue
                if ',' in link: # There's an S *and* a K script
                    link_list = link.split(',')
                    for _link in link_list:
                        _link = _link.strip()
                        if oper == 'enable_boot':
                            if _link.startswith('S'):
                                link = _link
                                break
                        else: # enable_halt
                            if _link.startswith('K'):
                                link = _link
                                break
                else:
                    if oper == 'enable_boot':
                        if link.startswith('S'):
                            pass
                        else: # Link starts with K
                            link = None
                    else: # enable_halt
                        if link.startswith('K'):
                            pass
                        else: # Link starts with S
                            link = None
                if oper == 'enable_boot' and link:
                    return jquery_ui_error_msg(_('%s is already enabled at boot as /etc/rc.d/%s.') % (script, link))
                elif oper == 'enable_halt' and link:
                    return jquery_ui_error_msg(_('%s is already enabled at shutdown as /etc/rc.d/%s.') % (script, link))
                else: # There's no existing link so we'll make one
                    if oper == 'enable_boot':
                        boot_link = '/etc/rc.d/S%s%s' % (start_value, script)
                        os.symlink('/etc/init.d/%s' % script, boot_link)
                        return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/init.d/%s was enabled at boot as %s.') % (script, boot_link))
                    else:
                        halt_link = '/etc/rc.d/K%s%s' % (start_value, script)
                        os.symlink('/etc/init.d/%s' % script, halt_link)
                        return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/init.d/%s was enabled at shutdown as %s.') % (script, halt_link))
            except Exception as e:
                return jquery_ui_error_msg(_('Could not enable %s at boot or shutdown.<br />%s') % (script, e))
        if oper == 'disable_boot' or oper == 'disable_halt':
            try:
                link = None
                # Find the link(s)
                for script_list in scripts_list:
                    if script_list[0] == script:
                        link = script_list[1]
                        break # No need to continue
                if ',' in link: # There's an S *and* a K script
                    link_list = link.split(',')
                    for _link in link_list:
                        _link = _link.strip()
                        if oper == 'disable_boot':
                            if _link.startswith('S'):
                                link = _link
                                break
                        else: # disable_halt
                            if _link.startswith('K'):
                                link = _link
                                break
                else:
                    if oper == 'disable_boot':
                        if link.startswith('S'):
                            pass
                        else: # Link starts with K
                            link = None
                    else: # enable_halt
                        if link.startswith('K'):
                            pass
                        else: # Link starts with S
                            link = None
                if oper == 'disable_boot' and not link:
                    return jquery_ui_error_msg(_('%s is already disabled at boot.') % script)
                elif oper == 'disable_halt' and not link:
                    return jquery_ui_error_msg(_('%s is already disabled at shutdown.') % script)
                else: # There's an existing link so we'll remove it
                    if oper == 'disable_boot':
                        boot_link = '/etc/rc.d/%s' % link
                        os.remove(boot_link)
                        return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/init.d/%s was disabled at boot.') % script)
                    else:
                        halt_link = '/etc/rc.d/%s' % link
                        os.remove(halt_link)
                        return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/init.d/%s was disabled at shutdown.') % script)
            except Exception as e:
                return jquery_ui_error_msg(_('Could not disable %s at boot or shutdown.<br />%s') % (script, e))
        if oper == "del":
            try:
                for script_list in scripts_list:
                    if script_list[0] == script:
                        link = script_list[1]
                        break # No need to continue
                if ',' in link: # There's an S *and* a K script
                    link_list = link.split(',')
                    for _link in link_list:
                        os.remove('/etc/rc.d/%s' % _link)
                else:
                    os.remove('/etc/rc.d/%s' % link)
                os.remove('/etc/init.d/%s' % script)
                return jquery_ui_highlight_msg(_('<strong>Success:</strong> /etc/init.d/%s has been deleted and all links to it from /etc/rc.d have been removed.') % id)
            except Exception as e:
                return jquery_ui_error_msg(_('Could not delete %s.  Please make sure PyCI is running as root.<br />%s') % (id,e))
        # At this point if we've run out of 'oper' operations so we must assume that it is simply jqGrid pushing an update
        if script_rename: # This is a rename operation
            # Rename the script
            os.rename('/etc/init.d/%s' % script, '/etc/init.d/%s' % script_rename)
            # Now update its links in /etc/rc.d (if any) to point to the new name
            link = None
            for script_list in scripts_list:
                if script_list[0] == script:
                    link = script_list[1]
                    start_value = script_list[4]
                    break # No need to continue
            if ',' in link: # There's an S *and* a K script
                link_list = link.split(',')
                for _link in link_list:
                    _link = _link.strip()
                    os.remove('/etc/rc.d/%s' % _link)
                    if _link.startswith('S'):
                        boot_link = '/etc/rc.d/S%s%s' % (start_value, script_rename)
                        os.symlink('/etc/init.d/%s' % script_rename, boot_link)
                    elif _link.startswith('K'):
                        halt_link = '/etc/rc.d/K%s%s' % (start_value, script_rename)
                        os.symlink('/etc/init.d/%s' % script_rename, halt_link)
            elif link:
                os.remove('/etc/rc.d/%s' % link)
                boot_link = '/etc/rc.d/S%s%s' % (start_value, script_rename)
                os.symlink('/etc/init.d/%s' % script_rename, boot_link)
        if enabled == 'false':
            # Check if this script is enabled and if it is, disable it
            for script_list in scripts_list:
                if script_list[0] == script:
                    link = script_list[1]
                    enabled = script_list[2]
                    break # No need to continue
            if enabled == 'true':
                # Disable it
                disable_initscript(script)
        if start: # We need to modify the script's START value
            output = ""
            matched = False
            for line in open('/etc/init.d/%s' % script):
                if start_regex.match(line):
                    matched = True
                    start_value = start_regex.split(line)[1]
                    output += line.replace('START=%s' % start_value, 'START=%s' % start)
                else:
                    output += line
            if not matched:
                return jquery_ui_error_msg(
                _('/etc/init.d/%s does not have a START variable (it is broken).  Please edit it and add "START=##" at the top (where ## is something like 50).') % script
                )
            f = open('/etc/init.d/%s' % script, 'w')
            f.write(output)
            f.close()


    @require(has_write('System'))
    def initscript_service_control(self, oper, service=None):
        """Stop, Start, or Restart the given service depending on the value of 'oper':
        'start'
        'stop'
        'restart'"""
        if not service:
            return jquery_ui_error_msg(
                _('You must specify a service to perform an operation on.')
            )
        from utils import service_control
        valid_operations = ['start', 'stop', 'restart']
        if oper in valid_operations:
            result = service_control(service, oper)
            if result == True:
                return jquery_ui_highlight_msg(
                    _('<strong>Success:</strong> "/etc/init.d/dropbear %s" executed successfully.') % oper
                )
            else:
                return jquery_ui_error_msg(
                    _('Could not execute /etc/init.d/%s %s.<br />%s') % (oper, service, result)
                )
        else:
            return jquery_ui_error_msg(
                _('Unrecognized operation: %s.<br /> Must be one of "start", "stop", or "restart".') % oper
            )

    @require(has_read('System'))
    def validate_option(self, config=None, section=None, **kwargs):
        """Returns a json-encoded error message if the given option:value pair does not validate"""
        config_obj = ConfigFileOpenWRT('/etc/config/%s' % config)
        result = None
        # This should only ever loop once
        for key, value in kwargs.items():
            result = config_obj.validate_option(section, key, value)
        if result:
            return json.dumps(result)

    @require(has_write('System'))
    def eftw_terminal_size(self, rows, cols, **kw):
        """Sets the server-side terminal window size and also stores it in the client's session.
        (Useful for when the browser window is resized)"""
        # First set the session rows/cols:
        from datetime import datetime
        rows = int(rows)
        cols = int(cols)
        #print "rows: %s" % `rows`
        #print "cols: %s" % `cols`
        username = cherrypy.session.get('_cp_username')
        cherrypy.session['eftw_rows'] = rows
        cherrypy.session['eftw_cols'] = cols
        cherrypy.session.save()
        # Next set the running terminal's rows/cols (if it exists)
        try:
            # Note:  This doesn't work properly (yet) for some reason
            term = eftw.sessions[username]['term']
            eftw.sessions[username]['multiplexer'].proc[term]['term'].width = cols
            eftw.sessions[username]['multiplexer'].proc[term]['term'].height = rows
            eftw.sessions[username]['multiplexer'].proc[term]['term'].redraw = True
            # This peforms the equivalent of control-l
            eftw.sessions[username]['multiplexer'].proc[term]['term'].reset()
            eftw.sessions[username]['multiplexer'].resize(term, rows, cols)
        except: # Doesn't exist yet, no biggie
            pass
        return "success"

    @require(has_write('System'))
    def eftw(self, chars=None, rows=None, cols=None, **kw):
        """
        Provides an Ajaxterm-like window...
        """
        from datetime import datetime, timedelta
        from time import sleep
        username = cherrypy.session.get('_cp_username')
        if username not in eftw.sessions:
            # Start a new root login (-l) shell
            # TODO:  Make this start a shell owned by the connected user
            eftw.sessions[username] = {}
            if 'eftw_rows' not in cherrypy.session or 'eftw_cols' not in cherrypy.session:
                # Default to 80x24
                rows, cols = 24, 80
                cherrypy.session['eftw_rows'] = rows
                cherrypy.session['eftw_cols'] = cols
                cherrypy.session.save()
            else:
                rows = cherrypy.session['eftw_rows']
                cols = cherrypy.session['eftw_cols']
            # Figure out the user's shell
            import pwd
            shell = pwd.getpwnam(username)[6]
            uid = pwd.getpwnam(username)[2]
            eftw.sessions[username]['multiplexer'] = eftw.Multiplex(shell, uid)
            # This makes the EFTW shell perform /bin/login:
            #eftw.sessions[username]['multiplexer'] = eftw.Multiplex()
            eftw.sessions[username]['term'] = eftw.sessions[username]['multiplexer'].create(cols, rows)
            # Start the keepalive thread so the session will time out if the user disconnects
            eftw.sessions[username]['keepalive_thread'] = EFTWKeepaliveThread(username, datetime.now())
            eftw.sessions[username]['keepalive_thread'].start()
        if eftw.sessions[username]['multiplexer'].alive:
            eftw.sessions[username]['keepalive_thread'].keepalive(datetime.now())
            if chars:
                eftw.sessions[username]['multiplexer'].proc_write(eftw.sessions[username]['term'], chars)
                sleep(0.02)
            output = eftw.sessions[username]['multiplexer'].dump(eftw.sessions[username]['term'])
            cherrypy.response.headers['Content-Type'] = "text/xml"
            return output

class EFTWKeepaliveThread(threading.Thread):
    """
    Kills a user's EFTW session if the client hasn't updated the keepalive within 30 seconds.
    Note: This is necessary to prevent shells from running eternally in the background.
    """
    def __init__(self, username, last_keepalive):
        threading.Thread.__init__(self, name="EFTWKeepaliveThread-%s" % (username,))
        self.last_keepalive = last_keepalive
        self.username = username
        self.quitting = False

    def keepalive(self, last_keepalive):
        """Updates the keepalive timeout"""
        self.last_keepalive = last_keepalive

    def run(self):
        from time import sleep
        from datetime import datetime, timedelta
        thirty_seconds = timedelta(seconds=30)

        while not self.quitting:
            if datetime.now() > self.last_keepalive + thirty_seconds:
                term = eftw.sessions[self.username]['term']
                # TODO: Change this to use logging
                print "{username} EFTW session timeout...  Killing EFTW shell.".format(username=self.username)
                eftw.sessions[self.username]['multiplexer'].proc_kill(term)
                eftw.sessions[self.username]['multiplexer'].die()
                # Clean up:
                del eftw.sessions[self.username]
                self.quitting = True
            sleep(2)