# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

import os, sys, time, gettext, threading, socket
try: # Note: The json module is only present in Python 2.6+
    import json
except: # If not found, try importing simplejson which will be required if using Python 2.4 or Python 2.5
    import simplejson as json # simplejson was included in Python 2.6 as 'json'
# 3rd party modules
import cherrypy
# PyCI modules
import plugin_utils
from auth import require, has_read, has_write
from pyci_db import check_user_permission, check_group_permission, change_user_permission, change_group_permission, connect_db, disconnect_db, delete_user_permission_by_id, delete_group_permission_by_id
from utils import lookup_template, search_cells, jqgrid_select_options, plugin_template, jquery_ui_error_msg, jquery_ui_highlight_msg, jqgrid_json
from odict import odict
import calendar
from datetime import datetime

"""Admin Interface Plugin (for configuring things like user/group permissions, themes, etc)"""

# TODO:  Make PasswdFile use the make_backup and undo_from_backup methods along with some timing events
# TODO:  Figure out a way to migrate all but the base Admin object out of this file (might need to re-do how base.mako figures out the plugin navigation tree)

# Take care of internationalization
trans = plugin_utils.get_translation(__file__)
_ = lambda s: trans.ugettext(s)

class Utilities(plugin_utils.Plugin):
    """A plugin for managing PyCI's configuration."""
    display = True
    permissions = ['Utilities',]
    sort_priority = 1
    navigable_functions = odict([
        (_('Network Tools'), 'NetworkTools'),
        (_('Speed Test'), 'SpeedTest'),
    ])

    @require(has_read(permissions[0]))
    def NetworkTools(self, **data):
        """Render the Network Utilities page"""
        _template = plugin_template(self, "templates/network_tools.mako")
        return self.standard_render(
            template=_template,
            help=_("Each tab represents a given command line utility.  Click on the tab related to the utility you wish to use."),
            long_name=_('Network Tools'),
            description=_("Various network tools related to testing connectivity, discovery, and enumeration."),
        )

    @require(has_read(permissions[0]))
    def Ping(self, **data):
        """Render the Ping utility tab"""
        _template = plugin_template(self, "templates/ping.mako")
        return self.standard_render(
            template=_template,
            help=_("This is a web-based interface into the ping utility."),
            long_name=_('Ping a host'),
            description=_("Ping utility."),
        )

    @require(has_write(permissions[0]))
    def ping(self, host, **data):
        """Ping a given host and return the responses in real-time.  Note the lower-case name."""
        from subprocess import Popen, PIPE
        cherrypy.response.headers['Content-Type'] = 'text/plain'
        command = "ping -c 4 '%s'" % host
        def run_command():
            process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
            while True:
                out = process.stdout.read(1)
                if out == '' and process.poll() != None:
                    break
                if out != '':
                    yield out
        return run_command()
    ping._cp_config = {'response.stream': True}

    @require(has_read(permissions[0]))
    def Traceroute(self, **data):
        """Render the Traceroute utility tab"""
        _template = plugin_template(self, "templates/traceroute.mako")
        return self.standard_render(
            template=_template,
            help=_("This is a web-based interface into the ping utility."),
            long_name=_('Traceroute to a host'),
            description=_("Traceroute utility."),
        )

    @require(has_write(permissions[0]))
    def traceroute(self, host, **data):
        """Traceroute to a given host and return the responses in real-time.  Note the lower-case name."""
        from subprocess import Popen, PIPE
        cherrypy.response.headers['Content-Type'] = 'text/plain'
        command = "traceroute '%s'" % host
        def run_command():
            process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
            while True:
                out = process.stdout.read(1)
                if out == '' and process.poll() != None:
                    break
                if out != '':
                    yield out
        return run_command()
    traceroute._cp_config = {'response.stream': True}

    @require(has_read(permissions[0]))
    def TCPDump(self, **data):
        """Render the Traceroute utility tab"""
        _template = plugin_template(self, "templates/tcpdump.mako")
        return self.standard_render(
            template=_template,
            help=_("This is a web-based interface into the tcpdump utility."),
            long_name=_('Run tcpdump'),
            description=_("tcpdump utility."),
        )

    @require(has_write(permissions[0]))
    def tcpdump(self, interface, bytes, **data):
        """Traceroute to a given host and return the responses in real-time.  Only outputs 'bytes' amount of data.  Note the lower-case name."""
        from subprocess import Popen, PIPE
        cherrypy.response.headers['Content-Type'] = 'text/plain'
        bytes = int(bytes)
        command = "tcpdump -Al -s 1516 -i '%s'" % interface
        def run_command():
            bytecount = 0
            process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
            while bytecount < bytes:
                bytecount += 1
                #print "bytecount: %s" % bytecount
                out = process.stdout.read(1)
                if out == '' and process.poll() != None:
                    break
                if out != '':
                    yield out
        return run_command()
    traceroute._cp_config = {'response.stream': True}

    @require(has_read(permissions[0]))
    def SpeedTest(self, **data):
        """Render the Speed Test page"""
        _template = plugin_template(self, "templates/speed_test.mako")
        return self.standard_render(
            template=_template,
            help=_("This speed test requires that you download a client utility that constantly uploads packets to a thread listening on port 8081 by default."),
            long_name=_('Utilities: Speed Test'),
            description=_('Download the <a href="/Utilities/_static?filepath=static/speed_test_client.py">Speed Test Client</a> to graph out the speed between your computer and this host.'),
        )

    @require(has_write('Services'))
    def speedtest_service_control(self, oper, address="0.0.0.0", port=8081):
        """Stop, Start, or Restart the SpeedTester thread based on oper which should be one of:
        'start'
        'stop'
        'restart'"""
        #print cherrypy.request.params
        import speedtester
        valid_operations = ['start', 'stop', 'restart', 'status']
        if oper in valid_operations:
            try:
                if oper == 'start':
                    speedtester.start_speedtest_server(address=address, port=port, verbose=False)
                    return jquery_ui_highlight_msg(
                    _('<strong>Success:</strong> SpeedTester thread started successfully.  You may now start your Speed Test Client.')
                    )
                elif oper == 'stop':
                    speedtester.stop_speedtest_server()
                    return jquery_ui_highlight_msg(
                    _('<strong>Success:</strong> SpeedTester thread stopped successfully.')
                    )
                elif oper == 'restart':
                    speedtester.stop_speedtest_server()
                    speedtester.start_speedtest_server(address=address, port=port)
                    return jquery_ui_highlight_msg(
                    _('<strong>Success:</strong> SpeedTester thread restarted successfully.')
                    )
                elif oper == 'status':
                    if speedtester.speedtest_server_status():
                        return _('SpeedTestServer Started.')
                    else:
                        return _('SpeedTestServer Stopped.')
            except Exception as e:
                return jquery_ui_error_msg(
                    _('Got an error trying to start or stop the SpeedTester thread.<br />%s<br /><b>Note:</b> It can sometimes take 30 seconds for the port to become available after the SpeedTestServer shuts down.') % e
                )

    def average_speed(self, **options):
        """Return the current average bandwidth speed recorded by the SpeedTester thread."""
        import speedtester
        if speedtester.speedtest_server_status():
            status = 'started'
        else:
            status = 'stopped'
        now = calendar.timegm(datetime.now().timetuple()) * 1000
        average_KBs = open('/tmp/speed_test_speed').read()
        json_out = []
        json_out.append({'data':[[now, average_KBs]], 'label': 'MB/s', 'server_status': status})
        cherrypy.response.headers['Expires'] = 'Sun, 19 Nov 1978 05:00:00 GMT'
        cherrypy.response.headers['Cache-Control'] = 'no-store, no-cache, must-revalidate, post-check=0, pre-check=0'
        cherrypy.response.headers['Pragma'] = 'no-cache'
        return json.dumps(json_out)